@@ -36,51 +36,33 @@
  * ***** END LICENSE BLOCK ***** */
 
 function runTestOnPrivacyPrefPane(testFunc) {
-  let ww = Cc["@mozilla.org/embedcomp/window-watcher;1"].
-           getService(Ci.nsIWindowWatcher);
   let observer = {
     observe: function(aSubject, aTopic, aData) {
       if (aTopic == "domwindowopened") {
-        ww.unregisterNotification(this);
+        Services.ww.unregisterNotification(this);
 
         let win = aSubject.QueryInterface(Ci.nsIDOMEventTarget);
         win.addEventListener("load", function() {
           win.removeEventListener("load", arguments.callee, false);
           testFunc(dialog.document.defaultView);
 
-          ww.registerNotification(observer);
+          Services.ww.registerNotification(observer);
           dialog.close();
         }, false);
       } else if (aTopic == "domwindowclosed") {
-        ww.unregisterNotification(this);
+        Services.ww.unregisterNotification(this);
         testRunner.runNext();
       }
     }
   };
-  ww.registerNotification(observer);
+  Services.ww.registerNotification(observer);
 
   let dialog = openDialog("chrome://browser/content/preferences/preferences.xul", "Preferences",
                           "chrome,titlebar,toolbar,centerscreen,dialog=no", "panePrivacy");
 }
 
-function controlChanged(doc, element) {
-  let event = doc.createEvent("Events");
-  event.initEvent("command", true, true);
-  element.dispatchEvent(event);
-}
-
-function test_locbar_emptyText(win) {
-  let texts = ["none", "bookmarkhistory", "history", "bookmark"];
-
-  let locbarlist = win.document.getElementById("locationBarSuggestion");
-  ok(locbarlist, "location bar suggestion menulist should exist");
-
-  for (let level = -1; level <= 2; ++level) {
-    locbarlist.value = level;
-    controlChanged(win.document, locbarlist);
-    is(gURLBar.emptyText, gURLBar.getAttribute(texts[level + 1] + "emptytext"),
-      "location bar empty text for for level " + level + " is correctly set");
-  }
+function controlChanged(element) {
+  element.doCommand();
 }
 
 function test_pane_visibility(win) {
@@ -97,7 +79,7 @@ function test_pane_visibility(win) {
 
   for (let mode in modes) {
     historymode.value = mode;
-    controlChanged(win.document, historymode);
+    controlChanged(historymode);
     is(historypane.selectedPanel, win.document.getElementById(modes[mode]),
       "The correct pane should be selected for the " + mode + " mode");
   }
@@ -109,9 +91,7 @@ function test_dependent_elements(win) {
   let pbautostart = win.document.getElementById("privateBrowsingAutoStart");
   ok(pbautostart, "the private browsing auto-start checkbox should exist");
   let controls = [
-    win.document.getElementById("rememberHistoryDays"),
-    win.document.getElementById("historyDays"),
-    win.document.getElementById("rememberAfter"),
+    win.document.getElementById("rememberHistory"),
     win.document.getElementById("rememberDownloads"),
     win.document.getElementById("rememberForms"),
     win.document.getElementById("keepUntil"),
@@ -134,7 +114,7 @@ function test_dependent_elements(win) {
   ok(keepuntil, "the keep cookies until menulist should exist");
   let alwaysclear = win.document.getElementById("alwaysClear");
   ok(alwaysclear, "the clear data on close checkbox should exist");
-  let rememberhistory = win.document.getElementById("rememberHistoryDays");
+  let rememberhistory = win.document.getElementById("rememberHistory");
   ok(rememberhistory, "the remember history checkbox should exist");
   let rememberdownloads = win.document.getElementById("rememberDownloads");
   ok(rememberdownloads, "the remember downloads checkbox should exist");
@@ -174,37 +154,37 @@ function test_dependent_elements(win) {
 
   // controls should only change in custom mode
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_disabled(false);
   check_independents(false);
 
   // setting the mode to custom shouldn't change anything
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_disabled(false);
   check_independents(false);
 
   // controls should only change in custom mode
   historymode.value = "dontremember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_disabled(false);
   check_independents(false);
 
   // controls should only change in custom mode
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_disabled(true);
   check_independents(false);
 
   // dependent controls should follow pbautostart
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(false);
   check_independents(false);
 
   // dependent controls should follow pbautostart
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
   check_independents(false);
 }
@@ -233,46 +213,46 @@ function test_dependent_cookie_elements(win) {
   }
 
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(false);
 
   acceptcookies.checked = false;
-  controlChanged(win.document, acceptcookies);
+  controlChanged(acceptcookies);
   expect_disabled(true);
 
   // pbautostart shouldn't change anything now
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
 
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
 
   acceptcookies.checked = true;
-  controlChanged(win.document, acceptcookies);
+  controlChanged(acceptcookies);
   expect_disabled(false);
 
   let accessthirdparty = controls.shift();
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
   ok(!accessthirdparty.disabled, "access third party button should be enabled");
 
   acceptcookies.checked = false;
-  controlChanged(win.document, acceptcookies);
+  controlChanged(acceptcookies);
   expect_disabled(true);
   ok(accessthirdparty.disabled, "access third party button should be disabled");
 
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
   ok(accessthirdparty.disabled, "access third party button should be disabled");
 
   acceptcookies.checked = true;
-  controlChanged(win.document, acceptcookies);
+  controlChanged(acceptcookies);
   expect_disabled(false);
   ok(!accessthirdparty.disabled, "access third party button should be enabled");
 }
@@ -293,27 +273,27 @@ function test_dependent_clearonclose_elements(win) {
   }
 
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   alwaysclear.checked = false;
-  controlChanged(win.document, alwaysclear);
+  controlChanged(alwaysclear);
   expect_disabled(true);
 
   alwaysclear.checked = true;
-  controlChanged(win.document, alwaysclear);
+  controlChanged(alwaysclear);
   expect_disabled(false);
 
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(true);
 
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   expect_disabled(false);
 
   alwaysclear.checked = false;
-  controlChanged(win.document, alwaysclear);
+  controlChanged(alwaysclear);
   expect_disabled(true);
 }
 
@@ -321,7 +301,7 @@ function test_dependent_prefs(win) {
   let historymode = win.document.getElementById("historyMode");
   ok(historymode, "history mode menulist should exist");
   let controls = [
-    win.document.getElementById("rememberHistoryDays"),
+    win.document.getElementById("rememberHistory"),
     win.document.getElementById("rememberDownloads"),
     win.document.getElementById("rememberForms"),
     win.document.getElementById("acceptCookies"),
@@ -340,19 +320,19 @@ function test_dependent_prefs(win) {
 
   // controls should be checked in remember mode
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_checked(true);
 
   // even if they're unchecked in custom mode
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   controls.forEach(function(control) {
     control.checked = false;
-    controlChanged(win.document, control);
+    controlChanged(control);
   });
   expect_checked(false);
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   expect_checked(true);
 }
 
@@ -367,7 +347,7 @@ function test_historymode_retention(mode, expect) {
     }
 
     historymode.value = mode;
-    controlChanged(win.document, historymode);
+    controlChanged(historymode);
   };
 }
 
@@ -382,7 +362,7 @@ function test_custom_retention(controlToChange, expect, valueIncrement) {
     }
 
     historymode.value = "custom";
-    controlChanged(win.document, historymode);
+    controlChanged(historymode);
 
     controlToChange = win.document.getElementById(controlToChange);
     ok(controlToChange, "the control to change should exist");
@@ -397,7 +377,7 @@ function test_custom_retention(controlToChange, expect, valueIncrement) {
       controlToChange.value = valueIncrement;
       break;
     }
-    controlChanged(win.document, controlToChange);
+    controlChanged(controlToChange);
   };
 }
 
@@ -412,7 +392,7 @@ function test_locbar_suggestion_retention(mode, expect) {
     }
 
     locbarsuggest.value = mode;
-    controlChanged(win.document, locbarsuggest);
+    controlChanged(locbarsuggest);
   };
 }
 
@@ -427,31 +407,31 @@ function test_privatebrowsing_toggle(win) {
 
   // initial state
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
 
   // switch to dontremember mode
   historymode.value = "dontremember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(pbService.privateBrowsingEnabled, "private browsing should be activated");
 
   // switch to remember mode
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(!pbService.privateBrowsingEnabled, "private browsing should be deactivated");
 
   // switch to custom mode
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(!pbService.privateBrowsingEnabled, "private browsing should remain deactivated");
 
   // check the autostart checkbox
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   ok(pbService.privateBrowsingEnabled, "private browsing should be activated");
 
   // uncheck the autostart checkbox
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   ok(!pbService.privateBrowsingEnabled, "private browsing should be deactivated");
 }
 
@@ -468,7 +448,7 @@ function test_privatebrowsing_ui(win) {
 
   // initial state
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(!pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should not be initially disabled");
   ok(!pbcommand.hasAttribute("disabled"),
@@ -476,7 +456,7 @@ function test_privatebrowsing_ui(win) {
 
   // switch to dontremember mode
   historymode.value = "dontremember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should be disabled");
   ok(pbcommand.hasAttribute("disabled"),
@@ -484,7 +464,7 @@ function test_privatebrowsing_ui(win) {
 
   // switch to remember mode
   historymode.value = "remember";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(!pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should be enabled");
   ok(!pbcommand.hasAttribute("disabled"),
@@ -492,7 +472,7 @@ function test_privatebrowsing_ui(win) {
 
   // switch to custom mode
   historymode.value = "custom";
-  controlChanged(win.document, historymode);
+  controlChanged(historymode);
   ok(!pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should remain enabled");
   ok(!pbcommand.hasAttribute("disabled"),
@@ -500,7 +480,7 @@ function test_privatebrowsing_ui(win) {
 
   // check the autostart checkbox
   pbautostart.checked = true;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   ok(pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should be disabled");
   ok(pbcommand.hasAttribute("disabled"),
@@ -508,7 +488,7 @@ function test_privatebrowsing_ui(win) {
 
   // uncheck the autostart checkbox
   pbautostart.checked = false;
-  controlChanged(win.document, pbautostart);
+  controlChanged(pbautostart);
   ok(!pbmenuitem.hasAttribute("disabled"),
     "private browsing menu item should be enabled");
   ok(!pbcommand.hasAttribute("disabled"),
@@ -532,10 +512,8 @@ function reset_preferences(win) {
 
 let testRunner;
 function run_test_subset(subset) {
-  let psvc = Cc["@mozilla.org/preferences-service;1"].
-             getService(Ci.nsIPrefBranch);
-  let instantApplyOrig = psvc.getBoolPref("browser.preferences.instantApply");
-  psvc.setBoolPref("browser.preferences.instantApply", true);
+  let instantApplyOrig = Services.prefs.getBoolPref("browser.preferences.instantApply");
+  Services.prefs.setBoolPref("browser.preferences.instantApply", true);
 
   waitForExplicitFinish();
 
@@ -545,7 +523,7 @@ function run_test_subset(subset) {
     runNext: function() {
       if (this.counter == this.tests.length) {
         // cleanup
-        psvc.setBoolPref("browser.preferences.instantApply", instantApplyOrig);
+        Services.prefs.setBoolPref("browser.preferences.instantApply", instantApplyOrig);
         finish();
       } else {
         let self = this;
