@@ -36,6 +36,19 @@
 
 function test() {
   /** Test for Bug 346337 **/
+
+  var file = Components.classes["@mozilla.org/file/directory_service;1"]
+               .getService(Components.interfaces.nsIProperties)
+               .get("TmpD", Components.interfaces.nsILocalFile);
+  file.append("346337_test1.file");
+  file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0666);
+  filePath1 = file.path;
+  file = Components.classes["@mozilla.org/file/directory_service;1"]
+             .getService(Components.interfaces.nsIProperties)
+             .get("TmpD", Components.interfaces.nsILocalFile);
+  file.append("346337_test2.file");
+  file.createUnique(Components.interfaces.nsIFile.NORMAL_FILE_TYPE, 0666);
+  filePath2 = file.path;
   
   let fieldList = {
     "//input[@name='input']":     Date.now().toString(),
@@ -51,8 +64,8 @@ function test() {
     "//textarea[1]":              "",
     "//textarea[2]":              "Some text... " + Math.random(),
     "//textarea[3]":              "Some more text\n" + new Date(),
-    "//input[@type='file'][1]":   ["/dev/null"],
-    "//input[@type='file'][2]":   ["/dev/null", "/dev/stdin"]
+    "//input[@type='file'][1]":   [filePath1],
+    "//input[@type='file'][2]":   [filePath1, filePath2]
   };
   
   function getElementByXPath(aTab, aQuery) {
@@ -98,21 +111,23 @@ function test() {
   }
   
   // test setup
-  let tabbrowser = getBrowser();
+  let tabbrowser = gBrowser;
   waitForExplicitFinish();
   
   // make sure we don't save form data at all (except for tab duplication)
   gPrefService.setIntPref("browser.sessionstore.privacy_level", 2);
   
-  let testURL = "chrome://mochikit/content/browser/" +
-    "browser/components/sessionstore/test/browser/browser_346337_sample.html";
+  let rootDir = getRootDirectory(gTestPath);
+  let testURL = rootDir + "browser_346337_sample.html";
   let tab = tabbrowser.addTab(testURL);
   tab.linkedBrowser.addEventListener("load", function(aEvent) {
+    this.removeEventListener("load", arguments.callee, true);
     for (let xpath in fieldList)
       setFormValue(tab, xpath, fieldList[xpath]);
     
     let tab2 = tabbrowser.duplicateTab(tab);
     tab2.linkedBrowser.addEventListener("load", function(aEvent) {
+      this.removeEventListener("load", arguments.callee, true);
       for (let xpath in fieldList)
         ok(compareFormValue(tab2, xpath, fieldList[xpath]),
            "The value for \"" + xpath + "\" was correctly restored");
@@ -123,6 +138,7 @@ function test() {
       
       tab = undoCloseTab();
       tab.linkedBrowser.addEventListener("load", function(aEvent) {
+        this.removeEventListener("load", arguments.callee, true);
         for (let xpath in fieldList)
           if (fieldList[xpath])
             ok(!compareFormValue(tab, xpath, fieldList[xpath]),
@@ -132,7 +148,7 @@ function test() {
           gPrefService.clearUserPref("browser.sessionstore.privacy_level");
         // undoCloseTab can reuse a single blank tab, so we have to
         // make sure not to close the window when closing our last tab
-        if (tabbrowser.tabContainer.childNodes.length == 1)
+        if (tabbrowser.tabs.length == 1)
           tabbrowser.addTab();
         tabbrowser.removeTab(tab);
         finish();
