@@ -53,8 +53,10 @@
 #include "nsIFrame.h"
 #include "nsCycleCollectionParticipant.h"
 #include "nsIMarkupDocumentViewer.h"
+#include "nsIScrollableFrame.h"
+#include "nsFocusManager.h"
+#include "nsIDocument.h"
 
-class nsIScrollableView;
 class nsIPresShell;
 class nsIDocShell;
 class nsIDocShellTreeNode;
@@ -62,13 +64,11 @@ class nsIDocShellTreeItem;
 class imgIContainer;
 class nsDOMDataTransfer;
 
-// mac uses click-hold context menus, a holdover from 4.x
-// touch screens (like maemo) could use this also, 
-// perhaps we should move to NS_TOUCHSCREEN
-#if defined(XP_MACOSX) || defined(MOZ_PLATFORM_MAEMO)
-#define CLICK_HOLD_CONTEXT_MENUS 1
-#endif
-
+namespace mozilla {
+namespace dom {
+class TabParent;
+}
+}
 
 /*
  * Event listener manager
@@ -120,8 +120,9 @@ public:
   NS_IMETHOD GetEventTarget(nsIFrame **aFrame);
   NS_IMETHOD GetEventTargetContent(nsEvent* aEvent, nsIContent** aContent);
 
-  NS_IMETHOD GetContentState(nsIContent *aContent, PRInt32& aState);
-  virtual PRBool SetContentState(nsIContent *aContent, PRInt32 aState);
+  virtual nsEventStates GetContentState(nsIContent *aContent,
+                                        PRBool aFollowLabels = PR_FALSE);
+  virtual PRBool SetContentState(nsIContent *aContent, nsEventStates aState);
   NS_IMETHOD ContentRemoved(nsIDocument* aDocument, nsIContent* aContent);
   NS_IMETHOD EventStatusOK(nsGUIEvent* aEvent, PRBool *aOK);
 
@@ -151,9 +152,19 @@ public:
 
   NS_IMETHOD_(PRBool) IsHandlingUserInputExternal() { return IsHandlingUserInput(); }
   
+  nsPresContext* GetPresContext() { return mPresContext; }
+
   NS_DECL_CYCLE_COLLECTION_CLASS_AMBIGUOUS(nsEventStateManager,
                                            nsIEventStateManager)
 
+  static nsIDocument* sMouseOverDocument;
+
+  static nsIEventStateManager* GetActiveEventStateManager() { return sActiveESM; }
+
+  // Sets aNewESM to be the active event state manager, and
+  // if aContent is non-null, marks the object as active.
+  static void SetActiveManager(nsEventStateManager* aNewESM,
+                               nsIContent* aContent);
 protected:
   void UpdateCursor(nsPresContext* aPresContext, nsEvent* aEvent, nsIFrame* aTargetFrame, nsEventStatus* aStatus);
   /**
@@ -199,6 +210,12 @@ protected:
                            nsIContent* aRelatedTarget,
                            nsIContent* aTargetContent,
                            nsWeakFrame& aTargetFrame);
+  /**
+   * Update the initial drag session data transfer with any changes that occur
+   * on cloned data transfer objects used for events.
+   */
+  void UpdateDragDataTransfer(nsDragEvent* dragEvent);
+
   nsresult SetClickCount(nsPresContext* aPresContext, nsMouseEvent *aEvent, nsEventStatus* aStatus);
   nsresult CheckForAndDispatchClick(nsPresContext* aPresContext, nsMouseEvent *aEvent, nsEventStatus* aStatus);
   void EnsureDocument(nsPresContext* aPresContext);
@@ -250,10 +267,6 @@ protected:
   PRBool IsShellVisible(nsIDocShell* aShell);
 
   // These functions are for mousewheel and pixel scrolling
-  nsresult GetParentScrollingView(nsInputEvent* aEvent,
-                                  nsPresContext* aPresContext,
-                                  nsIFrame* &targetOuterFrame,
-                                  nsPresContext* &presCtxOuter);
   void SendLineScrollEvent(nsIFrame* aTargetFrame,
                            nsMouseScrollEvent* aEvent,
                            nsPresContext* aPresContext,
@@ -263,15 +276,9 @@ protected:
                             nsMouseScrollEvent* aEvent,
                             nsPresContext* aPresContext,
                             nsEventStatus* aStatus);
-  typedef enum {
-    eScrollByPixel,
-    eScrollByLine,
-    eScrollByPage
-  } ScrollQuantity;
-  nsresult DoScrollText(nsPresContext* aPresContext,
-                        nsIFrame* aTargetFrame,
+  nsresult DoScrollText(nsIFrame* aTargetFrame,
                         nsMouseScrollEvent* aMouseEvent,
-                        ScrollQuantity aScrollQuantity,
+                        nsIScrollableFrame::ScrollUnit aScrollQuantity,
                         PRBool aAllowScrollSpeedOverride);
   void DoScrollHistory(PRInt32 direction);
   void DoScrollZoom(nsIFrame *aTargetFrame, PRInt32 adjustment);
@@ -316,29 +323,37 @@ protected:
 
   /*
    * Perform the default handling for the dragstart/draggesture event and set up a
-   * drag for aDataTransfer if it contains any data.
+   * drag for aDataTransfer if it contains any data. Returns true if a drag has
+   * started.
    *
    * aDragEvent - the dragstart/draggesture event
    * aDataTransfer - the data transfer that holds the data to be dragged
    * aDragTarget - the target of the drag
    * aIsSelection - true if a selection is being dragged
    */
-  void DoDefaultDragStart(nsPresContext* aPresContext,
-                          nsDragEvent* aDragEvent,
-                          nsDOMDataTransfer* aDataTransfer,
-                          nsIContent* aDragTarget,
-                          PRBool aIsSelection);
+  PRBool DoDefaultDragStart(nsPresContext* aPresContext,
+                            nsDragEvent* aDragEvent,
+                            nsDOMDataTransfer* aDataTransfer,
+                            nsIContent* aDragTarget,
+                            PRBool aIsSelection);
 
   PRBool IsTrackingDragGesture ( ) const { return mGestureDownContent != nsnull; }
   /**
    * Set the fields of aEvent to reflect the mouse position and modifier keys
    * that were set when the user first pressed the mouse button (stored by
    * BeginTrackingDragGesture). aEvent->widget must be
-   * mCurrentTarget->GetWindow().
+   * mCurrentTarget->GetNearestWidget().
    */
   void FillInEventFromGestureDown(nsMouseEvent* aEvent);
 
   nsresult DoContentCommandEvent(nsContentCommandEvent* aEvent);
+  nsresult DoContentCommandScrollEvent(nsContentCommandEvent* aEvent);
+
+#ifdef MOZ_IPC
+  PRBool RemoteQueryContentEvent(nsEvent *aEvent);
+  mozilla::dom::TabParent *GetCrossProcessTarget();
+  PRBool IsTargetCrossProcess(nsGUIEvent *aEvent);
+#endif
 
   PRInt32     mLockCursor;
 
@@ -363,8 +378,11 @@ protected:
   PRPackedBool mGestureDownMeta;
 
   nsCOMPtr<nsIContent> mLastLeftMouseDownContent;
+  nsCOMPtr<nsIContent> mLastLeftMouseDownContentParent;
   nsCOMPtr<nsIContent> mLastMiddleMouseDownContent;
+  nsCOMPtr<nsIContent> mLastMiddleMouseDownContentParent;
   nsCOMPtr<nsIContent> mLastRightMouseDownContent;
+  nsCOMPtr<nsIContent> mLastRightMouseDownContentParent;
 
   nsCOMPtr<nsIContent> mActiveContent;
   nsCOMPtr<nsIContent> mHoverContent;
@@ -386,10 +404,11 @@ protected:
   PRUint32 mMClickCount;
   PRUint32 mRClickCount;
 
-  PRPackedBool mNormalLMouseEventInProcess;
-
   PRPackedBool m_haveShutdown;
 
+
+public:
+  static nsresult UpdateUserActivityTimer(void);
   // Array for accesskey support
   nsCOMArray<nsIContent> mAccessKeys;
 
@@ -397,30 +416,51 @@ protected:
   PRPackedBool mLastLineScrollConsumedX;
   PRPackedBool mLastLineScrollConsumedY;
 
-#ifdef CLICK_HOLD_CONTEXT_MENUS
-  enum { kClickHoldDelay = 500 } ;        // 500ms == 1/2 second
+  static PRInt32 sUserInputEventDepth;
+  
+  static PRBool sNormalLMouseEventInProcess;
 
+  static nsEventStateManager* sActiveESM;
+  
+  static void ClearGlobalActiveContent(nsEventStateManager* aClearer);
+
+  // Functions used for click hold context menus
+  PRBool mClickHoldContextMenu;
+  nsCOMPtr<nsITimer> mClickHoldTimer;
   void CreateClickHoldTimer ( nsPresContext* aPresContext, nsIFrame* inDownFrame,
                               nsGUIEvent* inMouseDownEvent ) ;
   void KillClickHoldTimer ( ) ;
   void FireContextClick ( ) ;
   static void sClickHoldCallback ( nsITimer* aTimer, void* aESM ) ;
-  
-  nsCOMPtr<nsITimer> mClickHoldTimer;
-#endif
-
-  static PRInt32 sUserInputEventDepth;
 };
 
-
+/**
+ * This class is used while processing real user input. During this time, popups
+ * are allowed. For mousedown events, mouse capturing is also permitted.
+ */
 class nsAutoHandlingUserInputStatePusher
 {
 public:
-  nsAutoHandlingUserInputStatePusher(PRBool aIsHandlingUserInput)
-    : mIsHandlingUserInput(aIsHandlingUserInput)
+  nsAutoHandlingUserInputStatePusher(PRBool aIsHandlingUserInput,
+                                     nsEvent* aEvent,
+                                     nsIDocument* aDocument)
+    : mIsHandlingUserInput(aIsHandlingUserInput),
+      mIsMouseDown(aEvent && aEvent->message == NS_MOUSE_BUTTON_DOWN),
+      mResetFMMouseDownState(PR_FALSE)
   {
     if (aIsHandlingUserInput) {
       nsEventStateManager::StartHandlingUserInput();
+      if (mIsMouseDown) {
+        nsIPresShell::SetCapturingContent(nsnull, 0);
+        nsIPresShell::AllowMouseCapture(PR_TRUE);
+        if (aDocument && NS_IS_TRUSTED_EVENT(aEvent)) {
+          nsFocusManager* fm = nsFocusManager::GetFocusManager();
+          if (fm) {
+            fm->SetMouseButtonDownHandlingDocument(aDocument);
+            mResetFMMouseDownState = PR_TRUE;
+          }
+        }
+      }
     }
   }
 
@@ -428,11 +468,22 @@ public:
   {
     if (mIsHandlingUserInput) {
       nsEventStateManager::StopHandlingUserInput();
+      if (mIsMouseDown) {
+        nsIPresShell::AllowMouseCapture(PR_FALSE);
+        if (mResetFMMouseDownState) {
+          nsFocusManager* fm = nsFocusManager::GetFocusManager();
+          if (fm) {
+            fm->SetMouseButtonDownHandlingDocument(nsnull);
+          }
+        }
+      }
     }
   }
 
 protected:
   PRBool mIsHandlingUserInput;
+  PRBool mIsMouseDown;
+  PRBool mResetFMMouseDownState;
 
 private:
   // Hide so that this class can only be stack-allocated
