@@ -38,13 +38,14 @@
 
 #include "nsSVGPathDataParser.h"
 #include "nsSVGDataParser.h"
-#include "nsSVGPathSeg.h"
 #include "nsSVGPathElement.h"
 #include "prdtoa.h"
 #include "nsSVGUtils.h"
 #include <stdlib.h>
 #include <math.h>
 
+using namespace mozilla;
+
 nsresult nsSVGPathDataParser::Match()
 {
   return MatchSvgPath();
@@ -866,197 +867,7 @@ PRBool nsSVGPathDataParser::IsTokenEllipticalArcArgStarter()
 //-----------------------------------------------------------------------
 
 
-// ---------------------------------------------------------------
-// nsSVGPathDataParserToInternal
 
-nsresult
-nsSVGPathDataParserToInternal::Parse(const nsAString &aValue)
-{
-  mPathData->Clear();
-  mPx = mPy = mCx = mCy = mStartX = mStartY = 0;
-  mNumCommands = mNumArguments = 0;
-  mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_UNKNOWN;
-
-  nsresult rv = nsSVGPathDataParser::Parse(aValue);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  rv = PathFini();
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  return rv;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreMoveTo(PRBool absCoords, float x, float y)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_MOVETO_ABS;
-  } else {
-    x += mPx;
-    y += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_MOVETO_REL;
-  }
-  return PathMoveTo(x, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreClosePath()
-{
-  mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CLOSEPATH;
-
-  return PathClose();
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreLineTo(PRBool absCoords, float x, float y)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_ABS;
-  } else {
-    x += mPx;
-    y += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_REL;
-  }
-  return PathLineTo(x, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreHLineTo(PRBool absCoords, float x)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_HORIZONTAL_ABS;
-  } else {
-    x += mPx;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_HORIZONTAL_REL;
-  }
-  return PathLineTo(x, mPy);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreVLineTo(PRBool absCoords, float y)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_VERTICAL_ABS;
-  } else {
-    y += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_LINETO_VERTICAL_REL;
-  }
-  return PathLineTo(mPx, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreCurveTo(PRBool absCoords,
-                                            float x, float y,
-                                            float x1, float y1,
-                                            float x2, float y2)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_ABS;
-  } else {
-    x += mPx;  x1 += mPx;  x2 += mPx;
-    y += mPy;  y1 += mPy;  y2 += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_REL;
-  }
-  mCx = x2;
-  mCy = y2;
-  return PathCurveTo(x1, y1, x2, y2, x, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreSmoothCurveTo(PRBool absCoords,
-                                                  float x, float y,
-                                                  float x2, float y2)
-{
-  float x1, y1;
-
-  // first controlpoint = reflection last one about current point
-  if (mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_REL        ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_ABS        ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_REL ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_ABS ) {
-    x1 = 2 * mPx - mCx;
-    y1 = 2 * mPy - mCy;
-  } else {
-    x1 = mPx;
-    y1 = mPy;
-  }
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_ABS;
-  } else {
-    x += mPx;
-    x2 += mPx;
-    y += mPy;
-    y2 += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_REL;
-  }
-  mCx = x2;
-  mCy = y2;
-  return PathCurveTo(x1, y1, x2, y2, x, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreQuadCurveTo(PRBool absCoords,
-                                                float x, float y,
-                                                float x1, float y1)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_ABS;
-  } else {
-    x += mPx;
-    x1 += mPx;
-    y += mPy;
-    y1 += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_REL;
-  }
-
-  float x31, y31, x32, y32;
-  // conversion of quadratic bezier curve to cubic bezier curve:
-  x31 = mPx + (x1 - mPx) * 2 / 3;
-  y31 = mPy + (y1 - mPy) * 2 / 3;
-  x32 = x1 + (x - x1) / 3;
-  y32 = y1 + (y - y1) / 3;
-
-  mCx = x1;
-  mCy = y1;
-  return PathCurveTo(x31, y31, x32, y32, x, y);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreSmoothQuadCurveTo(PRBool absCoords,
-                                                      float x, float y)
-{
-  float x1, y1;
-
-  // first controlpoint = reflection last one about current point
-  if (mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_REL        ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_ABS        ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL ||
-      mPrevSeg == nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS ) {
-    x1 = 2 * mPx - mCx;
-    y1 = 2 * mPy - mCy;
-  } else {
-    x1 = mPx;
-    y1 = mPy;
-  }
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS;
-  } else {
-    x += mPx;
-    y += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL;
-  }
-
-  float x31, y31, x32, y32;
-  // conversion of quadratic bezier curve to cubic bezier curve:
-  x31 = mPx + (x1 - mPx) * 2 / 3;
-  y31 = mPy + (y1 - mPy) * 2 / 3;
-  x32 = x1 + (x - x1) / 3;
-  y32 = y1 + (y - y1) / 3;
-
-  mCx = x1;
-  mCy = y1;
-  return PathCurveTo(x31, y31, x32, y32, x, y);
-}
 
 static double
 CalcVectorAngle(double ux, double uy, double vx, double vy)
@@ -1068,302 +879,26 @@ CalcVectorAngle(double ux, double uy, double vx, double vy)
   return 2 * M_PI - (ta-tb);
 }
 
-nsresult
-nsSVGPathDataParserToInternal::ConvertArcToCurves(float x2, float y2,
-                                                  float rx, float ry,
-                                                  float angle,
-                                                  PRBool largeArcFlag,
-                                                  PRBool sweepFlag)
-{
-  float x1=mPx, y1=mPy, x3, y3;
-  // Treat out-of-range parameters as described in
-  // http://www.w3.org/TR/SVG/implnote.html#ArcImplementationNotes
-  
-  // If the endpoints (x1, y1) and (x2, y2) are identical, then this
-  // is equivalent to omitting the elliptical arc segment entirely
-  if (x1 == x2 && y1 == y2) {
-    return NS_OK;
-  }
-  // If rX = 0 or rY = 0 then this arc is treated as a straight line
-  // segment (a "lineto") joining the endpoints.
-  if (rx == 0.0f || ry == 0.0f) {
-    return PathLineTo(x2, y2);
-  }
-  nsSVGArcConverter converter(x1, y1, x2, y2, rx, ry, angle,
-                              largeArcFlag, sweepFlag);
-  
-  while (converter.GetNextSegment(&x1, &y1, &x2, &y2, &x3, &y3)) {
-    // c) draw the cubic bezier:
-    nsresult rv = PathCurveTo(x1, y1, x2, y2, x3, y3);
-    NS_ENSURE_SUCCESS(rv, rv);
-  }
 
-  return NS_OK;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::StoreEllipticalArc(PRBool absCoords,
-                                                  float x, float y,
-                                                  float r1, float r2,
-                                                  float angle,
-                                                  PRBool largeArcFlag,
-                                                  PRBool sweepFlag)
-{
-  if (absCoords) {
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_ARC_ABS;
-  } else {
-    x += mPx;
-    y += mPy;
-    mPrevSeg = nsIDOMSVGPathSeg::PATHSEG_ARC_REL;
-  }
-  return ConvertArcToCurves(x, y, r1, r2, angle, largeArcFlag, sweepFlag);
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathEnsureSpace(PRUint32 aNumArgs)
-{
-  if (!(mNumCommands % 4) &&
-      !mCommands.AppendElement())
-    return NS_ERROR_OUT_OF_MEMORY;
-
-  if (!mArguments.SetLength(mArguments.Length()+aNumArgs))
-    return NS_ERROR_OUT_OF_MEMORY;
-
-  return NS_OK;
-}
-
-void
-nsSVGPathDataParserToInternal::PathAddCommandCode(PRUint8 aCommand)
-{
-  PRUint32 offset = mNumCommands / 4;
-  PRUint32 shift = 2 * (mNumCommands % 4);
-  if (shift == 0) {
-    // make sure we set the byte, to avoid false UMR reports
-    mCommands[offset] = aCommand;
-  } else {
-    mCommands[offset] |= aCommand << shift;
-  }
-  mNumCommands++;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathMoveTo(float x, float y)
-{
-  nsresult rv = PathEnsureSpace(2);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  PathAddCommandCode(nsSVGPathList::MOVETO);
-  mArguments[mNumArguments++] = x;
-  mArguments[mNumArguments++] = y;
-
-  mPx = mStartX = x;
-  mPy = mStartY = y;
-
-  return NS_OK;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathLineTo(float x, float y)
-{
-  nsresult rv = PathEnsureSpace(2);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  PathAddCommandCode(nsSVGPathList::LINETO);
-  mArguments[mNumArguments++] = x;
-  mArguments[mNumArguments++] = y;
-
-  mPx = x;
-  mPy = y;
-
-  return NS_OK;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathCurveTo(float x1, float y1,
-                                           float x2, float y2,
-                                           float x3, float y3)
-{
-  nsresult rv = PathEnsureSpace(6);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  PathAddCommandCode(nsSVGPathList::CURVETO);
-  mArguments[mNumArguments++] = x1;
-  mArguments[mNumArguments++] = y1;
-  mArguments[mNumArguments++] = x2;
-  mArguments[mNumArguments++] = y2;
-  mArguments[mNumArguments++] = x3;
-  mArguments[mNumArguments++] = y3;
-
-  mPx = x3;
-  mPy = y3;
-
-  return NS_OK;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathClose()
-{
-  nsresult rv = PathEnsureSpace(0);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  PathAddCommandCode(nsSVGPathList::CLOSEPATH);
-
-  mPx = mStartX;
-  mPy = mStartY;
-
-  return NS_OK;
-}
-
-nsresult
-nsSVGPathDataParserToInternal::PathFini()
-{
-  // We're done adding data to the arrays - copy to a straight array
-  // in mPathData, which allows us to remove the 8-byte overhead per
-  // nsTArray.  For a bonus savings we allocate a single array instead
-  // of two.
-  PRUint32 argArraySize;
-
-  argArraySize = mArguments.Length() * sizeof(float);
-  mPathData->mArguments = (float *)malloc(argArraySize + mCommands.Length());
-  if (!mPathData->mArguments)
-    return NS_ERROR_OUT_OF_MEMORY;
-
-  memcpy(mPathData->mArguments, mArguments.Elements(), argArraySize);
-  memcpy(mPathData->mArguments + mNumArguments,
-         mCommands.Elements(),
-         mCommands.Length());
-  mPathData->mNumArguments = mNumArguments;
-  mPathData->mNumCommands = mNumCommands;
-
-  return NS_OK;
-}
-
-// ---------------------------------------------------------------
-// nsSVGPathDataParserToDOM
-
-nsresult
-nsSVGPathDataParserToDOM::AppendSegment(nsIDOMSVGPathSeg* seg)
-{
-  NS_ENSURE_TRUE(seg, NS_ERROR_OUT_OF_MEMORY);
-  mData->AppendObject(seg);
-  return NS_OK;
-}
-
-
-nsresult
-nsSVGPathDataParserToDOM::StoreMoveTo(PRBool absCoords, float x, float y)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegMovetoAbs(x, y)
-              : NS_NewSVGPathSegMovetoRel(x, y));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreClosePath()
-{
-  return AppendSegment(NS_NewSVGPathSegClosePath());
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreLineTo(PRBool absCoords, float x, float y)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegLinetoAbs(x, y)
-              : NS_NewSVGPathSegLinetoRel(x, y));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreHLineTo(PRBool absCoords, float x)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegLinetoHorizontalAbs(x)
-              : NS_NewSVGPathSegLinetoHorizontalRel(x));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreVLineTo(PRBool absCoords, float y)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegLinetoVerticalAbs(y)
-              : NS_NewSVGPathSegLinetoVerticalRel(y));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreCurveTo(PRBool absCoords,
-                                       float x, float y,
-                                       float x1, float y1,
-                                       float x2, float y2)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegCurvetoCubicAbs(x, y, x1, y1, x2, y2)
-              : NS_NewSVGPathSegCurvetoCubicRel(x, y, x1, y1, x2, y2));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreSmoothCurveTo(PRBool absCoords,
-                                             float x, float y,
-                                             float x2, float y2)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegCurvetoCubicSmoothAbs(x, y, x2, y2)
-              : NS_NewSVGPathSegCurvetoCubicSmoothRel(x, y, x2, y2));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreQuadCurveTo(PRBool absCoords,
-                                           float x, float y,
-                                           float x1, float y1)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegCurvetoQuadraticAbs(x, y, x1, y1)
-              : NS_NewSVGPathSegCurvetoQuadraticRel(x, y, x1, y1));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreSmoothQuadCurveTo(PRBool absCoords,
-                                                 float x, float y)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegCurvetoQuadraticSmoothAbs(x, y)
-              : NS_NewSVGPathSegCurvetoQuadraticSmoothRel(x, y));
-}
-
-nsresult
-nsSVGPathDataParserToDOM::StoreEllipticalArc(PRBool absCoords,
-                                             float x, float y,
-                                             float r1, float r2,
-                                             float angle,
-                                             PRBool largeArcFlag,
-                                             PRBool sweepFlag)
-{
-  return AppendSegment(
-    absCoords ? NS_NewSVGPathSegArcAbs(x, y, r1, r2, angle, 
-                                       largeArcFlag, sweepFlag)
-              : NS_NewSVGPathSegArcRel(x, y, r1, r2, angle,
-                                       largeArcFlag, sweepFlag));
-}
-
-nsSVGArcConverter::nsSVGArcConverter(float x1, float y1,
-                                     float x2, float y2,
-                                     float rx, float ry,
-                                     float angle,
+nsSVGArcConverter::nsSVGArcConverter(const gfxPoint &from,
+                                     const gfxPoint &to,
+                                     const gfxPoint &radii,
+                                     double angle,
                                      PRBool largeArcFlag,
                                      PRBool sweepFlag)
 {
   const double radPerDeg = M_PI/180.0;
 
-  // If rX or rY have negative signs, these are dropped; the absolute
-  // value is used instead.
-  mRx = fabs(rx);
-  mRy = fabs(ry);
-
   // Convert to center parameterization as shown in
   // http://www.w3.org/TR/SVG/implnote.html
+  mRx = fabs(radii.x);
+  mRy = fabs(radii.y);
+
   mSinPhi = sin(angle*radPerDeg);
   mCosPhi = cos(angle*radPerDeg);
 
-  double x1dash =  mCosPhi * (x1-x2)/2.0 + mSinPhi * (y1-y2)/2.0;
-  double y1dash = -mSinPhi * (x1-x2)/2.0 + mCosPhi * (y1-y2)/2.0;
+  double x1dash =  mCosPhi * (from.x-to.x)/2.0 + mSinPhi * (from.y-to.y)/2.0;
+  double y1dash = -mSinPhi * (from.x-to.x)/2.0 + mCosPhi * (from.y-to.y)/2.0;
 
   double root;
   double numerator = mRx*mRx*mRy*mRy - mRx*mRx*y1dash*y1dash -
@@ -1371,13 +906,13 @@ nsSVGArcConverter::nsSVGArcConverter(float x1, float y1,
 
   if (numerator < 0.0) {
     //  If mRx , mRy and are such that there is no solution (basically,
-    //  the ellipse is not big enough to reach from (x1, y1) to (x2,
-    //  y2)) then the ellipse is scaled up uniformly until there is
+    //  the ellipse is not big enough to reach from 'from' to 'to'
+    //  then the ellipse is scaled up uniformly until there is
     //  exactly one solution (until the ellipse is just big enough).
 
     // -> find factor s, such that numerator' with mRx'=s*mRx and
     //    mRy'=s*mRy becomes 0 :
-    float s = (float)sqrt(1.0 - numerator/(mRx*mRx*mRy*mRy));
+    double s = sqrt(1.0 - numerator/(mRx*mRx*mRy*mRy));
 
     mRx *= s;
     mRy *= s;
@@ -1392,8 +927,8 @@ nsSVGArcConverter::nsSVGArcConverter(float x1, float y1,
   double cxdash = root*mRx*y1dash/mRy;
   double cydash = -root*mRy*x1dash/mRx;
 
-  mCx = mCosPhi * cxdash - mSinPhi * cydash + (x1+x2)/2.0;
-  mCy = mSinPhi * cxdash + mCosPhi * cydash + (y1+y2)/2.0;
+  mC.x = mCosPhi * cxdash - mSinPhi * cydash + (from.x+to.x)/2.0;
+  mC.y = mSinPhi * cxdash + mCosPhi * cydash + (from.y+to.y)/2.0;
   mTheta = CalcVectorAngle(1.0, 0.0, (x1dash-cxdash)/mRx, (y1dash-cydash)/mRy);
   double dtheta = CalcVectorAngle((x1dash-cxdash)/mRx, (y1dash-cydash)/mRy,
                                   (-x1dash-cxdash)/mRx, (-y1dash-cydash)/mRy);
@@ -1407,43 +942,179 @@ nsSVGArcConverter::nsSVGArcConverter(float x1, float y1,
   mDelta = dtheta/mNumSegs;
   mT = 8.0/3.0 * sin(mDelta/4.0) * sin(mDelta/4.0) / sin(mDelta/2.0);
 
-  mX1 = x1;
-  mY1 = y1;
+  mFrom = from;
   mSegIndex = 0;
 }
 
 PRBool
-nsSVGArcConverter::GetNextSegment(float *x1, float *y1,
-                                  float *x2, float *y2,
-                                  float *x3, float *y3)
+nsSVGArcConverter::GetNextSegment(gfxPoint *cp1, gfxPoint *cp2, gfxPoint *to)
 {
   if (mSegIndex == mNumSegs) {
-     return PR_FALSE;
+    return PR_FALSE;
   }
   
-  float cosTheta1 = cos(mTheta);
-  float sinTheta1 = sin(mTheta);
-  float theta2 = mTheta + mDelta;
-  float cosTheta2 = cos(theta2);
-  float sinTheta2 = sin(theta2);
+  double cosTheta1 = cos(mTheta);
+  double sinTheta1 = sin(mTheta);
+  double theta2 = mTheta + mDelta;
+  double cosTheta2 = cos(theta2);
+  double sinTheta2 = sin(theta2);
 
   // a) calculate endpoint of the segment:
-  *x3 = mCosPhi * mRx*cosTheta2 - mSinPhi * mRy*sinTheta2 + mCx;
-  *y3 = mSinPhi * mRx*cosTheta2 + mCosPhi * mRy*sinTheta2 + mCy;
+  to->x = mCosPhi * mRx*cosTheta2 - mSinPhi * mRy*sinTheta2 + mC.x;
+  to->y = mSinPhi * mRx*cosTheta2 + mCosPhi * mRy*sinTheta2 + mC.y;
 
   // b) calculate gradients at start/end points of segment:
-  *x1 = mX1 + mT * ( - mCosPhi * mRx*sinTheta1 - mSinPhi * mRy*cosTheta1);
-  *y1 = mY1 + mT * ( - mSinPhi * mRx*sinTheta1 + mCosPhi * mRy*cosTheta1);
+  cp1->x = mFrom.x + mT * ( - mCosPhi * mRx*sinTheta1 - mSinPhi * mRy*cosTheta1);
+  cp1->y = mFrom.y + mT * ( - mSinPhi * mRx*sinTheta1 + mCosPhi * mRy*cosTheta1);
 
-  *x2 = *x3 + mT * ( mCosPhi * mRx*sinTheta2 + mSinPhi * mRy*cosTheta2);
-  *y2 = *y3 + mT * ( mSinPhi * mRx*sinTheta2 - mCosPhi * mRy*cosTheta2);
+  cp2->x = to->x + mT * ( mCosPhi * mRx*sinTheta2 + mSinPhi * mRy*cosTheta2);
+  cp2->y = to->y + mT * ( mSinPhi * mRx*sinTheta2 - mCosPhi * mRy*cosTheta2);
 
   // do next segment
   mTheta = theta2;
-  mX1 = *x3;
-  mY1 = *y3;
+  mFrom = *to;
   ++mSegIndex;
 
   return PR_TRUE;
 }
 
+
+// ---------------------------------------------------------------
+// nsSVGPathDataParserToInternal
+
+nsresult
+nsSVGPathDataParserToInternal::Parse(const nsAString &aValue)
+{
+  mPathSegList->Clear();
+  return nsSVGPathDataParser::Parse(aValue);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreMoveTo(PRBool absCoords, float x, float y)
+{
+  // Because our IDL compiler doesn't know any better, each seg type constant
+  // in nsIDOMSVGPathSeg is in a separate enum. This results in "warning:
+  // enumeral mismatch in conditional expression" under GCC if two bare
+  // nsIDOMSVGPathSeg constants are used as operands of the ?: operator below.
+  // In newer versions of GCC we would be able to turn off this warning using:
+  //
+  //#pragma GCC diagnostic push
+  //#pragma GCC diagnostic ignored "-Wenum-compare"
+  //...
+  //#pragma GCC diagnostic pop
+  //
+  // Unfortunately we need to support older versions of GCC. Instead, to
+  // eliminate this warning noise being sent to the console, we wrap the
+  // operands with PRUint32(...).
+
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_MOVETO_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_MOVETO_REL);
+
+  return mPathSegList->AppendSeg(type, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreClosePath()
+{
+  return mPathSegList->AppendSeg(nsIDOMSVGPathSeg::PATHSEG_CLOSEPATH);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreLineTo(PRBool absCoords, float x, float y)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_REL);
+
+  return mPathSegList->AppendSeg(type, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreHLineTo(PRBool absCoords, float x)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_HORIZONTAL_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_HORIZONTAL_REL);
+
+  return mPathSegList->AppendSeg(type, x);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreVLineTo(PRBool absCoords, float y)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_VERTICAL_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_LINETO_VERTICAL_REL);
+
+  return mPathSegList->AppendSeg(type, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreCurveTo(PRBool absCoords,
+                                            float x, float y,
+                                            float x1, float y1,
+                                            float x2, float y2)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_REL);
+
+  return mPathSegList->AppendSeg(type, x1, y1, x2, y2, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreSmoothCurveTo(PRBool absCoords,
+                                                  float x, float y,
+                                                  float x2, float y2)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_CUBIC_SMOOTH_REL);
+
+  return mPathSegList->AppendSeg(type, x2, y2, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreQuadCurveTo(PRBool absCoords,
+                                                float x, float y,
+                                                float x1, float y1)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_REL);
+
+  return mPathSegList->AppendSeg(type, x1, y1, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreSmoothQuadCurveTo(PRBool absCoords,
+                                                      float x, float y)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_CURVETO_QUADRATIC_SMOOTH_REL);
+
+  return mPathSegList->AppendSeg(type, x, y);
+}
+
+nsresult
+nsSVGPathDataParserToInternal::StoreEllipticalArc(PRBool absCoords,
+                                                  float x, float y,
+                                                  float r1, float r2,
+                                                  float angle,
+                                                  PRBool largeArcFlag,
+                                                  PRBool sweepFlag)
+{
+  PRUint32 type = absCoords ?
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_ARC_ABS) :
+    PRUint32(nsIDOMSVGPathSeg::PATHSEG_ARC_REL);
+
+  // We can only pass floats after 'type', and per the SVG spec for arc,
+  // non-zero args are treated at 'true'.
+  return mPathSegList->AppendSeg(type, r1, r2, angle,
+                                 largeArcFlag ? 1.0f : 0.0f,
+                                 sweepFlag ? 1.0f : 0.0f,
+                                 x, y);
+}
+
