@@ -41,7 +41,6 @@
 #ifndef nsDocShell_h__
 #define nsDocShell_h__
 
-#include "nsIPresShell.h"
 #include "nsIDOMNode.h"
 #include "nsIDOMNodeList.h"
 #include "nsIContentViewer.h"
@@ -109,8 +108,6 @@
 #include "nsIObserver.h"
 #include "nsDocShellLoadTypes.h"
 #include "nsPIDOMEventTarget.h"
-#include "nsIURIClassifier.h"
-#include "nsIChannelClassifier.h"
 #include "nsILoadContext.h"
 #include "nsIWidget.h"
 #include "nsIWebShellServices.h"
@@ -119,10 +116,10 @@
 #include "nsICommandManager.h"
 #include "nsCRT.h"
 
-class nsIScrollableView;
 class nsDocShell;
 class nsIController;
 class OnLinkClickEvent;
+class nsIScrollableFrame;
 
 /* load commands were moved to nsIDocShell.h */
 /* load types were moved to nsDocShellLoadTypes.h */
@@ -157,32 +154,6 @@ protected:
     virtual ~nsRefreshTimer();
 };
 
-class nsClassifierCallback : public nsIChannelClassifier
-                           , public nsIURIClassifierCallback
-                           , public nsIRunnable
-                           , public nsIChannelEventSink
-                           , public nsIInterfaceRequestor
-{
-public:
-    nsClassifierCallback() {}
-    ~nsClassifierCallback() {}
-
-    NS_DECL_ISUPPORTS
-    NS_DECL_NSICHANNELCLASSIFIER
-    NS_DECL_NSIURICLASSIFIERCALLBACK
-    NS_DECL_NSIRUNNABLE
-    NS_DECL_NSICHANNELEVENTSINK
-    NS_DECL_NSIINTERFACEREQUESTOR
-
-private:
-    nsCOMPtr<nsIChannel> mChannel;
-    nsCOMPtr<nsIChannel> mSuspendedChannel;
-    nsCOMPtr<nsIInterfaceRequestor> mNotificationCallbacks;
-
-    void MarkEntryClassified(nsresult status);
-    PRBool HasBeenClassified(nsIChannel *aChannel);
-};
-
 #define NS_ERROR_DOCSHELL_REQUEST_REJECTED  NS_ERROR_GENERATE_FAILURE(NS_ERROR_MODULE_GENERAL,1001)
 
 typedef enum {
@@ -215,7 +186,8 @@ class nsDocShell : public nsDocLoader,
                    public nsILoadContext,
                    public nsIWebShellServices,
                    public nsILinkHandler,
-                   public nsIClipboardCommands
+                   public nsIClipboardCommands,
+                   public nsIDocShell_MOZILLA_2_0_BRANCH
 {
     friend class nsDSURIContentListener;
 
@@ -230,6 +202,7 @@ public:
     NS_DECL_ISUPPORTS_INHERITED
 
     NS_DECL_NSIDOCSHELL
+    NS_DECL_NSIDOCSHELL_MOZILLA_2_0_BRANCH
     NS_DECL_NSIDOCSHELLTREEITEM
     NS_DECL_NSIDOCSHELLTREENODE
     NS_DECL_NSIDOCSHELLHISTORY
@@ -277,7 +250,6 @@ public:
         nsIURI* aURI,
         const PRUnichar* aTargetSpec);
     NS_IMETHOD OnLeaveLink();
-    NS_IMETHOD GetLinkState(nsIURI* aLinkURI, nsLinkState& aState);
 
     nsDocShellInfoLoadType ConvertLoadTypeToDocShellLoadInfo(PRUint32 aLoadType);
     PRUint32 ConvertDocShellLoadInfoToLoadType(nsDocShellInfoLoadType aDocShellLoadType);
@@ -299,6 +271,13 @@ public:
 
     friend class OnLinkClickEvent;
 
+    // We need dummy OnLocationChange in some cases to update the UI.
+    void FireDummyOnLocationChange()
+    {
+      FireOnLocationChange(this, nsnull, mCurrentURI);
+    }
+
+    nsresult HistoryTransactionRemoved(PRInt32 aIndex);
 protected:
     // Object Management
     virtual ~nsDocShell();
@@ -309,7 +288,8 @@ protected:
     // aPrincipal can be passed in if the caller wants.  If null is
     // passed in, the about:blank principal will end up being used.
     nsresult CreateAboutBlankContentViewer(nsIPrincipal* aPrincipal,
-                                           nsIURI* aBaseURI);
+                                           nsIURI* aBaseURI,
+                                           PRBool aTryToSaveOldPresentation = PR_TRUE);
     NS_IMETHOD CreateContentViewer(const char * aContentType, 
         nsIRequest * request, nsIStreamListener ** aContentHandler);
     NS_IMETHOD NewContentViewerObj(const char * aContentType, 
@@ -355,20 +335,12 @@ protected:
                                    nsIURILoader * aURILoader,
                                    PRBool aBypassClassifier);
 
-    // Check the channel load against the URI classifier service (if it
-    // exists).  The channel will be suspended until the classification is
-    // complete.
-    nsresult CheckClassifier(nsIChannel *aChannel);
-
     nsresult ScrollIfAnchor(nsIURI * aURI, PRBool * aWasAnchor,
-                            PRUint32 aLoadType, nscoord *cx, nscoord *cy,
-                            PRBool * aDoHashchange);
+                            PRUint32 aLoadType, PRBool * aDoHashchange);
 
-    // Dispatches the hashchange event to the current thread, if the document's
-    // readystate is "complete".
-    nsresult DispatchAsyncHashchange();
-
-    nsresult FireHashchange();
+    // Tries to stringify a given variant by converting it to JSON.  This only
+    // works if the variant is backed by a JSVal.
+    nsresult StringifyJSValVariant(nsIVariant *aData, nsAString &aResult);
 
     // Returns PR_TRUE if would have called FireOnLocationChange,
     // but did not because aFireOnLocationChange was false on entry.
@@ -411,10 +383,13 @@ protected:
     // |aReplaceEntry|.  |aSrcShell| is a (possibly null) docshell which
     // corresponds to |aSrcEntry| via its mLSHE or mOHE pointers, and will
     // have that pointer updated to point to the cloned history entry.
+    // If aCloneChildren is true then the children of the entry with id
+    // |aCloneID| will be cloned into |aReplaceEntry|.
     static nsresult CloneAndReplace(nsISHEntry *aSrcEntry,
                                     nsDocShell *aSrcShell,
                                     PRUint32 aCloneID,
                                     nsISHEntry *aReplaceEntry,
+                                    PRBool aCloneChildren,
                                     nsISHEntry **aDestEntry);
 
     // Child-walking callback for CloneAndReplace
@@ -470,9 +445,77 @@ protected:
                                        PRUint32 aRedirectFlags,
                                        PRUint32 aStateFlags);
 
-    // Global History
-    nsresult AddToGlobalHistory(nsIURI * aURI, PRBool aRedirect,
-                                nsIChannel * aChannel);
+    /**
+     * Helper function that determines if channel is an HTTP POST.
+     *
+     * @param aChannel
+     *        The channel to test
+     *
+     * @return True iff channel is an HTTP post.
+     */
+    bool ChannelIsPost(nsIChannel* aChannel);
+
+    /**
+     * Helper function that finds the last URI and its transition flags for a
+     * channel.
+     *
+     * This method first checks the channel's property bag to see if previous
+     * info has been saved.  If not, it gives back the referrer of the channel.
+     *
+     * @param aChannel
+     *        The channel we are transitioning to
+     * @param aURI
+     *        Output parameter with the previous URI, not addref'd
+     * @param aChannelRedirectFlags
+     *        If a redirect, output parameter with the previous redirect flags
+     *        from nsIChannelEventSink
+     */
+    void ExtractLastVisit(nsIChannel* aChannel,
+                          nsIURI** aURI,
+                          PRUint32* aChannelRedirectFlags);
+
+    /**
+     * Helper function that caches a URI and a transition for saving later.
+     *
+     * @param aChannel
+     *        Channel that will have these properties saved
+     * @param aURI
+     *        The URI to save for later
+     * @param aChannelRedirectFlags
+     *        The nsIChannelEventSink redirect flags to save for later
+     */
+    void SaveLastVisit(nsIChannel* aChannel,
+                       nsIURI* aURI,
+                       PRUint32 aChannelRedirectFlags);
+
+    /**
+     * Helper function for adding a URI visit using IHistory.  If IHistory is
+     * not available, the method tries nsIGlobalHistory2.
+     *
+     * The IHistory API maintains chains of visits, tracking both HTTP referrers
+     * and redirects for a user session. VisitURI requires the current URI and
+     * the previous URI in the chain.
+     *
+     * Visits can be saved either during a redirect or when the request has
+     * reached its final destination.  The previous URI in the visit may be
+     * from another redirect or it may be the referrer.
+     *
+     * @pre aURI is not null.
+     *
+     * @param aURI
+     *        The URI that was just visited
+     * @param aReferrerURI
+     *        The referrer URI of this request
+     * @param aPreviousURI
+     *        The previous URI of this visit (may be the same as aReferrerURI)
+     * @param aChannelRedirectFlags
+     *        For redirects, the redirect flags from nsIChannelEventSink
+     *        (0 otherwise)
+     */
+    void AddURIVisit(nsIURI* aURI,
+                     nsIURI* aReferrerURI,
+                     nsIURI* aPreviousURI,
+                     PRUint32 aChannelRedirectFlags);
 
     // Helper Routines
     nsresult   ConfirmRepost(PRBool * aRepost);
@@ -480,7 +523,7 @@ protected:
         nsIStringBundle ** aStringBundle);
     NS_IMETHOD GetChildOffset(nsIDOMNode * aChild, nsIDOMNode * aParent,
         PRInt32 * aOffset);
-    NS_IMETHOD GetRootScrollableView(nsIScrollableView ** aOutScrollView);
+    nsIScrollableFrame* GetRootScrollFrame();
     NS_IMETHOD EnsureScriptEnvironment();
     NS_IMETHOD EnsureEditorData();
     nsresult   EnsureTransferableHookData();
@@ -521,6 +564,11 @@ protected:
                                  nsIChannel * aChannel,
                                  nsresult aResult);
 
+    // Sets the current document's current state object to the given SHEntry's
+    // state object.  The current state object is eventually given to the page
+    // in the PopState event.
+    nsresult SetDocCurrentStateObj(nsISHEntry *shEntry);
+
     nsresult CheckLoadingPermissions();
 
     // Security checks to prevent frameset spoofing.  See comments at
@@ -611,6 +659,7 @@ protected:
     void ReattachEditorToWindow(nsISHEntry *aSHEntry);
 
     nsresult GetSessionStorageForURI(nsIURI* aURI,
+                                     const nsSubstring& aDocumentURI,
                                      PRBool create,
                                      nsIDOMStorage** aStorage);
 
@@ -620,11 +669,14 @@ protected:
     nsresult IsCommandEnabled(const char * inCommand, PRBool* outEnabled);
     nsresult DoCommand(const char * inCommand);
     nsresult EnsureCommandHandler();
-    
+
+    nsIChannel* GetCurrentDocChannel();
 protected:
     // Override the parent setter from nsDocLoader
     virtual nsresult SetDocLoaderParent(nsDocLoader * aLoader);
 
+    void ClearFrameHistory(nsISHEntry* aEntry);
+
     // Event type dispatched by RestorePresentation
     class RestorePresentationEvent : public nsRunnable {
     public:
@@ -687,9 +739,6 @@ protected:
     // Secure browser UI object
     nsCOMPtr<nsISecureBrowserUI> mSecurityUI;
 
-    // Suspends/resumes channels based on the URI classifier.
-    nsRefPtr<nsClassifierCallback> mClassifier;
-
     // The URI we're currently loading.  This is only relevant during the
     // firing of a pagehide/unload.  The caller of FirePageHideNotification()
     // is responsible for setting it and unsetting it.  It may be null if the
@@ -714,15 +763,7 @@ protected:
     eCharsetReloadState        mCharsetReloadState;
 
     // Offset in the parent's child list.
-    // XXXmats the line above is bogus, it's the offset in the parent's
-    // child list at the time this docshell was added to it,
-    // see nsDocShell::AddChild().  It isn't updated after that so if children
-    // with lower indices are removed this offset is no longer valid to be used
-    // as an index into the parent's child list (see bug 162283).  It MUST not
-    // be used for that purpose.  It's used as an index to get/add history
-    // entries into nsIDocShellHistory, although I very much doubt that it
-    // can be correct for that purpose as well...
-    // Try not to use it, we should get rid of it.
+    // -1 if the docshell is added dynamically to the parent shell.
     PRUint32                   mChildOffset;
     PRUint32                   mBusyFlags;
     PRUint32                   mAppType;
@@ -730,6 +771,9 @@ protected:
 
     PRInt32                    mMarginWidth;
     PRInt32                    mMarginHeight;
+
+    // This can either be a content docshell or a chrome docshell.  After
+    // Create() is called, the type is not expected to change.
     PRInt32                    mItemType;
 
     // Index into the SHTransaction list, indicating the previous and current
@@ -749,6 +793,9 @@ protected:
     PRPackedBool               mAllowAuth;
     PRPackedBool               mAllowKeywordFixup;
     PRPackedBool               mIsOffScreenBrowser;
+    PRPackedBool               mIsActive;
+    PRPackedBool               mIsAppTab;
+    PRPackedBool               mUseGlobalHistory;
 
     // This boolean is set to true right before we fire pagehide and generally
     // unset when we embed a new content viewer.  While it's true no navigation
@@ -773,9 +820,13 @@ protected:
     // presentation of the page, and to SetupNewViewer() that the old viewer
     // should be passed a SHEntry to save itself into.
     PRPackedBool               mSavingOldViewer;
+
+    // @see nsIDocShellHistory::createdDynamically
+    PRPackedBool               mDynamicallyCreated;
 #ifdef DEBUG
     PRPackedBool               mInEnsureScriptEnv;
 #endif
+    PRUint64                   mHistoryID;
 
     static nsIURIFixup *sURIFixup;
 
