@@ -43,10 +43,10 @@
 #include "nsIContentPolicy.h"
 #include "nsIHttpChannel.h"
 #include "nsIIOService.h"
+#include "nsIProtocolHandler.h"
 #include "nsIRequest.h"
 #include "nsIScriptSecurityManager.h"
 #include "nsIStreamLoader.h"
-#include "nsIChannelClassifier.h"
 
 // Other includes
 #include "nsAutoLock.h"
@@ -60,6 +60,9 @@
 #include "nsThreadUtils.h"
 #include "pratom.h"
 #include "nsDocShellCID.h"
+#include "nsIChannelPolicy.h"
+#include "nsChannelPolicy.h"
+#include "nsIContentSecurityPolicy.h"
 
 // DOMWorker includes
 #include "nsDOMWorkerPool.h"
@@ -88,7 +91,7 @@ NS_IMPL_ISUPPORTS_INHERITED2(nsDOMWorkerScriptLoader, nsDOMWorkerFeature,
 nsresult
 nsDOMWorkerScriptLoader::LoadScripts(JSContext* aCx,
                                      const nsTArray<nsString>& aURLs,
-                                     PRBool aForWorker)
+                                     PRBool aExecute)
 {
   NS_ASSERTION(!NS_IsMainThread(), "Wrong thread!");
   NS_ASSERTION(aCx, "Null context!");
@@ -100,8 +103,6 @@ nsDOMWorkerScriptLoader::LoadScripts(JSContext* aCx,
     return NS_ERROR_ABORT;
   }
 
-  mForWorker = aForWorker;
-
   mScriptCount = aURLs.Length();
   if (!mScriptCount) {
     return NS_ERROR_INVALID_ARG;
@@ -145,23 +146,26 @@ nsDOMWorkerScriptLoader::LoadScripts(JSContext* aCx,
     return rv;
   }
 
-  rv = ExecuteScripts(aCx);
-  if (NS_FAILED(rv)) {
-    return rv;
+  if (aExecute) {
+    rv = ExecuteScripts(aCx);
+    if (NS_FAILED(rv)) {
+      return rv;
+    }
   }
 
   return NS_OK;
 }
 
 nsresult
-nsDOMWorkerScriptLoader::LoadScript(JSContext* aCx,
-                                    const nsString& aURL,
-                                    PRBool aForWorker)
+nsDOMWorkerScriptLoader::LoadWorkerScript(JSContext* aCx,
+                                          const nsString& aURL)
 {
+  mForWorker = PR_TRUE;
+
   nsAutoTArray<nsString, 1> url;
   url.AppendElement(aURL);
 
-  return LoadScripts(aCx, url, aForWorker);
+  return LoadScripts(aCx, url, PR_FALSE);
 }
 
 nsresult
@@ -176,20 +180,9 @@ nsDOMWorkerScriptLoader::DoRunLoop(JSContext* aCx)
   nsresult rv = NS_DispatchToMainThread(this);
   NS_ENSURE_SUCCESS(rv, rv);
 
-  if (!(done || mCanceled)) {
-    // Since we're going to lock up this thread we might as well allow the
-    // thread service to schedule another worker on a new thread.
-    nsDOMThreadService* threadService = nsDOMThreadService::get();
-    PRBool changed = NS_SUCCEEDED(threadService->ChangeThreadPoolMaxThreads(1));
-
-    while (!(done || mCanceled)) {
-      JSAutoSuspendRequest asr(aCx);
-      NS_ProcessNextEvent(mTarget);
-    }
-
-    if (changed) {
-      threadService->ChangeThreadPoolMaxThreads(-1);
-    }
+  while (!(done || mCanceled)) {
+    JSAutoSuspendRequest asr(aCx);
+    NS_ProcessNextEvent(mTarget);
   }
 
   return mCanceled ? NS_ERROR_ABORT : NS_OK;
@@ -417,46 +410,57 @@ nsDOMWorkerScriptLoader::RunInternal()
 {
   NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
 
-  // Things we need to make all this work...
-  nsCOMPtr<nsIDocument> parentDoc = mWorker->Pool()->ParentDocument();
-  if (!parentDoc) {
-    // Must have been canceled.
-    return NS_ERROR_ABORT;
-  }
-
-  nsIPrincipal* principal;
-  nsIURI* baseURI;
-
   if (mForWorker) {
     NS_ASSERTION(mScriptCount == 1, "Bad state!");
+  }
 
-    nsRefPtr<nsDOMWorker> parentWorker = mWorker->GetParent();
+  nsRefPtr<nsDOMWorker> parentWorker = mWorker->GetParent();
+
+  // Figure out which principal to use.
+  nsIPrincipal* principal = mWorker->GetPrincipal();
+  if (!principal) {
+    if (!parentWorker) {
+      NS_ERROR("Must have a principal if this is not a subworker!");
+    }
+    principal = parentWorker->GetPrincipal();
+  }
+  NS_ASSERTION(principal, "This should never be null here!");
+
+  // Figure out our base URI.
+  nsCOMPtr<nsIURI> baseURI;
+  if (mForWorker) {
     if (parentWorker) {
-      principal = parentWorker->GetPrincipal();
-      NS_ENSURE_STATE(principal);
-
-      baseURI = parentWorker->GetURI();
-      NS_ENSURE_STATE(baseURI);
+      baseURI = parentWorker->GetBaseURI();
+      NS_ASSERTION(baseURI, "Should have been set already!");
     }
     else {
-      principal = parentDoc->NodePrincipal();
-      NS_ENSURE_STATE(principal);
+      // May be null.
+      baseURI = mWorker->GetBaseURI();
 
-      baseURI = parentDoc->GetBaseURI();
+      // Don't leave a temporary URI hanging around.
+      mWorker->ClearBaseURI();
     }
+    NS_ASSERTION(!mWorker->GetBaseURI(), "Should not be set here!");
   }
   else {
-    principal = mWorker->GetPrincipal();
-    baseURI = mWorker->GetURI();
-
-    NS_ASSERTION(principal && baseURI, "Should have been set already!");
+    baseURI = mWorker->GetBaseURI();
+    NS_ASSERTION(baseURI, "Should have been set already!");
   }
 
+  nsCOMPtr<nsIDocument> parentDoc = mWorker->Pool()->ParentDocument();
+
   // All of these can potentially be null, but that should be ok. We'll either
   // succeed without them or fail below.
-  nsCOMPtr<nsILoadGroup> loadGroup(parentDoc->GetDocumentLoadGroup());
+  nsCOMPtr<nsILoadGroup> loadGroup;
+  if (parentDoc) {
+    loadGroup = parentDoc->GetDocumentLoadGroup();
+  }
+
   nsCOMPtr<nsIIOService> ios(do_GetIOService());
 
+  nsIScriptSecurityManager* secMan = nsContentUtils::GetSecurityManager();
+  NS_ASSERTION(secMan, "This should never be null!");
+
   for (PRUint32 index = 0; index < mScriptCount; index++) {
     ScriptLoadInfo& loadInfo = mLoadInfos[index];
     nsresult& rv = loadInfo.result;
@@ -469,37 +473,30 @@ nsDOMWorkerScriptLoader::RunInternal()
       return rv;
     }
 
-    nsIScriptSecurityManager* secMan = nsContentUtils::GetSecurityManager();
-    NS_ENSURE_TRUE(secMan, NS_ERROR_FAILURE);
-
-    PRInt16 shouldLoad = nsIContentPolicy::ACCEPT;
-    rv = NS_CheckContentLoadPolicy(nsIContentPolicy::TYPE_SCRIPT, uri,
-                                   principal, parentDoc,
-                                   NS_LITERAL_CSTRING("text/javascript"),
-                                   nsnull, &shouldLoad,
-                                   nsContentUtils::GetContentPolicy(), secMan);
-    if (NS_FAILED(rv) || NS_CP_REJECTED(shouldLoad)) {
-      if (NS_FAILED(rv) || shouldLoad != nsIContentPolicy::REJECT_TYPE) {
-        return NS_ERROR_CONTENT_BLOCKED;
+    // If we're part of a document then check the content load policy.
+    if (parentDoc) {
+      PRInt16 shouldLoad = nsIContentPolicy::ACCEPT;
+      rv = NS_CheckContentLoadPolicy(nsIContentPolicy::TYPE_SCRIPT, uri,
+                                     principal, parentDoc,
+                                     NS_LITERAL_CSTRING("text/javascript"),
+                                     nsnull, &shouldLoad,
+                                     nsContentUtils::GetContentPolicy(),
+                                     secMan);
+      if (NS_FAILED(rv) || NS_CP_REJECTED(shouldLoad)) {
+        if (NS_FAILED(rv) || shouldLoad != nsIContentPolicy::REJECT_TYPE) {
+          return (rv = NS_ERROR_CONTENT_BLOCKED);
+        }
+        return (rv = NS_ERROR_CONTENT_BLOCKED_SHOW_ALT);
       }
-      return NS_ERROR_CONTENT_BLOCKED_SHOW_ALT;
     }
 
     // If this script loader is being used to make a new worker then we need to
     // do a same-origin check. Otherwise we need to clear the load with the
     // security manager.
-    if (mForWorker) {
-      rv = principal->CheckMayLoad(uri, PR_FALSE);
-      NS_ENSURE_SUCCESS(rv, rv);
-
-      // Set the principal and URI on the new worker.
-      mWorker->SetPrincipal(principal);
-      mWorker->SetURI(uri);
-    }
-    else {
-      rv = secMan->CheckLoadURIWithPrincipal(principal, uri, 0);
-      NS_ENSURE_SUCCESS(rv, rv);
-    }
+    rv = mForWorker ?
+         principal->CheckMayLoad(uri, PR_FALSE):
+         secMan->CheckLoadURIWithPrincipal(principal, uri, 0);
+    NS_ENSURE_SUCCESS(rv, rv);
 
     // We need to know which index we're on in OnStreamComplete so we know where
     // to put the result.
@@ -516,7 +513,27 @@ nsDOMWorkerScriptLoader::RunInternal()
     rv = NS_NewStreamLoader(getter_AddRefs(loader), this);
     NS_ENSURE_SUCCESS(rv, rv);
 
-    rv = NS_NewChannel(getter_AddRefs(loadInfo.channel), uri, ios, loadGroup);
+    // Get Content Security Policy from parent document to pass into channel
+    nsCOMPtr<nsIContentSecurityPolicy> csp;
+    rv = principal->GetCsp(getter_AddRefs(csp));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    nsCOMPtr<nsIChannelPolicy> channelPolicy;
+    if (csp) {
+      channelPolicy = do_CreateInstance("@mozilla.org/nschannelpolicy;1", &rv);
+      NS_ENSURE_SUCCESS(rv, rv);
+
+      rv = channelPolicy->SetContentSecurityPolicy(csp);
+      NS_ENSURE_SUCCESS(rv, rv);
+
+      rv = channelPolicy->SetLoadType(nsIContentPolicy::TYPE_SCRIPT);
+      NS_ENSURE_SUCCESS(rv, rv);
+    }
+
+    rv = NS_NewChannel(getter_AddRefs(loadInfo.channel),
+                       uri, ios, loadGroup, nsnull,
+                       nsIRequest::LOAD_NORMAL | nsIChannel::LOAD_CLASSIFY_URI,
+                       channelPolicy);
     NS_ENSURE_SUCCESS(rv, rv);
 
     rv = loadInfo.channel->AsyncOpen(loader, indexSupports);
@@ -525,18 +542,6 @@ nsDOMWorkerScriptLoader::RunInternal()
       loadInfo.channel = nsnull;
       return rv;
     }
-
-    // Check the load against the URI classifier
-    nsCOMPtr<nsIChannelClassifier> classifier =
-        do_CreateInstance(NS_CHANNELCLASSIFIER_CONTRACTID);
-    if (classifier) {
-        rv = classifier->Start(loadInfo.channel, PR_TRUE);
-        if (NS_FAILED(rv)) {
-            loadInfo.channel->Cancel(rv);
-            loadInfo.channel = nsnull;
-            return rv;
-        }
-    }
   }
 
   return NS_OK;
@@ -567,34 +572,11 @@ nsDOMWorkerScriptLoader::OnStreamCompleteInternal(nsIStreamLoader* aLoader,
   NS_ASSERTION(!loadInfo.done, "Got complete on the same load twice!");
   loadInfo.done = PR_TRUE;
 
-#ifdef DEBUG
-  // Make sure we're seeing the channel that we expect.
-  nsCOMPtr<nsIRequest> requestDebug;
-  nsresult rvDebug = aLoader->GetRequest(getter_AddRefs(requestDebug));
-
-  // When we cancel sometimes we get null here. That should be ok, but only if
-  // we're canceled.
-  NS_ASSERTION(NS_SUCCEEDED(rvDebug) || mCanceled, "GetRequest failed!");
-
-  if (NS_SUCCEEDED(rvDebug)) {
-    nsCOMPtr<nsIChannel> channel(do_QueryInterface(requestDebug));
-    NS_ASSERTION(channel, "QI failed!");
-
-    nsCOMPtr<nsISupports> thisChannel(do_QueryInterface(channel));
-    NS_ASSERTION(thisChannel, "QI failed!");
-
-    nsCOMPtr<nsISupports> ourChannel(do_QueryInterface(loadInfo.channel));
-    NS_ASSERTION(ourChannel, "QI failed!");
-
-    NS_ASSERTION(thisChannel == ourChannel, "Wrong channel!");
-  }
-#endif
-
   // Use an alias to keep rv and loadInfo.result in sync.
-  nsresult& rv = loadInfo.result = aStatus;
+  nsresult& rv = loadInfo.result;
 
-  if (NS_FAILED(rv)) {
-    return rv;
+  if (NS_FAILED(aStatus)) {
+    return rv = aStatus;
   }
 
   if (!(aStringLen && aString)) {
@@ -607,7 +589,7 @@ nsDOMWorkerScriptLoader::OnStreamCompleteInternal(nsIStreamLoader* aLoader,
   rv = aLoader->GetRequest(getter_AddRefs(request));
   NS_ENSURE_SUCCESS(rv, rv);
 
-  nsCOMPtr<nsIHttpChannel> httpChannel(do_QueryInterface(request));
+  nsCOMPtr<nsIHttpChannel> httpChannel = do_QueryInterface(request);
   if (httpChannel) {
     PRBool requestSucceeded;
     rv = httpChannel->GetRequestSucceeded(&requestSucceeded);
@@ -618,8 +600,8 @@ nsDOMWorkerScriptLoader::OnStreamCompleteInternal(nsIStreamLoader* aLoader,
     }
   }
 
+  // May be null.
   nsIDocument* parentDoc = mWorker->Pool()->ParentDocument();
-  NS_ASSERTION(parentDoc, "Null parent document?!");
 
   // Use the regular nsScriptLoader for this grunt work! Should be just fine
   // because we're running on the main thread.
@@ -635,7 +617,8 @@ nsDOMWorkerScriptLoader::OnStreamCompleteInternal(nsIStreamLoader* aLoader,
   }
 
   nsCString filename;
-  loadInfo.finalURI->GetSpec(filename);
+  rv = loadInfo.finalURI->GetSpec(filename);
+  NS_ENSURE_SUCCESS(rv, rv);
 
   if (filename.IsEmpty()) {
     filename.Assign(NS_LossyConvertUTF16toASCII(loadInfo.url));
@@ -646,6 +629,67 @@ nsDOMWorkerScriptLoader::OnStreamCompleteInternal(nsIStreamLoader* aLoader,
     loadInfo.url.Assign(NS_ConvertUTF8toUTF16(filename));
   }
 
+  // Update the principal of the worker and its base URI if we just loaded the
+  // worker's primary script.
+  if (mForWorker) {
+    nsCOMPtr<nsIChannel> channel = do_QueryInterface(request);
+    NS_ASSERTION(channel, "This should never fail!");
+
+    // Take care of the base URI first.
+    nsCOMPtr<nsIURI> finalURI;
+    rv = NS_GetFinalChannelURI(channel, getter_AddRefs(finalURI));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    mWorker->SetBaseURI(finalURI);
+
+    // Now to figure out which principal to give this worker.
+    nsRefPtr<nsDOMWorker> parent = mWorker->GetParent();
+    NS_ASSERTION(mWorker->GetPrincipal() || parent, "Must have one of these!");
+
+    nsCOMPtr<nsIPrincipal> loadPrincipal = mWorker->GetPrincipal() ?
+                                           mWorker->GetPrincipal() :
+                                           parent->GetPrincipal();
+
+    nsIScriptSecurityManager* ssm = nsContentUtils::GetSecurityManager();
+    NS_ASSERTION(ssm, "Should never be null!");
+
+    nsCOMPtr<nsIPrincipal> channelPrincipal;
+    rv = ssm->GetChannelPrincipal(channel, getter_AddRefs(channelPrincipal));
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    // See if this is a resource URI. Since JSMs usually come from resource://
+    // URIs we're currently considering all URIs with the URI_IS_UI_RESOURCE
+    // flag as valid for creating privileged workers.
+    if (!nsContentUtils::IsSystemPrincipal(channelPrincipal)) {
+      PRBool isResource;
+      rv = NS_URIChainHasFlags(finalURI,
+                               nsIProtocolHandler::URI_IS_UI_RESOURCE,
+                               &isResource);
+      NS_ENSURE_SUCCESS(rv, rv);
+
+      if (isResource) {
+        rv = ssm->GetSystemPrincipal(getter_AddRefs(channelPrincipal));
+        NS_ENSURE_SUCCESS(rv, rv);
+      }
+    }
+
+    // If the load principal is the system principal then the channel principal
+    // must also be the system principal (we do not allow chrome code to create
+    // workers with non-chrome scripts). Otherwise this channel principal must
+    // be same origin with the load principal (we check again here in case
+    // redirects changed the location of the script).
+    if (nsContentUtils::IsSystemPrincipal(loadPrincipal)) {
+      if (!nsContentUtils::IsSystemPrincipal(channelPrincipal)) {
+        return rv = NS_ERROR_DOM_BAD_URI;
+      }
+    }
+    else if (NS_FAILED(loadPrincipal->CheckMayLoad(finalURI, PR_FALSE))) {
+      return rv = NS_ERROR_DOM_BAD_URI;
+    }
+
+    mWorker->SetPrincipal(channelPrincipal);
+  }
+
   nsRefPtr<ScriptCompiler> compiler =
     new ScriptCompiler(this, loadInfo.scriptText, filename, loadInfo.scriptObj);
   NS_ASSERTION(compiler, "Out of memory!");
