@@ -246,7 +246,8 @@ nsDOMWorkerXHRUpload::nsDOMWorkerXHRUpload(nsDOMWorkerXHR* aWorkerXHR)
 NS_IMPL_ISUPPORTS_INHERITED1(nsDOMWorkerXHRUpload, nsDOMWorkerXHREventTarget,
                                                    nsIXMLHttpRequestUpload)
 
-NS_IMPL_CI_INTERFACE_GETTER3(nsDOMWorkerXHRUpload, nsIDOMEventTarget,
+NS_IMPL_CI_INTERFACE_GETTER4(nsDOMWorkerXHRUpload, nsIDOMNSEventTarget,
+                                                   nsIDOMEventTarget,
                                                    nsIXMLHttpRequestEventTarget,
                                                    nsIXMLHttpRequestUpload)
 
@@ -257,25 +258,7 @@ nsDOMWorkerXHRUpload::AddEventListener(const nsAString& aType,
                                        nsIDOMEventListener* aListener,
                                        PRBool aUseCapture)
 {
-  NS_ASSERTION(!NS_IsMainThread(), "Wrong thread!");
-  NS_ENSURE_ARG_POINTER(aListener);
-
-  if (mWorkerXHR->mWorker->IsCanceled()) {
-    return NS_ERROR_ABORT;
-  }
-
-  nsresult rv = nsDOMWorkerXHREventTarget::AddEventListener(aType, aListener,
-                                                            aUseCapture);
-  NS_ENSURE_SUCCESS(rv, rv);
-
-  rv = mWorkerXHR->mXHRProxy->UploadEventListenerAdded();
-  if (NS_FAILED(rv)) {
-    NS_WARNING("UploadEventListenerAdded failed!");
-    RemoveEventListener(aType, aListener, aUseCapture);
-    return rv;
-  }
-
-  return NS_OK;
+  return AddEventListener(aType, aListener, aUseCapture, PR_FALSE, 0);
 }
 
 NS_IMETHODIMP
@@ -308,6 +291,36 @@ nsDOMWorkerXHRUpload::DispatchEvent(nsIDOMEvent* aEvent,
   return nsDOMWorkerXHREventTarget::DispatchEvent(aEvent, _retval);
 }
 
+NS_IMETHODIMP
+nsDOMWorkerXHRUpload::AddEventListener(const nsAString& aType,
+                                       nsIDOMEventListener* aListener,
+                                       PRBool aUseCapture,
+                                       PRBool aWantsUntrusted,
+                                       PRUint8 optional_argc)
+{
+  NS_ASSERTION(!NS_IsMainThread(), "Wrong thread!");
+  NS_ENSURE_ARG_POINTER(aListener);
+
+  if (mWorkerXHR->mWorker->IsCanceled()) {
+    return NS_ERROR_ABORT;
+  }
+
+  nsresult rv = nsDOMWorkerXHREventTarget::AddEventListener(aType, aListener,
+                                                            aUseCapture,
+                                                            aWantsUntrusted,
+                                                            optional_argc);
+  NS_ENSURE_SUCCESS(rv, rv);
+
+  rv = mWorkerXHR->mXHRProxy->UploadEventListenerAdded();
+  if (NS_FAILED(rv)) {
+    NS_WARNING("UploadEventListenerAdded failed!");
+    RemoveEventListener(aType, aListener, aUseCapture);
+    return rv;
+  }
+
+  return NS_OK;
+}
+
 nsresult
 nsDOMWorkerXHRUpload::SetOnXListener(const nsAString& aType,
                                      nsIDOMEventListener* aListener)
@@ -341,7 +354,7 @@ nsDOMWorkerXHR::~nsDOMWorkerXHR()
   if (mXHRProxy) {
     if (!NS_IsMainThread()) {
       nsCOMPtr<nsIRunnable> runnable =
-        NS_NEW_RUNNABLE_METHOD(nsDOMWorkerXHRProxy, mXHRProxy.get(), Destroy);
+        NS_NewRunnableMethod(mXHRProxy, &nsDOMWorkerXHRProxy::Destroy);
 
       if (runnable) {
         mXHRProxy = nsnull;
@@ -364,7 +377,8 @@ NS_IMPL_QUERY_INTERFACE_INHERITED2(nsDOMWorkerXHR, nsDOMWorkerXHREventTarget,
                                                    nsIXMLHttpRequest,
                                                    nsIXPCScriptable)
 
-NS_IMPL_CI_INTERFACE_GETTER3(nsDOMWorkerXHR, nsIDOMEventTarget,
+NS_IMPL_CI_INTERFACE_GETTER4(nsDOMWorkerXHR, nsIDOMNSEventTarget,
+                                             nsIDOMEventTarget,
                                              nsIXMLHttpRequestEventTarget,
                                              nsIXMLHttpRequest)
 
@@ -618,8 +632,9 @@ nsDOMWorkerXHR::OpenRequest(const nsACString& aMethod,
 }
 
 NS_IMETHODIMP
-nsDOMWorkerXHR::Open(const nsACString& aMethod,
-                     const nsACString& aUrl)
+nsDOMWorkerXHR::Open(const nsACString& aMethod, const nsACString& aUrl,
+                     PRBool aAsync, const nsAString& aUser,
+                     const nsAString& aPassword, PRUint8 optional_argc)
 {
   NS_ASSERTION(!NS_IsMainThread(), "Wrong thread!");
 
@@ -627,67 +642,11 @@ nsDOMWorkerXHR::Open(const nsACString& aMethod,
     return NS_ERROR_ABORT;
   }
 
-  PRBool async = PR_TRUE;
-  nsAutoString user, password;
+  if (!optional_argc) {
+      aAsync = PR_TRUE;
+  }
 
-  nsIXPConnect* xpc = nsContentUtils::XPConnect();
-  NS_ENSURE_TRUE(xpc, NS_ERROR_UNEXPECTED);
-
-  nsAXPCNativeCallContext* cc;
-  nsresult rv = xpc->GetCurrentNativeCallContext(&cc);
-
-  do {
-    if (NS_FAILED(rv) || !cc) {
-      break;
-    }
-
-    PRUint32 argc;
-    rv = cc->GetArgc(&argc);
-    NS_ENSURE_SUCCESS(rv, rv);
-
-    if (argc < 3) {
-      break;
-    }
-
-    jsval* argv;
-    rv = cc->GetArgvPtr(&argv);
-    NS_ENSURE_SUCCESS(rv, rv);
-
-    JSContext* cx;
-    rv = cc->GetJSContext(&cx);
-    NS_ENSURE_SUCCESS(rv, rv);
-
-    JSAutoRequest ar(cx);
-
-    JSBool asyncBool;
-    JS_ValueToBoolean(cx, argv[2], &asyncBool);
-    async = (PRBool)asyncBool;
-
-    if (argc < 4) {
-      break;
-    }
-
-    JSString* argStr;
-    if (!JSVAL_IS_NULL(argv[3]) && !JSVAL_IS_VOID(argv[3])) {
-      argStr = JS_ValueToString(cx, argv[3]);
-      if (argStr) {
-        user.Assign(nsDependentJSString(argStr));
-      }
-    }
-
-    if (argc < 5) {
-      break;
-    }
-
-    if (!JSVAL_IS_NULL(argv[4]) && !JSVAL_IS_VOID(argv[4])) {
-      argStr = JS_ValueToString(cx, argv[4]);
-      if (argStr) {
-        password.Assign(nsDependentJSString(argStr));
-      }
-    }
-  } while (PR_FALSE);
-
-  return OpenRequest(aMethod, aUrl, async, user, password);
+  return OpenRequest(aMethod, aUrl, aAsync, aUser, aPassword);
 }
 
 NS_IMETHODIMP
@@ -930,3 +889,10 @@ nsDOMWorkerXHR::SetWithCredentials(PRBool aWithCredentials)
 
   return NS_OK;
 }
+
+/* readonly attribute jsval (ArrayBuffer) mozResponseArrayBuffer; */
+NS_IMETHODIMP
+nsDOMWorkerXHR::GetMozResponseArrayBuffer(jsval *aResult)
+{
+    return NS_ERROR_NOT_IMPLEMENTED;
+}
