@@ -81,7 +81,6 @@
 #include "nsIDOMNSDocument.h"
 #include "nsIScriptContext.h"
 #include "imgIContainer.h"
-#include "nsPresContext.h"
 
 #if DEBUG
 //#define NOISY_DOC_LOADING  1
@@ -146,12 +145,12 @@ nsEditingSession::MakeWindowEditable(nsIDOMWindow *aWindow,
 {
   mEditorType.Truncate();
   mEditorFlags = 0;
-  mWindowToBeEdited = do_GetWeakReference(aWindow);
 
   // disable plugins
   nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
 
+  mDocShell = do_GetWeakReference(docShell);
   mInteractive = aInteractive;
   mMakeWholeDocumentEditable = aMakeWholeDocumentEditable;
 
@@ -175,15 +174,15 @@ nsEditingSession::MakeWindowEditable(nsIDOMWindow *aWindow,
   // if all this does is setup listeners and I don't need listeners, 
   // can't this step be ignored?? (based on aDoAfterURILoad)
   rv = PrepareForEditing(aWindow);
-  if (NS_FAILED(rv)) return rv;  
+  NS_ENSURE_SUCCESS(rv, rv);  
   
   nsCOMPtr<nsIEditorDocShell> editorDocShell;
   rv = GetEditorDocShellFromWindow(aWindow, getter_AddRefs(editorDocShell));
-  if (NS_FAILED(rv)) return rv;  
+  NS_ENSURE_SUCCESS(rv, rv);  
   
   // set the flag on the docShell to say that it's editable
   rv = editorDocShell->MakeEditable(aDoAfterUriLoad);
-  if (NS_FAILED(rv)) return rv;  
+  NS_ENSURE_SUCCESS(rv, rv);  
 
   // Setup commands common to plaintext and html editors,
   //  including the document creation observers
@@ -192,7 +191,7 @@ nsEditingSession::MakeWindowEditable(nsIDOMWindow *aWindow,
                                     aWindow,
                                     static_cast<nsIEditingSession*>(this),
                                     &mBaseCommandControllerId);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // The second is a controller to monitor doc state,
   // such as creation and "dirty flag"
@@ -200,7 +199,7 @@ nsEditingSession::MakeWindowEditable(nsIDOMWindow *aWindow,
                                     aWindow,
                                     static_cast<nsIEditingSession*>(this),
                                     &mDocStateControllerId);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // aDoAfterUriLoad can be false only when making an existing window editable
   if (!aDoAfterUriLoad)
@@ -220,7 +219,7 @@ NS_IMETHODIMP
 nsEditingSession::DisableJSAndPlugins(nsIDOMWindow *aWindow)
 {
   nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
 
   PRBool tmp;
   nsresult rv = docShell->GetAllowJavascript(&tmp);
@@ -248,13 +247,12 @@ nsEditingSession::DisableJSAndPlugins(nsIDOMWindow *aWindow)
 NS_IMETHODIMP
 nsEditingSession::RestoreJSAndPlugins(nsIDOMWindow *aWindow)
 {
-  if (!mDisabledJSAndPlugins)
-    return NS_OK;
+  NS_ENSURE_TRUE(mDisabledJSAndPlugins, NS_OK);
 
   mDisabledJSAndPlugins = PR_FALSE;
 
   nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
 
   nsresult rv = docShell->SetAllowJavascript(mScriptsEnabled);
   NS_ENSURE_SUCCESS(rv, rv);
@@ -263,6 +261,14 @@ nsEditingSession::RestoreJSAndPlugins(nsIDOMWindow *aWindow)
   return docShell->SetAllowPlugins(mPluginsEnabled);
 }
 
+NS_IMETHODIMP
+nsEditingSession::GetJsAndPluginsDisabled(PRBool *aResult)
+{
+  NS_ENSURE_ARG_POINTER(aResult);
+  *aResult = mDisabledJSAndPlugins;
+  return NS_OK;
+}
+
 /*---------------------------------------------------------------------------
 
   WindowIsEditable
@@ -275,7 +281,7 @@ nsEditingSession::WindowIsEditable(nsIDOMWindow *aWindow, PRBool *outIsEditable)
   nsCOMPtr<nsIEditorDocShell> editorDocShell;
   nsresult rv = GetEditorDocShellFromWindow(aWindow,
                                             getter_AddRefs(editorDocShell));
-  if (NS_FAILED(rv)) return rv;  
+  NS_ENSURE_SUCCESS(rv, rv);  
 
   return editorDocShell->GetEditable(outIsEditable);
 }
@@ -304,8 +310,7 @@ const char* const gSupportedTextTypes[] = {
 PRBool
 IsSupportedTextType(const char* aMIMEType)
 {
-  if (!aMIMEType)
-    return PR_FALSE;
+  NS_ENSURE_TRUE(aMIMEType, PR_FALSE);
 
   PRInt32 i = 0;
   while (gSupportedTextTypes[i])
@@ -419,7 +424,7 @@ nsEditingSession::SetupEditorOnWindow(nsIDOMWindow *aWindow)
   // This allows notification of error state
   //  even if we don't create an editor
   rv = mStateMaintainer->Init(aWindow);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   if (mEditorStatus != eEditorCreationInProgress)
   {
@@ -430,27 +435,34 @@ nsEditingSession::SetupEditorOnWindow(nsIDOMWindow *aWindow)
   // Create editor and do other things 
   //  only if we haven't found some error above,
   nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
-  if (!docShell) return NS_ERROR_FAILURE;  
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);  
 
   if (!mInteractive) {
     // Disable animation of images in this document:
     nsCOMPtr<nsIDOMWindowUtils> utils(do_GetInterface(aWindow));
-    if (!utils) return NS_ERROR_FAILURE;
+    NS_ENSURE_TRUE(utils, NS_ERROR_FAILURE);
 
     rv = utils->GetImageAnimationMode(&mImageAnimationMode);
-    if (NS_FAILED(rv)) return rv;
+    NS_ENSURE_SUCCESS(rv, rv);
     utils->SetImageAnimationMode(imgIContainer::kDontAnimMode);
   }
 
   // create and set editor
   nsCOMPtr<nsIEditorDocShell> editorDocShell = do_QueryInterface(docShell, &rv);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
-  nsCOMPtr<nsIEditor> editor = do_CreateInstance(classString, &rv);
-  if (NS_FAILED(rv)) return rv;
+  // Try to reuse an existing editor
+  nsCOMPtr<nsIEditor> editor = do_QueryReferent(mExistingEditor);
+  if (editor) {
+    editor->PreDestroy(PR_FALSE);
+  } else {
+    editor = do_CreateInstance(classString, &rv);
+    NS_ENSURE_SUCCESS(rv, rv);
+    mExistingEditor = do_GetWeakReference(editor);
+  }
   // set the editor on the docShell. The docShell now owns it.
   rv = editorDocShell->SetEditor(editor);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // setup the HTML editor command controller
   if (needHTMLController)
@@ -459,46 +471,46 @@ nsEditingSession::SetupEditorOnWindow(nsIDOMWindow *aWindow)
     rv = SetupEditorCommandController("@mozilla.org/editor/htmleditorcontroller;1",
                                       aWindow, editor,
                                       &mHTMLCommandControllerId);
-    if (NS_FAILED(rv)) return rv;
+    NS_ENSURE_SUCCESS(rv, rv);
   }
 
   // Set mimetype on editor
   rv = editor->SetContentsMIMEType(mimeCType.get());
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsIContentViewer> contentViewer;
   rv = docShell->GetContentViewer(getter_AddRefs(contentViewer));
-  if (NS_FAILED(rv)) return rv;
-  if (!contentViewer) return NS_ERROR_FAILURE;
+  NS_ENSURE_SUCCESS(rv, rv);
+  NS_ENSURE_TRUE(contentViewer, NS_ERROR_FAILURE);
 
   nsCOMPtr<nsIDOMDocument> domDoc;  
   rv = contentViewer->GetDOMDocument(getter_AddRefs(domDoc));
-  if (NS_FAILED(rv)) return rv;
-  if (!domDoc) return NS_ERROR_FAILURE;
+  NS_ENSURE_SUCCESS(rv, rv);
+  NS_ENSURE_TRUE(domDoc, NS_ERROR_FAILURE);
 
   // Set up as a doc state listener
   // Important! We must have this to broadcast the "obs_documentCreated" message
   rv = editor->AddDocumentStateListener(mStateMaintainer);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // XXXbz we really shouldn't need a presShell here!
   nsCOMPtr<nsIPresShell> presShell;
   rv = docShell->GetPresShell(getter_AddRefs(presShell));
-  if (NS_FAILED(rv)) return rv;
-  if (!presShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_SUCCESS(rv, rv);
+  NS_ENSURE_TRUE(presShell, NS_ERROR_FAILURE);
 
   nsCOMPtr<nsISelectionController> selCon = do_QueryInterface(presShell);
   rv = editor->Init(domDoc, presShell, nsnull /* root content */,
                     selCon, mEditorFlags);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsISelection> selection;
   editor->GetSelection(getter_AddRefs(selection));
   nsCOMPtr<nsISelectionPrivate> selPriv = do_QueryInterface(selection);
-  if (!selPriv) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(selPriv, NS_ERROR_FAILURE);
 
   rv = selPriv->AddSelectionListener(mStateMaintainer);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // and as a transaction listener
   nsCOMPtr<nsITransactionManager> txnMgr;
@@ -508,7 +520,7 @@ nsEditingSession::SetupEditorOnWindow(nsIDOMWindow *aWindow)
 
   // Set context on all controllers to be the editor
   rv = SetEditorOnControllers(aWindow, editor);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // Everything went fine!
   mEditorStatus = eEditorOK;
@@ -553,11 +565,9 @@ nsEditingSession::RemoveListenersAndControllers(nsIDOMWindow *aWindow,
 NS_IMETHODIMP
 nsEditingSession::TearDownEditorOnWindow(nsIDOMWindow *aWindow)
 {
-  if (!mDoneSetup)
-    return NS_OK;
+  NS_ENSURE_TRUE(mDoneSetup, NS_OK);
 
-  if (!aWindow)
-    return NS_ERROR_NULL_POINTER;
+  NS_ENSURE_TRUE(aWindow, NS_ERROR_NULL_POINTER);
 
   nsresult rv;
   
@@ -633,7 +643,7 @@ nsEditingSession::GetEditorForWindow(nsIDOMWindow *aWindow,
   nsCOMPtr<nsIEditorDocShell> editorDocShell;
   nsresult rv = GetEditorDocShellFromWindow(aWindow,
                                             getter_AddRefs(editorDocShell));
-  if (NS_FAILED(rv)) return rv;  
+  NS_ENSURE_SUCCESS(rv, rv);  
   
   return editorDocShell->GetEditor(outEditor);
 }
@@ -760,7 +770,7 @@ nsEditingSession::OnStateChange(nsIWebProgress *aWebProgress,
     }
   }
   //
-  // A network or document Request as finished...
+  // A network or document Request has finished...
   //
   else if (aStateFlags & nsIWebProgressListener::STATE_STOP)
   {
@@ -838,26 +848,26 @@ nsEditingSession::OnLocationChange(nsIWebProgress *aWebProgress,
 {
   nsCOMPtr<nsIDOMWindow> domWindow;
   nsresult rv = aWebProgress->GetDOMWindow(getter_AddRefs(domWindow));
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsIDOMDocument> domDoc;
   rv = domWindow->GetDocument(getter_AddRefs(domDoc));
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsIDocument> doc = do_QueryInterface(domDoc);
-  if (!doc) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(doc, NS_ERROR_FAILURE);
 
   doc->SetDocumentURI(aURI);
 
   // Notify the location-changed observer that
   //  the document URL has changed
   nsIDocShell *docShell = GetDocShellFromWindow(domWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
 
   nsCOMPtr<nsICommandManager> commandManager = do_GetInterface(docShell);
   nsCOMPtr<nsPICommandUpdater> commandUpdater =
                                   do_QueryInterface(commandManager);
-  if (!commandUpdater) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(commandUpdater, NS_ERROR_FAILURE);
 
   return commandUpdater->CommandStatusChanged("obs_documentLocationChanged");
 }
@@ -906,12 +916,8 @@ nsEditingSession::OnSecurityChange(nsIWebProgress *aWebProgress,
 PRBool
 nsEditingSession::IsProgressForTargetDocument(nsIWebProgress *aWebProgress)
 {
-  nsCOMPtr<nsIDOMWindow> domWindow;
-  if (aWebProgress)
-    aWebProgress->GetDOMWindow(getter_AddRefs(domWindow));
-  nsCOMPtr<nsIDOMWindow> editedDOMWindow = do_QueryReferent(mWindowToBeEdited);
-
-  return (domWindow && (domWindow == editedDOMWindow));
+  nsCOMPtr<nsIWebProgress> editedWebProgress = do_QueryReferent(mDocShell);
+  return editedWebProgress == aWebProgress;
 }
 
 
@@ -954,7 +960,7 @@ nsEditingSession::StartDocumentLoad(nsIWebProgress *aWebProgress,
   if (domWindow)
   {
     nsIDocShell *docShell = GetDocShellFromWindow(domWindow);
-    if (!docShell) return NS_ERROR_FAILURE;
+    NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
     docShell->DetachEditorFromWindow();
   }
     
@@ -990,7 +996,7 @@ nsEditingSession::EndDocumentLoad(nsIWebProgress *aWebProgress,
 
   // We want to call the base class EndDocumentLoad,
   // but avoid some of the stuff
-  // that nsWebShell does (need to refactor).
+  // that nsDocShell does (need to refactor).
   
   // OK, time to make an editor on this document
   nsCOMPtr<nsIDOMWindow> domWindow;
@@ -1004,7 +1010,7 @@ nsEditingSession::EndDocumentLoad(nsIWebProgress *aWebProgress,
   }
 
   nsIDocShell *docShell = GetDocShellFromWindow(domWindow);
-  if (!docShell) return NS_ERROR_FAILURE;       // better error handling?
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);       // better error handling?
 
   // cancel refresh from meta tags
   // we need to make sure that all pages in editor (whether editable or not)
@@ -1027,15 +1033,14 @@ nsEditingSession::EndDocumentLoad(nsIWebProgress *aWebProgress,
     {
       // To keep pre Gecko 1.9 behavior, setup editor always when
       // mMakeWholeDocumentEditable.
-      PRBool needsSetup;
+      bool needsSetup = false;
       if (mMakeWholeDocumentEditable) {
-        needsSetup = PR_TRUE;
+        needsSetup = true;
       } else {
         // do we already have an editor here?
         nsCOMPtr<nsIEditor> editor;
         rv = editorDocShell->GetEditor(getter_AddRefs(editor));
-        if (NS_FAILED(rv))
-           return rv;
+        NS_ENSURE_SUCCESS(rv, rv);
 
         needsSetup = !editor;
       }
@@ -1055,12 +1060,12 @@ nsEditingSession::EndDocumentLoad(nsIWebProgress *aWebProgress,
           }
   
           mLoadBlankDocTimer = do_CreateInstance("@mozilla.org/timer;1", &rv);
-          if (NS_FAILED(rv)) return rv;
+          NS_ENSURE_SUCCESS(rv, rv);
 
           mEditorStatus = eEditorCreationInProgress;
           mLoadBlankDocTimer->InitWithFuncCallback(
                                           nsEditingSession::TimerCallback,
-                                          (void*)docShell,
+                                          static_cast<void*> (mDocShell.get()),
                                           10, nsITimer::TYPE_ONE_SHOT);
         }
       }
@@ -1073,7 +1078,7 @@ nsEditingSession::EndDocumentLoad(nsIWebProgress *aWebProgress,
 void
 nsEditingSession::TimerCallback(nsITimer* aTimer, void* aClosure)
 {
-  nsCOMPtr<nsIDocShell> docShell = (nsIDocShell*)aClosure;
+  nsCOMPtr<nsIDocShell> docShell = do_QueryReferent(static_cast<nsIWeakReference*> (aClosure));
   if (docShell)
   {
     nsCOMPtr<nsIWebNavigation> webNav(do_QueryInterface(docShell));
@@ -1134,7 +1139,7 @@ nsEditingSession::EndPageLoad(nsIWebProgress *aWebProgress,
   aWebProgress->GetDOMWindow(getter_AddRefs(domWindow));
 
   nsIDocShell *docShell = GetDocShellFromWindow(domWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
 
   // cancel refresh from meta tags
   // we need to make sure that all pages in editor (whether editable or not)
@@ -1166,8 +1171,7 @@ nsIDocShell *
 nsEditingSession::GetDocShellFromWindow(nsIDOMWindow *aWindow)
 {
   nsCOMPtr<nsPIDOMWindow> window = do_QueryInterface(aWindow);
-  if (!window)
-    return nsnull;
+  NS_ENSURE_TRUE(window, nsnull);
 
   return window->GetDocShell();
 }
@@ -1184,7 +1188,7 @@ nsEditingSession::GetEditorDocShellFromWindow(nsIDOMWindow *aWindow,
                                               nsIEditorDocShell** outDocShell)
 {
   nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
-  if (!docShell) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
   
   return docShell->QueryInterface(NS_GET_IID(nsIEditorDocShell), 
                                   (void **)outDocShell);
@@ -1206,7 +1210,7 @@ nsEditingSession::PrepareForEditing(nsIDOMWindow *aWindow)
   
   // register callback
   nsCOMPtr<nsIWebProgress> webProgress = do_GetInterface(docShell);
-  if (!webProgress) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(webProgress, NS_ERROR_FAILURE);
 
   nsresult rv =
     webProgress->AddProgressListener(this,
@@ -1241,11 +1245,11 @@ nsEditingSession::SetupEditorCommandController(
   nsresult rv;
   nsCOMPtr<nsIDOMWindowInternal> domWindowInt =
                                     do_QueryInterface(aWindow, &rv);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
   
   nsCOMPtr<nsIControllers> controllers;      
   rv = domWindowInt->GetControllers(getter_AddRefs(controllers));
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   // We only have to create each singleton controller once
   // We know this has happened once we have a controllerId value
@@ -1254,17 +1258,17 @@ nsEditingSession::SetupEditorCommandController(
     nsresult rv;
     nsCOMPtr<nsIController> controller;
     controller = do_CreateInstance(aControllerClassName, &rv);
-    if (NS_FAILED(rv)) return rv;  
+    NS_ENSURE_SUCCESS(rv, rv);  
 
     // We must insert at head of the list to be sure our
     //   controller is found before other implementations
     //   (e.g., not-implemented versions by browser)
     rv = controllers->InsertControllerAt(0, controller);
-    if (NS_FAILED(rv)) return rv;  
+    NS_ENSURE_SUCCESS(rv, rv);  
 
     // Remember the ID for the controller
     rv = controllers->GetControllerId(controller, aControllerId);
-    if (NS_FAILED(rv)) return rv;  
+    NS_ENSURE_SUCCESS(rv, rv);  
   }  
 
   // Set the context
@@ -1286,25 +1290,25 @@ nsEditingSession::SetEditorOnControllers(nsIDOMWindow *aWindow,
   // set the editor on the controller
   nsCOMPtr<nsIDOMWindowInternal> domWindowInt =
                                      do_QueryInterface(aWindow, &rv);
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
   
   nsCOMPtr<nsIControllers> controllers;      
   rv = domWindowInt->GetControllers(getter_AddRefs(controllers));
-  if (NS_FAILED(rv)) return rv;
+  NS_ENSURE_SUCCESS(rv, rv);
 
   nsCOMPtr<nsISupports> editorAsISupports = do_QueryInterface(aEditor);
   if (mBaseCommandControllerId)
   {
     rv = SetContextOnControllerById(controllers, editorAsISupports,
                                     mBaseCommandControllerId);
-    if (NS_FAILED(rv)) return rv;
+    NS_ENSURE_SUCCESS(rv, rv);
   }
 
   if (mDocStateControllerId)
   {
     rv = SetContextOnControllerById(controllers, editorAsISupports,
                                     mDocStateControllerId);
-    if (NS_FAILED(rv)) return rv;
+    NS_ENSURE_SUCCESS(rv, rv);
   }
 
   if (mHTMLCommandControllerId)
@@ -1328,7 +1332,7 @@ nsEditingSession::SetContextOnControllerById(nsIControllers* aControllers,
   // ok with nil controller
   nsCOMPtr<nsIControllerContext> editorController =
                                        do_QueryInterface(controller);
-  if (!editorController) return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(editorController, NS_ERROR_FAILURE);
 
   return editorController->SetCommandContext(aContext);
 }
@@ -1404,8 +1408,7 @@ nsEditingSession::RestoreAnimationMode(nsIDOMWindow *aWindow)
 nsresult
 nsEditingSession::DetachFromWindow(nsIDOMWindow* aWindow)
 {
-  if (!mDoneSetup)
-    return NS_OK;
+  NS_ENSURE_TRUE(mDoneSetup, NS_OK);
 
   NS_ASSERTION(mStateMaintainer, "mStateMaintainer should exist.");
 
@@ -1425,7 +1428,7 @@ nsEditingSession::DetachFromWindow(nsIDOMWindow* aWindow)
 
   // Kill our weak reference to our original window, in case
   // it changes on restore, or otherwise dies.
-  mWindowToBeEdited = nsnull;
+  mDocShell = nsnull;
 
   return NS_OK;
 }
@@ -1433,8 +1436,7 @@ nsEditingSession::DetachFromWindow(nsIDOMWindow* aWindow)
 nsresult
 nsEditingSession::ReattachToWindow(nsIDOMWindow* aWindow)
 {
-  if (!mDoneSetup)
-    return NS_OK;
+  NS_ENSURE_TRUE(mDoneSetup, NS_OK);
 
   NS_ASSERTION(mStateMaintainer, "mStateMaintainer should exist.");
 
@@ -1442,7 +1444,9 @@ nsEditingSession::ReattachToWindow(nsIDOMWindow* aWindow)
   // old editor ot the window.
   nsresult rv;
 
-  mWindowToBeEdited = do_GetWeakReference(aWindow);
+  nsIDocShell *docShell = GetDocShellFromWindow(aWindow);
+  NS_ENSURE_TRUE(docShell, NS_ERROR_FAILURE);
+  mDocShell = do_GetWeakReference(docShell);
 
   // Disable plugins.
   if (!mInteractive)
@@ -1477,14 +1481,13 @@ nsEditingSession::ReattachToWindow(nsIDOMWindow* aWindow)
   // Get editor
   nsCOMPtr<nsIEditor> editor;
   rv = GetEditorForWindow(aWindow, getter_AddRefs(editor));
-  if (!editor)
-    return NS_ERROR_FAILURE;
+  NS_ENSURE_TRUE(editor, NS_ERROR_FAILURE);
 
   if (!mInteractive)
   {
     // Disable animation of images in this document:
     nsCOMPtr<nsIDOMWindowUtils> utils(do_GetInterface(aWindow));
-    if (!utils) return NS_ERROR_FAILURE;
+    NS_ENSURE_TRUE(utils, NS_ERROR_FAILURE);
 
     rv = utils->GetImageAnimationMode(&mImageAnimationMode);
     NS_ENSURE_SUCCESS(rv, rv);
