@@ -661,13 +661,6 @@ gtk_moz_embed_set_comp_path(const char *aPath)
   EmbedPrivate::SetCompPath(aPath);
 }
 
-void
-gtk_moz_embed_set_app_components(const nsModuleComponentInfo* aComps,
-                                 int aNumComponents)
-{
-  EmbedPrivate::SetAppComponents(aComps, aNumComponents);
-}
-
 void
 gtk_moz_embed_set_profile_path(const char *aDir, const char *aName)
 {
