@@ -38,39 +38,28 @@
 
 #include "nsUCSupport.h"
 
-#include "nsIPrefBranch.h"
-#include "nsIPrefService.h"
-
 #include "japanese.map"
 
 #include "nsICharsetConverterManager.h"
 #include "nsIServiceManager.h"
 static NS_DEFINE_CID(kCharsetConverterManagerCID, NS_ICHARSETCONVERTERMANAGER_CID);
 
-#define SJIS_INDEX mMapIndex[0]
-#define JIS0208_INDEX mMapIndex[1]
+#ifdef XP_OS2
+  // HTML5-incompliant behavior for OS/2, see bug 108136
+  // This is bogus. The right fix would be working around the font problems
+  // in OS/2 gfx, since this "fix" introduces script-visible DOM differences
+  // between the platforms.
+  #define SJIS_INDEX gIBM943Index[0]
+  #define JIS0208_INDEX gIBM943Index[1]
+#else
+  // HTML5 says to use Windows-31J instead of the real Shift_JIS for decoding
+  #define SJIS_INDEX gCP932Index[0]
+  #define JIS0208_INDEX gCP932Index[1]
+#endif
+
 #define JIS0212_INDEX gJIS0212Index
 #define SJIS_UNMAPPED	0x30fb
 
-void nsJapaneseToUnicode::setMapMode()
-{
-  nsresult res;
-
-  mMapIndex = gIndex;
-
-  nsCOMPtr<nsIPrefBranch> prefBranch = do_GetService(NS_PREFSERVICE_CONTRACTID);
-  if (!prefBranch) return;
-  nsXPIDLCString prefMap;
-  res = prefBranch->GetCharPref("intl.jis0208.map", getter_Copies(prefMap));
-  if (!NS_SUCCEEDED(res)) return;
-  nsCaseInsensitiveCStringComparator comparator;
-  if ( prefMap.Equals(NS_LITERAL_CSTRING("cp932"), comparator) ) {
-    mMapIndex = gCP932Index;
-  } else if ( prefMap.Equals(NS_LITERAL_CSTRING("ibm943"), comparator) ) {
-    mMapIndex = gIBM943Index;
-  }
-}
-
 NS_IMETHODIMP nsShiftJISToUnicode::Convert(
    const char * aSrc, PRInt32 * aSrcLen,
      PRUnichar * aDest, PRInt32 * aDestLen)
@@ -344,10 +333,10 @@ NS_IMETHODIMP nsEUCJPToUnicodeV2::Convert(
                 goto error_invalidchar;
               *dest++ = 0xFFFD;
                // if the first byte is valid for EUC-JP but the second 
-               // is not while being a valid US-ASCII(i.e. < 0xc0), save it
+               // is not while being a valid US-ASCII, save it
                // instead of eating it up !
-               if ( ! (*src & 0xc0)  )
-                 *dest++ = (PRUnichar) *src;;
+              if ( (PRUint8)*src < (PRUint8)0x7f )
+                --src;
             } else {
                *dest++ = gJapaneseMap[mData+off];
             }
@@ -368,7 +357,7 @@ NS_IMETHODIMP nsEUCJPToUnicodeV2::Convert(
               // if 0x8e is not followed by a valid JIS X 0201 byte
               // but by a valid US-ASCII, save it instead of eating it up.
               if ( (PRUint8)*src < (PRUint8)0x7f )
-                 *dest++ = (PRUnichar) *src;
+                --src;
             }
             mState = 0;
             if(dest >= destEnd)
