@@ -20,7 +20,7 @@
  * Portions created by the Initial Developer are Copyright (C) 2006
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s): Igor Bukanov
+ * Contributor(s): Jesse Ruderman
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -36,33 +36,23 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-gTestfile = 'regress-354145-03.js';
 
-var BUGNUMBER = 354145;
-var summary = 'Immutable XML';
+var summary = "Do not crash in js_IsXMLName";
+var BUGNUMBER = 327691;
 var actual = 'No Crash';
 var expect = 'No Crash';
 
 printBugNumber(BUGNUMBER);
 START(summary);
 
-var list = <></>
-var N = 10;
-for (var i = 0; i != N; ++i)
-  list[i] = <{"a"+i}/>;
+var A=<y/>;
+var B=A.h;
+var D=B.h;
+B.h=D.h;
+B[0]=B.h;
 
-function prepare()
-{
-  for (var i = N - 1; i >= 0; --i)
-    delete list[i];
-  gc();
-  return "test";
-}
+B.h=D.h; // Crashes at js_IsXMLName.
 
-print("Before list.contains");
-list.contains({ toString: prepare });
-print("After list.contains");
 TEST(1, expect, actual);
-print("After TEST");
 
 END();
