@@ -16,7 +16,7 @@
  *
  * The Initial Developer of the Original Code is
  * Mozilla Foundation.
- * Portions created by the Initial Developer are Copyright (C) 2005
+ * Portions created by the Initial Developer are Copyright (C) 2006
  * the Initial Developer. All Rights Reserved.
  *
  * Contributor(s): Igor Bukanov
@@ -35,41 +35,38 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-316885-03.js';
 //-----------------------------------------------------------------------------
-var BUGNUMBER = 316885;
-var summary = 'Unrooted access in jsinterp.c';
-var actual = '';
-var expect = '';
+var BUGNUMBER = 341877;
+var summary = 'GC hazard with for-in loop';
+var actual = 'No Crash';
+var expect = 'No Crash';
 
 printBugNumber(BUGNUMBER);
 printStatus (summary);
 
-var str = "test";
+var obj = { };
 
-var lval = {
-  valueOf: function() {
-    return str+"1";
+var prop = "xsomePropety".substr(1);
+
+obj.first = "first"
+
+  obj[prop] = 1;
+
+for (var elem in obj) {
+  var tmp = elem.toString();
+  delete obj[prop];
+  // ensure that prop is cut from all roots
+  prop = "xsomePropety".substr(2);
+  obj[prop] = 2;
+  delete obj[prop];
+  prop = null;
+  if (typeof gc == 'function')
+    gc();
+  for (var i = 0; i != 50000; ++i) {
+    var tmp = 1 / 3;
+    tmp /= 10;
   }
-};
-
-var ONE = 1;
-
-var rval = {
-  valueOf: function() {
-    // Make sure that the result of the previous lval.valueOf
-    // is not GC-rooted.
-    var tmp = "x"+lval;
-    if (typeof gc == "function")
-      gc();
-    for (var i = 0; i != 40000; ++i) {
-      tmp = 1e100 / ONE;
-    }
-    return str;
-  }
-};
-
-expect = (str+"1")+str;
-actual = lval+rval;
+}
 
+ 
 reportCompare(expect, actual, summary);
