@@ -16,10 +16,10 @@
  *
  * The Initial Developer of the Original Code is
  * Mozilla Foundation.
- * Portions created by the Initial Developer are Copyright (C) 2006
+ * Portions created by the Initial Developer are Copyright (C) 2007
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s):
+ * Contributor(s): Jesse Ruderman
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -35,15 +35,13 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-360969-03.js';
+
 //-----------------------------------------------------------------------------
-var BUGNUMBER = 360969;
-var summary = '2^17: local const';
+var BUGNUMBER = 385393;
+var summary = 'Regression test for bug 385393';
 var actual = 'No Crash';
 var expect = 'No Crash';
 
-var global = this;
-
 //-----------------------------------------------------------------------------
 test();
 //-----------------------------------------------------------------------------
@@ -54,21 +52,9 @@ function test()
   printBugNumber(BUGNUMBER);
   printStatus (summary);
  
-  var start = new Date();
-  var p;
-  var i;
-  var limit = 2 << 16;
-
-  for (var i = 0; i < limit; i++)
-  {
-    eval('const pv' + i + ';');
-  }
-
   reportCompare(expect, actual, summary);
 
-  var stop = new Date();
-
-  print('Elapsed time: ' + Math.floor((stop - start)/1000) + ' seconds');
+  true.watch("x", function(){});
 
   exitFunc ('test');
 }
