@@ -16,10 +16,10 @@
  *
  * The Initial Developer of the Original Code is
  * Mozilla Foundation.
- * Portions created by the Initial Developer are Copyright (C) 2008
+ * Portions created by the Initial Developer are Copyright (C) 2006
  * the Initial Developer. All Rights Reserved.
  *
- * Contributor(s): Igor Bukanov
+ * Contributor(s): Georgi Guninski
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either the GNU General Public License Version 2 or later (the "GPL"), or
@@ -35,46 +35,46 @@
  *
  * ***** END LICENSE BLOCK ***** */
 
-var gTestfile = 'regress-369696-03.js';
 //-----------------------------------------------------------------------------
-var BUGNUMBER = 396696;
-var summary = 'Do not assert: map->depth > 0" in js_LeaveSharpObject';
-var actual = '';
-var expect = '';
+var BUGNUMBER = 336409;
+var summary = 'Integer overflow in js_obj_toSource';
+var actual = 'No Crash';
+var expect = /(No Crash|InternalError: script stack space quota is exhausted|InternalError: allocation size overflow)/;
 
+printBugNumber(BUGNUMBER);
+printStatus (summary);
 
-//-----------------------------------------------------------------------------
-test();
-//-----------------------------------------------------------------------------
+expectExitCode(0);
+expectExitCode(5);
 
-function test()
+function createString(n)
 {
-  enterFunc ('test');
-  printBugNumber(BUGNUMBER);
-  printStatus (summary);
+  var l = n*1024*1024;
+  var r = 'r';
 
-  var x = [[[ { toSource: function() { gc();  }}]]];
-
-  var a = [];
-  a[0] = a;
-  a.toSource = a.toString;
-  Array.prototype.toSource.call(a);
-
-//cx->sharpObjectMap.depth == -2
-
-  (function() {
-    var tmp = [];
-    for (var i = 0; i != 30*1000; ++i) {
-      var tmp2 = [];
-      tmp.push(tmp2);
-      tmp2.toSource();
-    }
-  })();
-
-  gc();
-  x.toSource();
-
-  reportCompare(expect, actual, summary);
-
-  exitFunc ('test');
+  while (r.length < l)
+  {
+    r = r + r;
+  }
+  return r;
 }
+
+try
+{
+  var n = 128;
+  printStatus('Creating ' + n + 'MB string');
+  var r = createString(n);
+  printStatus('Done. length = ' + r.length);
+  printStatus('Creating object');
+  var o = {f1: r, f2: r, f3: r,f4: r,f5: r, f6: r, f7: r, f8: r,f9: r};
+  printStatus('object.toSource()');
+  var rr = o.toSource();
+  printStatus('Done.');
+}
+catch(ex)
+{
+  actual = ex + '';
+  print(actual);
+}
+
+reportMatch(expect, actual, summary);
