@@ -1,4 +1,5 @@
 /* -*- Mode: C++; tab-width: 8; indent-tabs-mode: nil; c-basic-offset: 4 -*-
+ * vim: set ts=8 sw=4 et tw=78:
  *
  * ***** BEGIN LICENSE BLOCK *****
  * Version: MPL 1.1/GPL 2.0/LGPL 2.1
@@ -42,20 +43,24 @@
 /* Sharable code and data for wrapper around JSObjects. */
 
 #include "xpcprivate.h"
-#include "XPCWrapper.h"
+#include "nsArrayEnumerator.h"
 #include "nsWrapperCache.h"
+#include "XPCWrapper.h"
+#include "AccessCheck.h"
+#include "nsJSUtils.h"
 
 NS_IMPL_THREADSAFE_ISUPPORTS1(nsXPCWrappedJSClass, nsIXPCWrappedJSClass)
 
 // the value of this variable is never used - we use its address as a sentinel
 static uint32 zero_methods_descriptor;
 
-void AutoScriptEvaluate::StartEvaluating(JSErrorReporter errorReporter)
+PRBool AutoScriptEvaluate::StartEvaluating(JSObject *scope, JSErrorReporter errorReporter)
 {
     NS_PRECONDITION(!mEvaluated, "AutoScriptEvaluate::Evaluate should only be called once");
 
-    if(!mJSContext)
-        return;
+    if (!mJSContext)
+        return PR_TRUE;
+
     mEvaluated = PR_TRUE;
     if(!mJSContext->errorReporter)
     {
@@ -66,6 +71,9 @@ void AutoScriptEvaluate::StartEvaluating(JSErrorReporter errorReporter)
     if (mContextHasThread)
         JS_BeginRequest(mJSContext);
 
+    if (!mEnterCompartment.enter(mJSContext, scope))
+        return PR_FALSE;
+
     // Saving the exception state keeps us from interfering with another script
     // that may also be running on this context.  This occurred first with the
     // js debugger, as described in
@@ -81,6 +89,8 @@ void AutoScriptEvaluate::StartEvaluating(JSErrorReporter errorReporter)
         mState = JS_SaveExceptionState(mJSContext);
         JS_ClearPendingException(mJSContext);
     }
+
+    return PR_TRUE;
 }
 
 AutoScriptEvaluate::~AutoScriptEvaluate()
@@ -247,6 +257,24 @@ nsXPCWrappedJSClass::CallQueryInterfaceOnJSObject(XPCCallContext& ccx,
     jsid funid;
     jsval fun;
 
+    // Don't call the actual function on a content object. We'll determine
+    // whether or not a content object is capable of implementing the
+    // interface (i.e. whether the interface is scriptable) and most content
+    // objects don't have QI implementations anyway. Also see bug 503926.
+    if(XPCPerThreadData::IsMainThread(ccx) &&
+       !xpc::AccessCheck::isChrome(jsobj->compartment()))
+    {
+        return nsnull;
+    }
+
+    // OK, it looks like we'll be calling into JS code.
+    AutoScriptEvaluate scriptEval(cx);
+
+    // XXX we should install an error reporter that will send reports to
+    // the JS error console service.
+    if (!scriptEval.StartEvaluating(jsobj))
+        return nsnull;
+
     // check upfront for the existence of the function property
     funid = mRuntime->GetStringID(XPCJSRuntime::IDX_QUERY_INTERFACE);
     if(!JS_GetPropertyById(cx, jsobj, funid, &fun) || JSVAL_IS_PRIMITIVE(fun))
@@ -272,14 +300,6 @@ nsXPCWrappedJSClass::CallQueryInterfaceOnJSObject(XPCCallContext& ccx,
             return nsnull;
     }
 
-    // OK, it looks like we'll be calling into JS code.
-
-    AutoScriptEvaluate scriptEval(cx);
-
-    // XXX we should install an error reporter that will send reports to
-    // the JS error console service.
-    scriptEval.StartEvaluating();
-
     id = xpc_NewIDObject(cx, jsobj, aIID);
     if(id)
     {
@@ -333,7 +353,7 @@ nsXPCWrappedJSClass::CallQueryInterfaceOnJSObject(XPCCallContext& ccx,
                 {
                     // JS often throws an nsresult.
                     if(JSVAL_IS_DOUBLE(jsexception))
-                        rv = (nsresult)(*JSVAL_TO_DOUBLE(jsexception));
+                        rv = (nsresult)(JSVAL_TO_DOUBLE(jsexception));
                     else
                         rv = (nsresult)(JSVAL_TO_INT(jsexception));
 
@@ -384,7 +404,8 @@ nsXPCWrappedJSClass::GetNamedPropertyAsVariant(XPCCallContext& ccx,
     nsresult rv = NS_ERROR_FAILURE;
 
     AutoScriptEvaluate scriptEval(cx);
-    scriptEval.StartEvaluating();
+    if (!scriptEval.StartEvaluating(aJSObj))
+        return NS_ERROR_FAILURE;
 
     ok = JS_ValueToId(cx, aName, &id) && 
          GetNamedPropertyAsVariantRaw(ccx, aJSObj, id, aResult, &rv);
@@ -403,22 +424,18 @@ nsXPCWrappedJSClass::BuildPropertyEnumerator(XPCCallContext& ccx,
     JSContext* cx = ccx.GetJSContext();
     nsresult retval = NS_ERROR_FAILURE;
     JSIdArray* idArray = nsnull;
-    xpcPropertyBagEnumerator* enumerator = nsnull;
     int i;
 
     // Saved state must be restored, all exits through 'out'...
     AutoScriptEvaluate scriptEval(cx);
-    scriptEval.StartEvaluating();
+    if (!scriptEval.StartEvaluating(aJSObj))
+        return NS_ERROR_FAILURE;
 
     idArray = JS_Enumerate(cx, aJSObj);
     if(!idArray)
-        goto out;
+        return retval;
     
-    enumerator = new xpcPropertyBagEnumerator(idArray->length);
-    if(!enumerator)
-        goto out;
-    NS_ADDREF(enumerator);
-        
+    nsCOMArray<nsIProperty> propertyArray(idArray->length);
     for(i = 0; i < idArray->length; i++)
     {
         nsCOMPtr<nsIVariant> value;
@@ -441,24 +458,24 @@ nsXPCWrappedJSClass::BuildPropertyEnumerator(XPCCallContext& ccx,
         if(!name)
             goto out;
 
+        size_t length;
+        const jschar *chars = JS_GetStringCharsAndLength(cx, name, &length);
+        if (!chars)
+            goto out;
+
         nsCOMPtr<nsIProperty> property = 
-            new xpcProperty((const PRUnichar*) JS_GetStringChars(name), 
-                            (PRUint32) JS_GetStringLength(name),
-                            value);
+            new xpcProperty(chars, (PRUint32) length, value);
         if(!property)
             goto out;
 
-        if(!enumerator->AppendElement(property))
+        if(!propertyArray.AppendObject(property))
             goto out;
     }
 
-    NS_ADDREF(*aEnumerate = enumerator);
-    retval = NS_OK;
+    retval = NS_NewArrayEnumerator(aEnumerate, propertyArray);
 
 out:
-    NS_IF_RELEASE(enumerator);
-    if(idArray)
-        JS_DestroyIdArray(cx, idArray);
+    JS_DestroyIdArray(cx, idArray);
 
     return retval;
 }
@@ -487,44 +504,6 @@ NS_IMETHODIMP xpcProperty::GetValue(nsIVariant * *aValue)
     return NS_OK;
 }
 
-/***************************************************************************/
-
-NS_IMPL_ISUPPORTS1(xpcPropertyBagEnumerator, nsISimpleEnumerator)
-
-xpcPropertyBagEnumerator::xpcPropertyBagEnumerator(PRUint32 count)
-    : mIndex(0), mCount(0)
-{
-    mArray.SizeTo(count);
-}
-
-JSBool xpcPropertyBagEnumerator::AppendElement(nsISupports* element)
-{
-    if(!mArray.AppendElement(element))
-        return JS_FALSE;
-    mCount++;
-    return JS_TRUE;
-}
-
-/* boolean hasMoreElements (); */
-NS_IMETHODIMP xpcPropertyBagEnumerator::HasMoreElements(PRBool *_retval)
-{
-    *_retval = mIndex < mCount;
-    return NS_OK;
-}
-
-/* nsISupports getNext (); */
-NS_IMETHODIMP xpcPropertyBagEnumerator::GetNext(nsISupports **_retval)
-{
-    if(!(mIndex < mCount))
-    {
-        NS_ERROR("Bad nsISimpleEnumerator caller!");
-        return NS_ERROR_FAILURE;    
-    }
-    
-    *_retval = mArray.ElementAt(mIndex++);
-    return *_retval ? NS_OK : NS_ERROR_FAILURE;
-}
-
 /***************************************************************************/
 // This 'WrappedJSIdentity' class and singleton allow us to figure out if
 // any given nsISupports* is implemented by a WrappedJS object. This is done
@@ -585,6 +564,10 @@ GetContextFromObject(JSObject *obj)
     XPCCallContext ccx(NATIVE_CALLER);
     if(!ccx.IsValid())
         return nsnull;
+
+    JSAutoEnterCompartment ac;
+    if(!ac.enter(ccx, obj))
+        return nsnull;
     XPCWrappedNativeScope* scope =
         XPCWrappedNativeScope::FindInJSObjectScope(ccx, obj);
     XPCContext *xpcc = scope->GetContext();
@@ -801,7 +784,7 @@ nsXPCWrappedJSClass::DelegatedQueryInterface(nsXPCWrappedJS* self,
         PRBool isSystem;
         rv = secMan->IsSystemPrincipal(objPrin, &isSystem);
         if((NS_FAILED(rv) || !isSystem) &&
-           !IS_WRAPPER_CLASS(STOBJ_GET_CLASS(selfObj)))
+           !IS_WRAPPER_CLASS(selfObj->getClass()))
         {
             // A content object.
             nsRefPtr<SameOriginCheckedComponent> checked =
@@ -1138,7 +1121,7 @@ nsXPCWrappedJSClass::CheckForException(XPCCallContext & ccx,
                     JSStackFrame * fp = nsnull;
                     while((fp = JS_FrameIterator(cx, &fp)))
                     {
-                        if(!JS_IsNativeFrame(cx, fp))
+                        if(JS_IsScriptFrame(cx, fp))
                         {
                             onlyNativeStackFrames = PR_FALSE;
                             break;
@@ -1238,11 +1221,14 @@ nsXPCWrappedJSClass::CheckForException(XPCCallContext & ccx,
                                     rv = location->GetFilename(getter_Copies(sourceName));
                                 }
 
-                                rv = scriptError->Init(newMessage.get(),
-                                                       NS_ConvertASCIItoUTF16(sourceName).get(),
-                                                       nsnull,
-                                                       lineNumber, 0, 0,
-                                                       "XPConnect JavaScript");
+                                nsCOMPtr<nsIScriptError2> scriptError2 =
+                                    do_QueryInterface(scriptError);
+                                rv = scriptError2->InitWithWindowID(newMessage.get(),
+                                                                    NS_ConvertASCIItoUTF16(sourceName).get(),
+                                                                    nsnull,
+                                                                    lineNumber, 0, 0,
+                                                                    "XPConnect JavaScript",
+                                                                    nsJSUtils::GetCurrentlyRunningCodeWindowID(cx));
                                 if(NS_FAILED(rv))
                                     scriptError = nsnull;
                             }
@@ -1272,28 +1258,35 @@ nsXPCWrappedJSClass::CheckForException(XPCCallContext & ccx,
     return NS_ERROR_FAILURE;
 }
 
+class ContextPrincipalGuard
+{
+    nsIScriptSecurityManager *ssm;
+    XPCCallContext &ccx;
+  public:
+    ContextPrincipalGuard(XPCCallContext &ccx)
+      : ssm(nsnull), ccx(ccx) {}
+    void principalPushed(nsIScriptSecurityManager *ssm) { this->ssm = ssm; }
+    ~ContextPrincipalGuard() { if (ssm) ssm->PopContextPrincipal(ccx); }
+};
+
 NS_IMETHODIMP
 nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
                                 const XPTMethodDescriptor* info,
                                 nsXPTCMiniVariant* nativeParams)
 {
-    jsval* stackbase = nsnull;
     jsval* sp = nsnull;
+    jsval* argv = nsnull;
     uint8 i;
     uint8 argc=0;
-    uint8 stack_size;
-    jsval result;
     uint8 paramCount=0;
     nsresult retval = NS_ERROR_FAILURE;
     nsresult pending_result = NS_OK;
     JSBool success;
     JSBool readyToDoTheCall = JS_FALSE;
     nsID  param_iid;
-    uint8 outConversionFailedIndex;
     JSObject* obj;
     const char* name = info->name;
     jsval fval;
-    void* mark;
     JSBool foundDependentParam;
     XPCContext* xpcc;
     JSContext* cx;
@@ -1316,54 +1309,69 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
         cx = nsnull;
     }
 
-    AutoScriptEvaluate scriptEval(cx);
-#ifdef DEBUG_stats_jband
-    PRIntervalTime startTime = PR_IntervalNow();
-    PRIntervalTime endTime = 0;
-    static int totalTime = 0;
-
-
-    static int count = 0;
-    static const int interval = 10;
-    if(0 == (++count % interval))
-        printf("<<<<<<<< %d calls on nsXPCWrappedJSs made.  (%d)\n", count, PR_IntervalToMilliseconds(totalTime));
-#endif
+    if(!cx || !xpcc || !IsReflectable(methodIndex))
+        return NS_ERROR_FAILURE;
 
     obj = thisObj = wrapper->GetJSObject();
 
+    JSAutoEnterCompartment ac;
+    if (!ac.enter(cx, obj))
+        return NS_ERROR_FAILURE;
+
+    ccx.SetScopeForNewJSObjects(obj);
+
+    js::AutoValueVector args(cx);
+    AutoScriptEvaluate scriptEval(cx);
+    ContextPrincipalGuard principalGuard(ccx);
+
     // XXX ASSUMES that retval is last arg. The xpidl compiler ensures this.
     paramCount = info->num_args;
     argc = paramCount -
         (paramCount && XPT_PD_IS_RETVAL(info->params[paramCount-1].flags) ? 1 : 0);
 
-    if(!cx || !xpcc || !IsReflectable(methodIndex))
+    if (!scriptEval.StartEvaluating(obj, xpcWrappedJSErrorReporter))
         goto pre_call_clean_up;
 
-    scriptEval.StartEvaluating(xpcWrappedJSErrorReporter);
-
     xpcc->SetPendingResult(pending_result);
     xpcc->SetException(nsnull);
     ccx.GetThreadData()->SetException(nsnull);
 
-    // We use js_AllocStack, js_Invoke, and js_FreeStack so that the gcthings
-    // we use as args will be rooted by the engine as we do conversions and
-    // prepare to do the function call. This adds a fair amount of complexity,
-    // but is a good optimization compared to calling JS_AddRoot for each item.
+    if(XPCPerThreadData::IsMainThread(ccx))
+    {
+        // TODO Remove me in favor of security wrappers.
+        nsIScriptSecurityManager *ssm = XPCWrapper::GetSecurityManager();
+        if(ssm)
+        {
+            nsIPrincipal *objPrincipal =
+                xpc::AccessCheck::getPrincipal(obj->compartment());
+            if(objPrincipal)
+            {
+                JSStackFrame* fp = nsnull;
+                nsresult rv =
+                    ssm->PushContextPrincipal(ccx, JS_FrameIterator(ccx, &fp),
+                                              objPrincipal);
+                if(NS_FAILED(rv))
+                {
+                    JS_ReportOutOfMemory(ccx);
+                    retval = NS_ERROR_OUT_OF_MEMORY;
+                    goto pre_call_clean_up;
+                }
 
-    js_LeaveTrace(cx);
+                principalGuard.principalPushed(ssm);
+            }
+        }
+    }
+
+    // We use js_Invoke so that the gcthings we use as args will be rooted by
+    // the engine as we do conversions and prepare to do the function call.
+
+    js::LeaveTrace(cx);
 
     // setup stack
 
     // if this isn't a function call then we don't need to push extra stuff
-    if(XPT_MD_IS_GETTER(info->flags) || XPT_MD_IS_SETTER(info->flags))
+    if (!(XPT_MD_IS_SETTER(info->flags) || XPT_MD_IS_GETTER(info->flags)))
     {
-        stack_size = argc;
-    }
-    else
-    {
-        // allocate extra space for function and 'this'
-        stack_size = argc + 2;
-
         // We get fval before allocating the stack to avoid gc badness that can
         // happen if the GetProperty call leaves our request and the gc runs
         // while the stack we allocate contains garbage.
@@ -1434,10 +1442,11 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
                             if(newThis)
                             {
                                 jsval v;
+                                xpcObjectHelper helper(newThis);
                                 JSBool ok =
                                   XPCConvert::NativeInterface2JSObject(ccx,
-                                        &v, nsnull, newThis, newWrapperIID,
-                                        nsnull, nsnull, obj, PR_FALSE, PR_FALSE,
+                                        &v, nsnull, helper, newWrapperIID,
+                                        nsnull, PR_FALSE, PR_FALSE,
                                         nsnull);
                                 if(newWrapperIID)
                                     nsMemory::Free(newWrapperIID);
@@ -1446,6 +1455,8 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
                                     goto pre_call_clean_up;
                                 }
                                 thisObj = JSVAL_TO_OBJECT(v);
+                                if(!JS_WrapObject(cx, &thisObj))
+                                    goto pre_call_clean_up;
                             }
                         }
                     }
@@ -1463,30 +1474,21 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
         }
     }
 
-    // if stack_size is zero then we won't be needing a stack
-    if(stack_size && !(stackbase = sp = js_AllocStack(cx, stack_size, &mark)))
+    if (!args.resize(argc))
     {
         retval = NS_ERROR_OUT_OF_MEMORY;
         goto pre_call_clean_up;
     }
 
-    NS_ASSERTION(XPT_MD_IS_GETTER(info->flags) || sp,
-                 "Only a getter needs no stack.");
-
-    // this is a function call, so push function and 'this'
-    if(stack_size != argc)
-    {
-        *sp++ = fval;
-        *sp++ = OBJECT_TO_JSVAL(thisObj);
-    }
-
-    // make certain we leave no garbage in the stack
-    for(i = 0; i < argc; i++)
-    {
-        sp[i] = JSVAL_VOID;
-    }
+    argv = args.jsval_begin();
+    sp = argv;
 
     // build the args
+    // NB: This assignment *looks* wrong because we haven't yet called our
+    // function. However, we *have* already entered the compartmen that we're
+    // about to call, and that's the global that we want here. In other words:
+    // we're trusting the JS engine to come up with a good global to use for
+    // our object (whatever it was).
     for(i = 0; i < argc; i++)
     {
         const nsXPTParamInfo& param = info->params[i];
@@ -1541,45 +1543,13 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
                     goto pre_call_clean_up;
             }
 
-            // Figure out what our callee is
-            if(XPT_MD_IS_GETTER(info->flags) || XPT_MD_IS_SETTER(info->flags))
-            {
-                // Pull the getter or setter off of |obj|
-                uintN attrs;
-                JSBool found;
-                JSPropertyOp getter;
-                JSPropertyOp setter;
-                JSBool ok =
-                    JS_GetPropertyAttrsGetterAndSetter(cx, obj, name,
-                                                       &attrs, &found,
-                                                       &getter, &setter);
-                if(ok)
-                {
-                    if(XPT_MD_IS_GETTER(info->flags) && (attrs & JSPROP_GETTER))
-                    {
-                        // JSPROP_GETTER means the getter is actually a
-                        // function object.
-                        ccx.SetCallee(JS_FUNC_TO_DATA_PTR(JSObject*, getter));
-                    }
-                    else if(XPT_MD_IS_SETTER(info->flags) && (attrs & JSPROP_SETTER))
-                    {
-                        // JSPROP_SETTER means the setter is actually a
-                        // function object.
-                        ccx.SetCallee(JS_FUNC_TO_DATA_PTR(JSObject*, setter));
-                    }
-                }
-            }
-            else if(JSVAL_IS_OBJECT(fval))
-            {
-                ccx.SetCallee(JSVAL_TO_OBJECT(fval));
-            }
-
             if(isArray)
             {
-
-                if(!XPCConvert::NativeArray2JS(ccx, &val, (const void**)&pv->val,
+                XPCLazyCallContext lccx(ccx);
+                if(!XPCConvert::NativeArray2JS(lccx, &val,
+                                               (const void**)&pv->val,
                                                datum_type, &param_iid,
-                                               array_count, obj, nsnull))
+                                               array_count, nsnull))
                     goto pre_call_clean_up;
             }
             else if(isSizedString)
@@ -1593,12 +1563,12 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
             else
             {
                 if(!XPCConvert::NativeData2JS(ccx, &val, &pv->val, type,
-                                              &param_iid, obj, nsnull))
+                                              &param_iid, nsnull))
                     goto pre_call_clean_up;
             }
         }
 
-        if(param.IsOut())
+        if(param.IsOut() || param.IsDipper())
         {
             // create an 'out' object
             JSObject* out_obj = NewOutObject(cx, obj);
@@ -1623,8 +1593,6 @@ nsXPCWrappedJSClass::CallMethod(nsXPCWrappedJS* wrapper, uint16 methodIndex,
             *sp++ = val;
     }
 
-
-
     readyToDoTheCall = JS_TRUE;
 
 pre_call_clean_up:
@@ -1676,26 +1644,34 @@ pre_call_clean_up:
     // Make sure "this" doesn't get deleted during this call.
     nsCOMPtr<nsIXPCWrappedJSClass> kungFuDeathGrip(this);
 
-    result = JSVAL_NULL;
-    AUTO_MARK_JSVAL(ccx, &result);
-
     if(!readyToDoTheCall)
-        goto done;
+        return retval;
 
     // do the deed - note exceptions
 
     JS_ClearPendingException(cx);
 
+    jsval rval;
     if(XPT_MD_IS_GETTER(info->flags))
-        success = JS_GetProperty(cx, obj, name, &result);
+    {
+        success = JS_GetProperty(cx, obj, name, argv);
+        rval = *argv;
+    }
     else if(XPT_MD_IS_SETTER(info->flags))
-        success = JS_SetProperty(cx, obj, name, sp-1);
+    {
+        success = JS_SetProperty(cx, obj, name, argv);
+        rval = *argv;
+    }
     else
     {
         if(!JSVAL_IS_PRIMITIVE(fval))
         {
-            success = js_Invoke(cx, argc, stackbase, 0);
-            result = *stackbase;
+            uint32 oldOpts = JS_GetOptions(cx);
+            JS_SetOptions(cx, oldOpts | JSOPTION_DONT_REPORT_UNCAUGHT);
+
+            success = JS_CallFunctionValue(cx, thisObj, fval, argc, argv, &rval);
+
+            JS_SetOptions(cx, oldOpts);
         }
         else
         {
@@ -1722,7 +1698,7 @@ pre_call_clean_up:
         }
     }
 
-    if (!success)
+    if(!success)
     {
         PRBool forceReport;
         if(NS_FAILED(mInfo->IsFunction(&forceReport)))
@@ -1731,15 +1707,11 @@ pre_call_clean_up:
         // May also want to check if we're moving from content->chrome and force
         // a report in that case.
 
-        retval = CheckForException(ccx, name, GetInterfaceName(), forceReport);
-        goto done;
+        return CheckForException(ccx, name, GetInterfaceName(), forceReport);
     }
 
     ccx.GetThreadData()->SetException(nsnull); // XXX necessary?
 
-#define HANDLE_OUT_CONVERSION_FAILURE       \
-        {outConversionFailedIndex = i; break;}
-
     // convert out args and result
     // NOTE: this is the total number of native params, not just the args
     // Convert independent params only.
@@ -1747,7 +1719,6 @@ pre_call_clean_up:
     // the params upon which they depend will have already been converted -
     // regardless of ordering.
 
-    outConversionFailedIndex = paramCount;
     foundDependentParam = JS_FALSE;
     for(i = 0; i < paramCount; i++)
     {
@@ -1773,12 +1744,12 @@ pre_call_clean_up:
             pv = (nsXPTCMiniVariant*) nativeParams[i].val.p;
 
         if(param.IsRetval())
-            val = result;
-        else if(JSVAL_IS_PRIMITIVE(stackbase[i+2]) ||
-                !JS_GetPropertyById(cx, JSVAL_TO_OBJECT(stackbase[i+2]),
+            val = rval;
+        else if(JSVAL_IS_PRIMITIVE(argv[i]) ||
+                !JS_GetPropertyById(cx, JSVAL_TO_OBJECT(argv[i]),
                     mRuntime->GetStringID(XPCJSRuntime::IDX_VALUE),
                     &val))
-            HANDLE_OUT_CONVERSION_FAILURE
+            break;
 
         // setup allocator and/or iid
 
@@ -1787,18 +1758,18 @@ pre_call_clean_up:
             if(NS_FAILED(GetInterfaceInfo()->
                             GetIIDForParamNoAlloc(methodIndex, &param,
                                                   &param_iid)))
-                HANDLE_OUT_CONVERSION_FAILURE
+                break;
         }
         else if(type.IsPointer() && !param.IsShared() && !param.IsDipper())
             useAllocator = JS_TRUE;
 
         if(!XPCConvert::JSData2Native(ccx, &pv->val, val, type,
                                       useAllocator, &param_iid, nsnull))
-            HANDLE_OUT_CONVERSION_FAILURE
+            break;
     }
 
     // if any params were dependent, then we must iterate again to convert them.
-    if(foundDependentParam && outConversionFailedIndex == paramCount)
+    if(foundDependentParam && i == paramCount)
     {
         for(i = 0; i < paramCount; i++)
         {
@@ -1824,11 +1795,11 @@ pre_call_clean_up:
             pv = (nsXPTCMiniVariant*) nativeParams[i].val.p;
 
             if(param.IsRetval())
-                val = result;
-            else if(!JS_GetPropertyById(cx, JSVAL_TO_OBJECT(stackbase[i+2]),
+                val = rval;
+            else if(!JS_GetPropertyById(cx, JSVAL_TO_OBJECT(argv[i]),
                         mRuntime->GetStringID(XPCJSRuntime::IDX_VALUE),
                         &val))
-                HANDLE_OUT_CONVERSION_FAILURE
+                break;
 
             // setup allocator and/or iid
 
@@ -1836,7 +1807,7 @@ pre_call_clean_up:
             {
                 if(NS_FAILED(mInfo->GetTypeForParam(methodIndex, &param, 1,
                                                     &datum_type)))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                    break;
             }
             else
                 datum_type = type;
@@ -1846,7 +1817,7 @@ pre_call_clean_up:
                if(!GetInterfaceTypeFromParam(cx, info, param, methodIndex,
                                              datum_type, nativeParams,
                                              &param_iid))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                   break;
             }
             else if(type.IsPointer() && !param.IsShared())
                 useAllocator = JS_TRUE;
@@ -1856,7 +1827,7 @@ pre_call_clean_up:
                 if(!GetArraySizeFromParam(cx, info, param, methodIndex,
                                           i, GET_LENGTH, nativeParams,
                                           &array_count))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                    break;
             }
 
             if(isArray)
@@ -1867,7 +1838,7 @@ pre_call_clean_up:
                                                datum_type,
                                                useAllocator, &param_iid,
                                                nsnull))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                    break;
             }
             else if(isSizedString)
             {
@@ -1876,19 +1847,19 @@ pre_call_clean_up:
                                                    array_count, array_count,
                                                    datum_type, useAllocator,
                                                    nsnull))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                    break;
             }
             else
             {
                 if(!XPCConvert::JSData2Native(ccx, &pv->val, val, type,
                                               useAllocator, &param_iid,
                                               nsnull))
-                    HANDLE_OUT_CONVERSION_FAILURE
+                    break;
             }
         }
     }
 
-    if(outConversionFailedIndex != paramCount)
+    if(i != paramCount)
     {
         // We didn't manage all the result conversions!
         // We have to cleanup any junk that *did* get converted.
@@ -1937,15 +1908,6 @@ pre_call_clean_up:
         retval = pending_result;
     }
 
-done:
-    if(sp)
-        js_FreeStack(cx, mark);
-
-#ifdef DEBUG_stats_jband
-    endTime = PR_IntervalNow();
-    printf("%s::%s %d ( c->js ) \n", GetInterfaceName(), info->GetName(), PR_IntervalToMilliseconds(endTime-startTime));
-    totalTime += endTime-startTime;
-#endif
     return retval;
 }
 
@@ -1979,7 +1941,7 @@ nsXPCWrappedJSClass::DebugDump(PRInt16 depth)
         char * iid = mIID.ToString();
         XPC_LOG_ALWAYS(("IID number is %s", iid ? iid : "invalid"));
         if(iid)
-            PR_Free(iid);
+            NS_Free(iid);
         XPC_LOG_ALWAYS(("InterfaceInfo @ %x", mInfo));
         uint16 methodCount = 0;
         if(depth)
