@@ -22,7 +22,7 @@
  *
  * Contributor(s):
  *   Pierre Phaneuf <pp@ludusdesign.com>
- *   Mats Palmgren <mats.palmgren@bredband.net>
+ *   Mats Palmgren <matspal@gmail.com>
  *
  * Alternatively, the contents of this file may be used under the terms of
  * either of the GNU General Public License Version 2 or later (the "GPL"),
@@ -59,9 +59,6 @@
 #include "nsIPresShell.h"
 #include "nsIRenderingContext.h"
 #include "nsIDeviceContext.h"
-#include "nsIView.h"
-#include "nsIScrollableView.h"
-#include "nsIViewManager.h"
 #include "nsPresContext.h"
 #include "nsILookAndFeel.h"
 #include "nsBlockFrame.h"
@@ -84,6 +81,71 @@ static const PRInt32 kMinBidiIndicatorPixels = 2;
 #include "nsContentUtils.h"
 #endif //IBMBIDI
 
+/**
+ * Find the first frame in an in-order traversal of the frame subtree rooted
+ * at aFrame which is either a text frame logically at the end of a line,
+ * or which is aStopAtFrame. Return null if no such frame is found. We don't
+ * descend into the children of non-eLineParticipant frames.
+ */
+static nsIFrame*
+CheckForTrailingTextFrameRecursive(nsIFrame* aFrame, nsIFrame* aStopAtFrame)
+{
+  if (aFrame == aStopAtFrame ||
+      ((aFrame->GetType() == nsGkAtoms::textFrame &&
+       (static_cast<nsTextFrame*>(aFrame))->IsAtEndOfLine())))
+    return aFrame;
+  if (!aFrame->IsFrameOfType(nsIFrame::eLineParticipant))
+    return nsnull;
+
+  for (nsIFrame* f = aFrame->GetFirstChild(nsnull); f; f = f->GetNextSibling())
+  {
+    nsIFrame* r = CheckForTrailingTextFrameRecursive(f, aStopAtFrame);
+    if (r)
+      return r;
+  }
+  return nsnull;
+}
+
+static nsLineBox*
+FindContainingLine(nsIFrame* aFrame)
+{
+  while (aFrame && aFrame->IsFrameOfType(nsIFrame::eLineParticipant))
+  {
+    nsIFrame* parent = aFrame->GetParent();
+    nsBlockFrame* blockParent = nsLayoutUtils::GetAsBlock(parent);
+    if (blockParent)
+    {
+      PRBool isValid;
+      nsBlockInFlowLineIterator iter(blockParent, aFrame, &isValid);
+      return isValid ? iter.GetLine().get() : nsnull;
+    }
+    aFrame = parent;
+  }
+  return nsnull;
+}
+
+static void
+AdjustCaretFrameForLineEnd(nsIFrame** aFrame, PRInt32* aOffset)
+{
+  nsLineBox* line = FindContainingLine(*aFrame);
+  if (!line)
+    return;
+  PRInt32 count = line->GetChildCount();
+  for (nsIFrame* f = line->mFirstChild; count > 0; --count, f = f->GetNextSibling())
+  {
+    nsIFrame* r = CheckForTrailingTextFrameRecursive(f, *aFrame);
+    if (r == *aFrame)
+      return;
+    if (r)
+    {
+      *aFrame = r;
+      NS_ASSERTION(r->GetType() == nsGkAtoms::textFrame, "Expected text frame");
+      *aOffset = (static_cast<nsTextFrame*>(r))->GetContentEnd();
+      return;
+    }
+  }
+}
+
 //-----------------------------------------------------------------------------
 
 nsCaret::nsCaret()
@@ -91,6 +153,7 @@ nsCaret::nsCaret()
 , mBlinkRate(500)
 , mVisible(PR_FALSE)
 , mDrawn(PR_FALSE)
+, mPendingDraw(PR_FALSE)
 , mReadOnly(PR_FALSE)
 , mShowDuringSelection(PR_FALSE)
 , mIgnoreUserModify(PR_TRUE)
@@ -193,7 +256,7 @@ nsCaret::Metrics nsCaret::ComputeMetrics(nsIFrame* aFrame, PRInt32 aOffset, nsco
     caretWidth += nsPresContext::CSSPixelsToAppUnits(1);
   }
   nscoord bidiIndicatorSize = nsPresContext::CSSPixelsToAppUnits(kMinBidiIndicatorPixels);
-  bidiIndicatorSize = PR_MAX(caretWidth, bidiIndicatorSize);
+  bidiIndicatorSize = NS_MAX(caretWidth, bidiIndicatorSize);
 
   // Round them to device pixels. Always round down, except that anything
   // between 0 and 1 goes up to 1 so we don't let the caret disappear.
@@ -255,8 +318,8 @@ void nsCaret::SetCaretVisible(PRBool inMakeVisible)
 {
   mVisible = inMakeVisible;
   if (mVisible) {
-    StartBlinking();
     SetIgnoreUserModify(PR_TRUE);
+    StartBlinking();
   } else {
     StopBlinking();
     SetIgnoreUserModify(PR_FALSE);
@@ -279,94 +342,88 @@ void nsCaret::SetCaretReadOnly(PRBool inMakeReadonly)
   mReadOnly = inMakeReadonly;
 }
 
-
-//-----------------------------------------------------------------------------
-nsresult nsCaret::GetCaretCoordinates(EViewCoordinates aRelativeToType,
-                                      nsISelection *aDOMSel,
-                                      nsRect *outCoordinates,
-                                      PRBool *outIsCollapsed,
-                                      nsIView **outView)
+nsresult
+nsCaret::GetGeometryForFrame(nsIFrame* aFrame,
+                             PRInt32   aFrameOffset,
+                             nsRect*   aRect,
+                             nscoord*  aBidiIndicatorSize)
 {
-  if (!mPresShell)
-    return NS_ERROR_NOT_INITIALIZED;
-  if (!outCoordinates || !outIsCollapsed)
-    return NS_ERROR_NULL_POINTER;
+  nsPoint framePos(0, 0);
+  nsresult rv = aFrame->GetPointFromOffset(aFrameOffset, &framePos);
+  if (NS_FAILED(rv))
+    return rv;
 
-  nsCOMPtr<nsISelection> domSelection = aDOMSel;
+  nsIFrame *frame = aFrame->GetContentInsertionFrame();
+  NS_ASSERTION(frame, "We should not be in the middle of reflow");
+  nscoord baseline = frame->GetCaretBaseline();
+  nscoord ascent = 0, descent = 0;
+  nsCOMPtr<nsIFontMetrics> fm;
+  nsLayoutUtils::GetFontMetricsForFrame(aFrame, getter_AddRefs(fm));
+  NS_ASSERTION(fm, "We should be able to get the font metrics");
+  if (fm) {
+    fm->GetMaxAscent(ascent);
+    fm->GetMaxDescent(descent);
+  }
+  nscoord height = ascent + descent;
+  framePos.y = baseline - ascent;
+  Metrics caretMetrics = ComputeMetrics(aFrame, aFrameOffset, height);
+  *aRect = nsRect(framePos, nsSize(caretMetrics.mCaretWidth, height));
 
-  if (outView)
-    *outView = nsnull;
+  // Clamp the x-position to be within our scroll frame. If we don't, then it
+  // clips us, and we don't appear at all. See bug 335560.
+  nsIFrame *scrollFrame =
+    nsLayoutUtils::GetClosestFrameOfType(aFrame, nsGkAtoms::scrollFrame);
+  if (scrollFrame) {
+    // First, use the scrollFrame to get at the scrollable view that we're in.
+    nsIScrollableFrame *sf = do_QueryFrame(scrollFrame);
+    nsIFrame *scrolled = sf->GetScrolledFrame();
+    nsRect caretInScroll = *aRect + aFrame->GetOffsetTo(scrolled);
+
+    // Now see if thet caret extends beyond the view's bounds. If it does,
+    // then snap it back, put it as close to the edge as it can.
+    nscoord overflow = caretInScroll.XMost() -
+      scrolled->GetVisualOverflowRectRelativeToSelf().width;
+    if (overflow > 0)
+      aRect->x -= overflow;
+  }
+
+  if (aBidiIndicatorSize)
+    *aBidiIndicatorSize = caretMetrics.mBidiIndicatorSize;
+
+  return NS_OK;
+}
+
+nsIFrame* nsCaret::GetGeometry(nsISelection* aSelection, nsRect* aRect,
+                               nscoord* aBidiIndicatorSize)
+{
+  nsCOMPtr<nsIDOMNode> focusNode;
+  nsresult rv = aSelection->GetFocusNode(getter_AddRefs(focusNode));
+  if (NS_FAILED(rv) || !focusNode)
+    return nsnull;
 
-  // fill in defaults for failure
-  outCoordinates->x = -1;
-  outCoordinates->y = -1;
-  outCoordinates->width = -1;
-  outCoordinates->height = -1;
-  *outIsCollapsed = PR_FALSE;
-  
-  nsresult err = domSelection->GetIsCollapsed(outIsCollapsed);
-  if (NS_FAILED(err)) 
-    return err;
-    
-  nsCOMPtr<nsIDOMNode>  focusNode;
-  
-  err = domSelection->GetFocusNode(getter_AddRefs(focusNode));
-  if (NS_FAILED(err))
-    return err;
-  if (!focusNode)
-    return NS_ERROR_FAILURE;
-  
   PRInt32 focusOffset;
-  err = domSelection->GetFocusOffset(&focusOffset);
-  if (NS_FAILED(err))
-    return err;
+  rv = aSelection->GetFocusOffset(&focusOffset);
+  if (NS_FAILED(rv))
+    return nsnull;
     
   nsCOMPtr<nsIContent> contentNode = do_QueryInterface(focusNode);
   if (!contentNode)
-    return NS_ERROR_FAILURE;
-
-  // find the frame that contains the content node that has focus
-  nsIFrame*       theFrame = nsnull;
-  PRInt32         theFrameOffset = 0;
+    return nsnull;
 
   nsCOMPtr<nsFrameSelection> frameSelection = GetFrameSelection();
   if (!frameSelection)
-    return NS_ERROR_FAILURE;
+    return nsnull;
   PRUint8 bidiLevel = frameSelection->GetCaretBidiLevel();
-  
-  err = GetCaretFrameForNodeOffset(contentNode, focusOffset,
-                                   frameSelection->GetHint(), bidiLevel,
-                                   &theFrame, &theFrameOffset);
-  if (NS_FAILED(err) || !theFrame)
-    return err;
-  
-  nsPoint   viewOffset(0, 0);
-  nsIView   *drawingView;     // views are not refcounted
+  nsIFrame* frame;
+  PRInt32 frameOffset;
+  rv = GetCaretFrameForNodeOffset(contentNode, focusOffset,
+                                  frameSelection->GetHint(), bidiLevel,
+                                  &frame, &frameOffset);
+  if (NS_FAILED(rv) || !frame)
+    return nsnull;
 
-  GetViewForRendering(theFrame, aRelativeToType, viewOffset, &drawingView, outView);
-  if (!drawingView)
-    return NS_ERROR_UNEXPECTED;
- 
-  nsPoint   framePos(0, 0);
-  err = theFrame->GetPointFromOffset(theFrameOffset, &framePos);
-  if (NS_FAILED(err))
-    return err;
-
-  // we don't need drawingView anymore so reuse that; reset viewOffset values for our purposes
-  if (aRelativeToType == eClosestViewCoordinates)
-  {
-    theFrame->GetOffsetFromView(viewOffset, &drawingView);
-    if (outView)
-      *outView = drawingView;
-  }
-  // now add the frame offset to the view offset, and we're done
-  viewOffset += framePos;
-  outCoordinates->x = viewOffset.x;
-  outCoordinates->y = viewOffset.y;
-  outCoordinates->height = theFrame->GetSize().height;
-  outCoordinates->width = ComputeMetrics(theFrame, theFrameOffset, outCoordinates->height).mCaretWidth;
-  
-  return NS_OK;
+  GetGeometryForFrame(frame, frameOffset, aRect, aBidiIndicatorSize);
+  return frame;
 }
 
 void nsCaret::DrawCaretAfterBriefDelay()
@@ -425,7 +482,7 @@ nsresult nsCaret::DrawAtPosition(nsIDOMNode* aNode, PRInt32 aOffset)
   return rv;
 }
 
-nsIFrame * nsCaret::GetCaretFrame()
+nsIFrame * nsCaret::GetCaretFrame(PRInt32 *aOffset)
 {
   // Return null if we're not drawn to prevent anybody from trying to draw us.
   if (!mDrawn)
@@ -433,14 +490,17 @@ nsIFrame * nsCaret::GetCaretFrame()
 
   // Recompute the frame that we're supposed to draw in to guarantee that
   // we're not going to try to draw into a stale (dead) frame.
-  PRInt32 unused;
+  PRInt32 offset;
   nsIFrame *frame = nsnull;
   nsresult rv = GetCaretFrameForNodeOffset(mLastContent, mLastContentOffset,
                                            mLastHint, mLastBidiLevel, &frame,
-                                           &unused);
+                                           &offset);
   if (NS_FAILED(rv))
     return nsnull;
 
+  if (aOffset) {
+    *aOffset = offset;
+  }
   return frame;
 }
 
@@ -449,7 +509,7 @@ void nsCaret::InvalidateOutsideCaret()
   nsIFrame *frame = GetCaretFrame();
 
   // Only invalidate if we are not fully contained by our frame's rect.
-  if (frame && !frame->GetOverflowRect().Contains(GetCaretRect()))
+  if (frame && !frame->GetVisualOverflowRect().Contains(GetCaretRect()))
     InvalidateRects(mCaretRect, GetHookRect(), frame);
 }
 
@@ -473,7 +533,10 @@ void nsCaret::PaintCaret(nsDisplayListBuilder *aBuilder,
   NS_ASSERTION(mDrawn, "The caret shouldn't be drawing");
 
   const nsRect drawCaretRect = mCaretRect + aOffset;
-  nscolor cssColor = aForFrame->GetStyleColor()->mColor;
+  PRInt32 contentOffset;
+  nsIFrame* frame = GetCaretFrame(&contentOffset);
+  NS_ASSERTION(frame == aForFrame, "We're referring different frame");
+  nscolor foregroundColor = aForFrame->GetCaretColorAt(contentOffset);
 
   // Only draw the native caret if the foreground color matches that of
   // -moz-fieldtext (the color of the text in a textbox). If it doesn't match
@@ -487,7 +550,7 @@ void nsCaret::PaintCaret(nsDisplayListBuilder *aBuilder,
       nsILookAndFeel* lookAndFeel = presContext->LookAndFeel();
       nscolor fieldText;
       if (NS_SUCCEEDED(lookAndFeel->GetColor(nsILookAndFeel::eColor__moz_fieldtext, fieldText)) &&
-          fieldText == cssColor) {
+          fieldText == foregroundColor) {
         theme->DrawWidgetBackground(aCtx, aForFrame, NS_THEME_TEXTFIELD_CARET,
                                     drawCaretRect, drawCaretRect);
         return;
@@ -495,7 +558,7 @@ void nsCaret::PaintCaret(nsDisplayListBuilder *aBuilder,
     }
   }
 
-  aCtx->SetColor(cssColor);
+  aCtx->SetColor(foregroundColor);
   aCtx->FillRect(drawCaretRect);
   if (!GetHookRect().IsEmpty())
     aCtx->FillRect(GetHookRect() + aOffset);
@@ -644,8 +707,7 @@ nsCaret::DrawAtPositionWithHint(nsIDOMNode*             aNode,
     }
 
     // Only update the caret's rect when we're not currently drawn.
-    rv = UpdateCaretRects(theFrame, theFrameOffset);
-    if (NS_FAILED(rv))
+    if (!UpdateCaretRects(theFrame, theFrameOffset))
       return PR_FALSE;
   }
 
@@ -655,71 +717,6 @@ nsCaret::DrawAtPositionWithHint(nsIDOMNode*             aNode,
   return PR_TRUE;
 }
 
-/**
- * Find the first frame in an in-order traversal of the frame subtree rooted
- * at aFrame which is either a text frame logically at the end of a line,
- * or which is aStopAtFrame. Return null if no such frame is found. We don't
- * descend into the children of non-eLineParticipant frames.
- */
-static nsIFrame*
-CheckForTrailingTextFrameRecursive(nsIFrame* aFrame, nsIFrame* aStopAtFrame)
-{
-  if (aFrame == aStopAtFrame ||
-      ((aFrame->GetType() == nsGkAtoms::textFrame &&
-       (static_cast<nsTextFrame*>(aFrame))->IsAtEndOfLine())))
-    return aFrame;
-  if (!aFrame->IsFrameOfType(nsIFrame::eLineParticipant))
-    return nsnull;
-
-  for (nsIFrame* f = aFrame->GetFirstChild(nsnull); f; f = f->GetNextSibling())
-  {
-    nsIFrame* r = CheckForTrailingTextFrameRecursive(f, aStopAtFrame);
-    if (r)
-      return r;
-  }
-  return nsnull;
-}
-
-static nsLineBox*
-FindContainingLine(nsIFrame* aFrame)
-{
-  while (aFrame && aFrame->IsFrameOfType(nsIFrame::eLineParticipant))
-  {
-    nsIFrame* parent = aFrame->GetParent();
-    nsBlockFrame* blockParent = nsLayoutUtils::GetAsBlock(parent);
-    if (blockParent)
-    {
-      PRBool isValid;
-      nsBlockInFlowLineIterator iter(blockParent, aFrame, &isValid);
-      return isValid ? iter.GetLine().get() : nsnull;
-    }
-    aFrame = parent;
-  }
-  return nsnull;
-}
-
-static void
-AdjustCaretFrameForLineEnd(nsIFrame** aFrame, PRInt32* aOffset)
-{
-  nsLineBox* line = FindContainingLine(*aFrame);
-  if (!line)
-    return;
-  PRInt32 count = line->GetChildCount();
-  for (nsIFrame* f = line->mFirstChild; count > 0; --count, f = f->GetNextSibling())
-  {
-    nsIFrame* r = CheckForTrailingTextFrameRecursive(f, *aFrame);
-    if (r == *aFrame)
-      return;
-    if (r)
-    {
-      *aFrame = r;
-      NS_ASSERTION(r->GetType() == nsGkAtoms::textFrame, "Expected text frame");
-      *aOffset = (static_cast<nsTextFrame*>(r))->GetContentEnd();
-      return;
-    }
-  }
-}
-
 nsresult 
 nsCaret::GetCaretFrameForNodeOffset(nsIContent*             aContentNode,
                                     PRInt32                 aOffset,
@@ -734,6 +731,10 @@ nsCaret::GetCaretFrameForNodeOffset(nsIContent*             aContentNode,
   if (!presShell)
     return NS_ERROR_FAILURE;
 
+  if (!aContentNode || !aContentNode->IsInDoc() ||
+      presShell->GetDocument() != aContentNode->GetCurrentDoc())
+    return NS_ERROR_FAILURE;
+
   nsCOMPtr<nsFrameSelection> frameSelection = GetFrameSelection();
   if (!frameSelection)
     return NS_ERROR_FAILURE;
@@ -789,8 +790,8 @@ nsCaret::GetCaretFrameForNodeOffset(nsIContent*             aContentNode,
 
         if ((levelBefore != levelAfter) || (aBidiLevel != levelBefore))
         {
-          aBidiLevel = PR_MAX(aBidiLevel, PR_MIN(levelBefore, levelAfter));                                  // rule c3
-          aBidiLevel = PR_MIN(aBidiLevel, PR_MAX(levelBefore, levelAfter));                                  // rule c4
+          aBidiLevel = NS_MAX(aBidiLevel, NS_MIN(levelBefore, levelAfter));                                  // rule c3
+          aBidiLevel = NS_MIN(aBidiLevel, NS_MAX(levelBefore, levelAfter));                                  // rule c4
           if (aBidiLevel == levelBefore                                                                      // rule c1
               || (aBidiLevel > levelBefore && aBidiLevel < levelAfter && !((aBidiLevel ^ levelBefore) & 1))    // rule c5
               || (aBidiLevel < levelBefore && aBidiLevel > levelAfter && !((aBidiLevel ^ levelBefore) & 1)))  // rule c9
@@ -886,86 +887,27 @@ nsCaret::GetCaretFrameForNodeOffset(nsIContent*             aContentNode,
       }
     }
   }
+
+  NS_ASSERTION(!theFrame || theFrame->PresContext()->PresShell() == presShell,
+               "caret frame is in wrong document");
   *aReturnFrame = theFrame;
   *aReturnOffset = theFrameOffset;
   return NS_OK;
 }
 
-
-//-----------------------------------------------------------------------------
-void nsCaret::GetViewForRendering(nsIFrame *caretFrame,
-                                  EViewCoordinates coordType,
-                                  nsPoint &viewOffset,
-                                  nsIView **outRenderingView,
-                                  nsIView **outRelativeView)
+nsresult nsCaret::CheckCaretDrawingState()
 {
-  if (!caretFrame || !outRenderingView)
-    return;
-
-  *outRenderingView = nsnull;
-  if (outRelativeView)
-    *outRelativeView = nsnull;
-  
-  NS_ASSERTION(caretFrame, "Should have a frame here");
- 
-  viewOffset.x = 0;
-  viewOffset.y = 0;
-  
-  nsPoint withinViewOffset(0, 0);
-  // get the offset of this frame from its parent view (walks up frame hierarchy)
-  nsIView* theView = nsnull;
-  caretFrame->GetOffsetFromView(withinViewOffset, &theView);
-  if (!theView)
-      return;
-
-  if (outRelativeView && coordType == eClosestViewCoordinates)
-    *outRelativeView = theView;
-
-  // Note: views are not refcounted.
-  nsIView* returnView = nsIView::GetViewFor(theView->GetNearestWidget(nsnull));
-  
-  // This gets uses the first view with a widget
-  if (coordType == eRenderingViewCoordinates) {
-    if (returnView) {
-      // Now adjust the view offset for this view.
-      withinViewOffset += theView->GetOffsetTo(returnView);
-      
-      // Account for the view's origin not lining up with the widget's
-      // (bug 190290)
-      withinViewOffset += returnView->GetPosition() -
-                          returnView->GetBounds().TopLeft();
-      viewOffset = withinViewOffset;
-
-      if (outRelativeView)
-        *outRelativeView = returnView;
-    }
+  if (mDrawn) {
+    // The caret is drawn; if it shouldn't be, erase it.
+    if (!mVisible || !MustDrawCaret(PR_TRUE))
+      EraseCaret();
   }
-  else {
-    // window-relative coordinates. Done for us by the view.
-    withinViewOffset += theView->GetOffsetTo(nsnull);
-    viewOffset = withinViewOffset;
-
-    // Get the relative view for top level window coordinates
-    if (outRelativeView && coordType == eTopLevelWindowCoordinates) {
-      nsCOMPtr<nsIPresShell> presShell = do_QueryReferent(mPresShell);
-      if (presShell) {
-        nsIViewManager* vm =
-          presShell->GetPresContext()->RootPresContext()->PresShell()->GetViewManager();
-        if (vm) {
-          vm->GetRootView(*outRelativeView);
-        }
-      }
-    }
+  else
+  {
+    // The caret is not drawn; if it should be, draw it.
+    if (mPendingDraw && (mVisible && MustDrawCaret(PR_TRUE)))
+      DrawCaret(PR_TRUE);
   }
-
-  *outRenderingView = returnView;
-}
-
-nsresult nsCaret::CheckCaretDrawingState() 
-{
-  // If the caret's drawn when it shouldn't be, erase it.
-  if (mDrawn && (!mVisible || !MustDrawCaret(PR_TRUE)))
-    EraseCaret();
   return NS_OK;
 }
 
@@ -983,22 +925,14 @@ nsresult nsCaret::CheckCaretDrawingState()
 ----------------------------------------------------------------------------- */
 PRBool nsCaret::MustDrawCaret(PRBool aIgnoreDrawnState)
 {
-  nsCOMPtr<nsIPresShell> presShell = do_QueryReferent(mPresShell);
-  if (presShell) {
-    PRBool isPaintingSuppressed;
-    presShell->IsPaintingSuppressed(&isPaintingSuppressed);
-    if (isPaintingSuppressed)
-      return PR_FALSE;
-  }
-
   if (!aIgnoreDrawnState && mDrawn)
     return PR_TRUE;
 
   nsCOMPtr<nsISelection> domSelection = do_QueryReferent(mDomSelectionWeak);
   if (!domSelection)
     return PR_FALSE;
-  PRBool isCollapsed;
 
+  PRBool isCollapsed;
   if (NS_FAILED(domSelection->GetIsCollapsed(&isCollapsed)))
     return PR_FALSE;
 
@@ -1059,18 +993,27 @@ PRBool nsCaret::IsMenuPopupHidingCaret()
   return PR_FALSE;
 }
 
-/*-----------------------------------------------------------------------------
-
-  DrawCaret
-    
------------------------------------------------------------------------------ */
-
 void nsCaret::DrawCaret(PRBool aInvalidate)
 {
-  // do we need to draw the caret at all?
+  // Do we need to draw the caret at all?
   if (!MustDrawCaret(PR_FALSE))
     return;
   
+  // Can we draw the caret now?
+  nsCOMPtr<nsIPresShell> presShell = do_QueryReferent(mPresShell);
+  NS_ENSURE_TRUE(presShell, /**/);
+  {
+    if (presShell->IsPaintingSuppressed())
+    {
+      if (!mDrawn)
+        mPendingDraw = PR_TRUE;
+
+      // PresShell::UnsuppressAndInvalidate() will call CheckCaretDrawingState()
+      // to get us drawn.
+      return;
+    }
+  }
+
   nsCOMPtr<nsIDOMNode> node;
   PRInt32 offset;
   nsFrameSelection::HINT hint;
@@ -1102,7 +1045,9 @@ void nsCaret::DrawCaret(PRBool aInvalidate)
     nsCOMPtr<nsFrameSelection> frameSelection = GetFrameSelection();
     if (!frameSelection)
       return;
+
     bidiLevel = frameSelection->GetCaretBidiLevel();
+    mPendingDraw = PR_FALSE;
   }
   else
   {
@@ -1111,7 +1056,8 @@ void nsCaret::DrawCaret(PRBool aInvalidate)
       mDrawn = PR_FALSE;
       return;
     }
-    if (!mLastContent->IsInDoc())
+    if (!mLastContent->IsInDoc() ||
+        presShell->GetDocument() != mLastContent->GetCurrentDoc())
     {
       mLastContent = nsnull;
       mDrawn = PR_FALSE;
@@ -1127,99 +1073,16 @@ void nsCaret::DrawCaret(PRBool aInvalidate)
   ToggleDrawnStatus();
 }
 
-static PRBool
-FramesOnSameLineHaveZeroHeight(nsIFrame* aFrame)
-{
-  nsLineBox* line = FindContainingLine(aFrame);
-  if (!line)
-    return aFrame->GetRect().height == 0;
-  PRInt32 count = line->GetChildCount();
-  for (nsIFrame* f = line->mFirstChild; count > 0; --count, f = f->GetNextSibling())
-  {
-   if (f->GetRect().height != 0)
-     return PR_FALSE;
-  }
-  return PR_TRUE;
-}
-
-nsresult nsCaret::UpdateCaretRects(nsIFrame* aFrame, PRInt32 aFrameOffset)
+PRBool
+nsCaret::UpdateCaretRects(nsIFrame* aFrame, PRInt32 aFrameOffset)
 {
   NS_ASSERTION(aFrame, "Should have a frame here");
 
-  nsRect frameRect = aFrame->GetRect();
-  frameRect.x = 0;
-  frameRect.y = 0;
-
-  nsCOMPtr<nsIPresShell> presShell = do_QueryReferent(mPresShell);
-  if (!presShell) return NS_ERROR_FAILURE;
-
-  // If we got a zero-height frame we should figure out a height. We have to do
-  // this after we've got an RC.
-  if (frameRect.height == 0)
-  {
-    nsCOMPtr<nsIFontMetrics> fm;
-    nsLayoutUtils::GetFontMetricsForFrame(aFrame, getter_AddRefs(fm));
-
-    if (fm)
-    {
-      nscoord ascent, descent;
-      fm->GetMaxAscent(ascent);
-      fm->GetMaxDescent(descent);
-      frameRect.height = ascent + descent;
-
-      // Place the caret on the baseline for inline frames, except when there is
-      // a frame on the line with non-zero height.  XXXmats why the exception? --
-      // I don't know but it seems to be necessary, see bug 503531.
-      if (aFrame->GetStyleDisplay()->IsInlineOutside() &&
-          !FramesOnSameLineHaveZeroHeight(aFrame))
-        frameRect.y -= ascent;
-    }
-  }
-
-  mCaretRect = frameRect;
-  nsCOMPtr<nsISelection> domSelection = do_QueryReferent(mDomSelectionWeak);
-  nsCOMPtr<nsISelectionPrivate> privateSelection = do_QueryInterface(domSelection);
-
-  nsPoint framePos;
-
-  // if cache in selection is available, apply it, else refresh it
-  nsresult rv = privateSelection->GetCachedFrameOffset(aFrame, aFrameOffset,
-                                                       framePos);
-  if (NS_FAILED(rv))
-  {
-    mCaretRect.Empty();
-    return rv;
-  }
-
-  mCaretRect += framePos;
-  Metrics metrics = ComputeMetrics(aFrame, aFrameOffset, mCaretRect.height);
-  mCaretRect.width = metrics.mCaretWidth;
-
-  // Clamp our position to be within our scroll frame. If we don't, then it
-  // clips us, and we don't appear at all. See bug 335560.
-  nsIFrame *scrollFrame =
-    nsLayoutUtils::GetClosestFrameOfType(aFrame, nsGkAtoms::scrollFrame);
-  if (scrollFrame)
-  {
-    // First, use the scrollFrame to get at the scrollable view that we're in.
-    nsIScrollableFrame *scrollable = do_QueryFrame(scrollFrame);
-    nsIScrollableView *scrollView = scrollable->GetScrollableView();
-    nsIView *view;
-    scrollView->GetScrolledView(view);
-
-    // Compute the caret's coordinates in the enclosing view's coordinate
-    // space. To do so, we need to correct for both the original frame's
-    // offset from the scrollframe, and the scrollable view's offset from the
-    // scrolled frame's view.
-    nsPoint toScroll = aFrame->GetOffsetTo(scrollFrame) -
-      view->GetOffsetTo(scrollFrame->GetView());
-    nsRect caretInScroll = mCaretRect + toScroll;
-
-    // Now see if thet caret extends beyond the view's bounds. If it does,
-    // then snap it back, put it as close to the edge as it can.
-    nscoord overflow = caretInScroll.XMost() - view->GetBounds().width;
-    if (overflow > 0)
-      mCaretRect.x -= overflow;
+  nscoord bidiIndicatorSize;
+  nsresult rv =
+    GetGeometryForFrame(aFrame, aFrameOffset, &mCaretRect, &bidiIndicatorSize);
+  if (NS_FAILED(rv)) {
+    return PR_FALSE;
   }
 
   // on RTL frames the right edge of mCaretRect must be equal to framePos
@@ -1227,27 +1090,18 @@ nsresult nsCaret::UpdateCaretRects(nsIFrame* aFrame, PRInt32 aFrameOffset)
   if (NS_STYLE_DIRECTION_RTL == vis->mDirection)
     mCaretRect.x -= mCaretRect.width;
 
-  return UpdateHookRect(presShell->GetPresContext(), metrics);
-}
-
-nsresult nsCaret::UpdateHookRect(nsPresContext* aPresContext,
-                                 const Metrics& aMetrics)
-{
+#ifdef IBMBIDI
   mHookRect.Empty();
 
-#ifdef IBMBIDI
   // Simon -- make a hook to draw to the left or right of the caret to show keyboard language direction
-  PRBool isCaretRTL=PR_FALSE;
+  PRBool isCaretRTL = PR_FALSE;
   nsIBidiKeyboard* bidiKeyboard = nsContentUtils::GetBidiKeyboard();
-  if (!bidiKeyboard || NS_FAILED(bidiKeyboard->IsLangRTL(&isCaretRTL)))
-    // if bidiKeyboard->IsLangRTL() failed, there is no way to tell the
-    // keyboard direction, or the user has no right-to-left keyboard
-    // installed, so we  never draw the hook.
-    return NS_OK;
-  if (mBidiUI)
-  {
-    if (isCaretRTL != mKeyboardRTL)
-    {
+  // if bidiKeyboard->IsLangRTL() fails, there is no way to tell the
+  // keyboard direction, or the user has no right-to-left keyboard
+  // installed, so we never draw the hook.
+  if (bidiKeyboard && NS_SUCCEEDED(bidiKeyboard->IsLangRTL(&isCaretRTL)) &&
+      mBidiUI) {
+    if (isCaretRTL != mKeyboardRTL) {
       /* if the caret bidi level and the keyboard language direction are not in
        * synch, the keyboard language must have been changed by the
        * user, and if the caret is in a boundary condition (between left-to-right and
@@ -1258,18 +1112,13 @@ nsresult nsCaret::UpdateHookRect(nsPresContext* aPresContext,
        */ 
       mKeyboardRTL = isCaretRTL;
       nsCOMPtr<nsISelection> domSelection = do_QueryReferent(mDomSelectionWeak);
-      if (domSelection)
-      {
-        if (NS_SUCCEEDED(domSelection->SelectionLanguageChange(mKeyboardRTL)))
-        {
-          return NS_ERROR_FAILURE;
-        }
-      }
+      if (!domSelection ||
+          NS_SUCCEEDED(domSelection->SelectionLanguageChange(mKeyboardRTL)))
+        return PR_FALSE;
     }
     // If keyboard language is RTL, draw the hook on the left; if LTR, to the right
     // The height of the hook rectangle is the same as the width of the caret
     // rectangle.
-    nscoord bidiIndicatorSize = aMetrics.mBidiIndicatorSize;
     mHookRect.SetRect(mCaretRect.x + ((isCaretRTL) ?
                       bidiIndicatorSize * -1 :
                       mCaretRect.width),
@@ -1278,8 +1127,7 @@ nsresult nsCaret::UpdateHookRect(nsPresContext* aPresContext,
                       mCaretRect.width);
   }
 #endif //IBMBIDI
-
-  return NS_OK;
+  return PR_TRUE;
 }
 
 // static
