@@ -54,49 +54,73 @@
 class NS_STACK_CLASS ChildIterator
 {
 protected:
-  // We could just use a mNodes for the case when we're not dealing with XBL
-  // anon content too, but in practice that would basically replace our current
-  // branch on mNodes with an extra virtual function call in the common case.
-
   nsIContent* mContent;
-  PRUint32 mIndex;
+  // If mNodes is non-null (so XBLInvolved() is true), mIndex is the
+  // index into mNodes for our current position.  Otherwise, mChild is
+  // our current child (which might be null if we're done iterating).
+  union {
+    PRUint32 mIndex;
+    nsIContent* mChild;
+  };
   nsINodeList* mNodes;
 
 public:
   ChildIterator()
-    : mContent(nsnull), mIndex(0), mNodes(nsnull) {}
+    : mContent(nsnull), mChild(0), mNodes(nsnull) {}
 
   ChildIterator(const ChildIterator& aOther)
     : mContent(aOther.mContent),
-      mIndex(aOther.mIndex),
-      mNodes(aOther.mNodes) {}
+      mNodes(aOther.mNodes) {
+    if (XBLInvolved()) {
+      mIndex = aOther.mIndex;
+    } else {
+      mChild = aOther.mChild;
+    }
+  }
 
   ChildIterator& operator=(const ChildIterator& aOther) {
     mContent = aOther.mContent;
-    mIndex = aOther.mIndex;
     mNodes = aOther.mNodes;
+    if (XBLInvolved()) {
+      mIndex = aOther.mIndex;
+    } else {
+      mChild = aOther.mChild;
+    }
     return *this;
   }
 
   ChildIterator& operator++() {
-    ++mIndex;
+    if (XBLInvolved()) {
+      ++mIndex;
+    } else {
+      NS_ASSERTION(mChild, "Walking off end of list?");
+      mChild = mChild->GetNextSibling();
+    }
+
     return *this;
   }
 
   ChildIterator operator++(int) {
     ChildIterator result(*this);
-    ++mIndex;
+    ++(*this);
     return result;
   }
 
   ChildIterator& operator--() {
-    --mIndex;
+    if (XBLInvolved()) {
+      --mIndex;
+    } else if (mChild) {
+      mChild = mChild->GetPreviousSibling();
+      NS_ASSERTION(mChild, "Walking off beginning of list");
+    } else {
+      mChild = mContent->GetLastChild();
+    }
     return *this;
   }
 
   ChildIterator operator--(int) {
     ChildIterator result(*this);
-    --mIndex;
+    --(*this);
     return result;
   }
 
@@ -105,59 +129,46 @@ public:
       return mNodes->GetNodeAt(mIndex);
     }
 
-    return mContent->GetChildAt(mIndex);
+    return mChild;
   }
 
   nsIContent* operator*() const { return get(); }
 
   PRBool operator==(const ChildIterator& aOther) const {
-    return mContent == aOther.mContent && mIndex == aOther.mIndex;
+    if (XBLInvolved()) {
+      return mContent == aOther.mContent && mIndex == aOther.mIndex;
+    }
+
+    return mContent == aOther.mContent && mChild == aOther.mChild;
   }
 
   PRBool operator!=(const ChildIterator& aOther) const {
     return !aOther.operator==(*this);
   }
 
-  PRUint32 position() {
-    return mIndex;
-  }
-
-  void seek(PRUint32 aIndex) {
-    // Make sure that aIndex is reasonable.  This should be |#ifdef
-    // DEBUG|, but we need these numbers for the temporary workaround
-    // for bug 133219.
-    PRUint32 l = length();
-
-    NS_ASSERTION(PRInt32(aIndex) >= 0 && aIndex <= l, "out of bounds");
-
-    // Temporary workaround for bug 133219.
-    if (aIndex > l)
-      aIndex = l;
-
-    mIndex = aIndex;
-  }
-
   void seek(nsIContent* aContent) {
-    PRInt32 index;
     if (XBLInvolved()) {
-      index = mNodes->IndexOf(aContent);
+      PRInt32 index = mNodes->IndexOf(aContent);
+      // XXXbz I wish we could assert that index != -1, but it seems to not be
+      // the case in some XBL cases with filtered insertion points and no
+      // default insertion point.  I will now claim that XBL's management of
+      // its insertion points is broken in those cases, since it's returning an
+      // insertion parent for a node that doesn't actually have the node in its
+      // child list according to ChildIterator.  See bug 474324.
+      if (index != -1) {
+        mIndex = index;
+      } else {
+        // If aContent isn't going to get hit by this iterator, just seek to the
+        // end of the list for lack of anything better to do.
+        mIndex = length();
+      }
+    } else if (aContent->GetParent() == mContent) {
+      mChild = aContent;
     } else {
-      index = mContent->IndexOf(aContent);
-    }
-    // XXXbz I wish we could assert that index != -1, but I think that's not
-    // necessarily the case when called from ContentInserted if first-letter
-    // frames are about.  It also seems to not be the case in some XBL cases
-    // with filtered insertion points and no default insertion point.  I will
-    // now claim that XBL's management of its insertion points is broken in
-    // those cases, since it's returning an insertion parent for a node that
-    // doesn't actually have the node in its child list according to
-    // ChildIterator.  See bug 474324.
-    if (index != -1) {
-      mIndex = index;
-    } else {
-      // If aContent isn't going to get hit by this iterator, just seek to the
-      // end of the list for lack of anything better to do.
-      mIndex = length();
+      // XXXbz I wish we could assert this doesn't happen, but I think that's
+      // not necessarily the case when called from ContentInserted if
+      // first-letter frames are about.
+      mChild = nsnull;
     }
   }
 
@@ -174,12 +185,9 @@ public:
 
 private:
   PRUint32 length() {
+    NS_PRECONDITION(XBLInvolved(), "Don't call me");
     PRUint32 l;
-    if (XBLInvolved()) {
-      mNodes->GetLength(&l);
-    } else {
-      l = mContent->GetChildCount();
-    }
+    mNodes->GetLength(&l);
     return l;
   }
 };
