@@ -50,6 +50,8 @@
 
 #include "nsRuleNode.h"
 #include "nsStyleContext.h"
+#include "prlog.h"
+#include "nsStyleAnimation.h"
 
 #ifdef DEBUG
 // #define NOISY_DEBUG
@@ -60,6 +62,7 @@
 
 nsStyleContext::nsStyleContext(nsStyleContext* aParent,
                                nsIAtom* aPseudoTag,
+                               nsCSSPseudoElements::Type aPseudoType,
                                nsRuleNode* aRuleNode,
                                nsPresContext* aPresContext)
   : mParent(aParent),
@@ -67,9 +70,14 @@ nsStyleContext::nsStyleContext(nsStyleContext* aParent,
     mEmptyChild(nsnull),
     mPseudoTag(aPseudoTag),
     mRuleNode(aRuleNode),
-    mBits(0),
+    mAllocations(nsnull),
+    mCachedResetData(nsnull),
+    mBits(((PRUint32)aPseudoType) << NS_STYLE_CONTEXT_TYPE_SHIFT),
     mRefCnt(0)
 {
+  PR_STATIC_ASSERT((PR_UINT32_MAX >> NS_STYLE_CONTEXT_TYPE_SHIFT) >
+                   nsCSSPseudoElements::ePseudo_MAX);
+
   mNextSibling = this;
   mPrevSibling = this;
   if (mParent) {
@@ -91,6 +99,8 @@ nsStyleContext::nsStyleContext(nsStyleContext* aParent,
   NS_ASSERTION(NS_STYLE_INHERIT_MASK & NS_STYLE_INHERIT_BIT(LastItem),
                "NS_STYLE_INHERIT_MASK must be bigger, and other bits shifted");
   #undef eStyleStruct_LastItem
+
+  mRuleNode->AddRef();
 }
 
 nsStyleContext::~nsStyleContext()
@@ -99,6 +109,8 @@ nsStyleContext::~nsStyleContext()
 
   nsPresContext *presContext = mRuleNode->GetPresContext();
 
+  mRuleNode->Release();
+
   presContext->PresShell()->StyleSet()->
     NotifyStyleContextDestroyed(presContext, this);
 
@@ -108,9 +120,12 @@ nsStyleContext::~nsStyleContext()
   }
 
   // Free up our data structs.
-  if (mCachedStyleData.mResetData || mCachedStyleData.mInheritedData) {
-    mCachedStyleData.Destroy(mBits, presContext);
+  mCachedInheritedData.DestroyStructs(mBits, presContext);
+  if (mCachedResetData) {
+    mCachedResetData->Destroy(mBits, presContext);
   }
+
+  FreeAllocations(presContext);
 }
 
 void nsStyleContext::AddChild(nsStyleContext* aChild)
@@ -156,8 +171,12 @@ void nsStyleContext::RemoveChild(nsStyleContext* aChild)
 
 already_AddRefed<nsStyleContext>
 nsStyleContext::FindChildWithRules(const nsIAtom* aPseudoTag, 
-                                   nsRuleNode* aRuleNode)
+                                   nsRuleNode* aRuleNode,
+                                   nsRuleNode* aRulesIfVisited,
+                                   PRBool aRelevantLinkVisited)
 {
+  NS_ABORT_IF_FALSE(aRulesIfVisited || !aRelevantLinkVisited,
+    "aRelevantLinkVisited should only be set when we have a separate style");
   PRUint32 threshold = 10; // The # of siblings we're willing to examine
                            // before just giving this whole thing up.
 
@@ -167,9 +186,21 @@ nsStyleContext::FindChildWithRules(const nsIAtom* aPseudoTag,
   if (list) {
     nsStyleContext *child = list;
     do {
-      if (child->mRuleNode == aRuleNode && child->mPseudoTag == aPseudoTag) {
-        result = child;
-        break;
+      if (child->mRuleNode == aRuleNode &&
+          child->mPseudoTag == aPseudoTag &&
+          !child->IsStyleIfVisited() &&
+          child->RelevantLinkVisited() == aRelevantLinkVisited) {
+        PRBool match = PR_FALSE;
+        if (aRulesIfVisited) {
+          match = child->GetStyleIfVisited() &&
+                  child->GetStyleIfVisited()->mRuleNode == aRulesIfVisited;
+        } else {
+          match = !child->GetStyleIfVisited();
+        }
+        if (match) {
+          result = child;
+          break;
+        }
       }
       child = child->mNextSibling;
       threshold--;
@@ -192,35 +223,34 @@ nsStyleContext::FindChildWithRules(const nsIAtom* aPseudoTag,
   return result;
 }
 
+const void* nsStyleContext::GetCachedStyleData(nsStyleStructID aSID)
+{
+  const void* cachedData;
+  PRBool isReset = nsCachedStyleData::IsReset(aSID);
+  if (isReset) {
+    if (mCachedResetData) {
+      char* slot = reinterpret_cast<char*>(mCachedResetData) +
+                   nsCachedStyleData::gInfo[aSID].mInheritResetOffset;
+      cachedData = *reinterpret_cast<void**>(slot);
+    } else {
+      cachedData = nsnull;
+    }
+  } else {
+    char* slot = reinterpret_cast<char*>(&mCachedInheritedData) +
+                 nsCachedStyleData::gInfo[aSID].mInheritResetOffset;
+    cachedData = *reinterpret_cast<void**>(slot);
+  }
+  return cachedData;
+}
+
 const void* nsStyleContext::GetStyleData(nsStyleStructID aSID)
 {
-  const void* cachedData = mCachedStyleData.GetStyleData(aSID); 
+  const void* cachedData = GetCachedStyleData(aSID);
   if (cachedData)
     return cachedData; // We have computed data stored on this node in the context tree.
   return mRuleNode->GetStyleData(aSID, this, PR_TRUE); // Our rule node will take care of it for us.
 }
 
-#define STYLE_STRUCT(name_, checkdata_cb_, ctor_args_)                      \
-  const nsStyle##name_ * nsStyleContext::GetStyle##name_ ()                 \
-  {                                                                         \
-    const nsStyle##name_ * cachedData = mCachedStyleData.GetStyle##name_(); \
-    if (cachedData)                                                         \
-      return cachedData; /* We have computed data stored on this node */    \
-                         /* in the context tree. */                         \
-    /* Else our rule node will take care of it for us. */                   \
-    return mRuleNode->GetStyle##name_(this, PR_TRUE);                       \
-  }
-#include "nsStyleStructList.h"
-#undef STYLE_STRUCT
-
-const void* nsStyleContext::PeekStyleData(nsStyleStructID aSID)
-{
-  const void* cachedData = mCachedStyleData.GetStyleData(aSID); 
-  if (cachedData)
-    return cachedData; // We have computed data stored on this node in the context tree.
-  return mRuleNode->GetStyleData(aSID, this, PR_FALSE); // Our rule node will take care of it for us.
-}
-
 // This is an evil evil function, since it forces you to alloc your own separate copy of
 // style data!  Do not use this function unless you absolutely have to!  You should avoid
 // this at all costs! -dwh
@@ -235,7 +265,7 @@ nsStyleContext::GetUniqueStyleData(const nsStyleStructID& aSID)
   const void *current = GetStyleData(aSID);
   if (!mChild && !mEmptyChild &&
       !(mBits & nsCachedStyleData::GetBitForSID(aSID)) &&
-      mCachedStyleData.GetStyleData(aSID))
+      GetCachedStyleData(aSID))
     return const_cast<void*>(current);
 
   void* result;
@@ -256,7 +286,7 @@ nsStyleContext::GetUniqueStyleData(const nsStyleStructID& aSID)
 #undef UNIQUE_CASE
 
   default:
-    NS_ERROR("Struct type not supported.  Please find another way to do this if you can!\n");
+    NS_ERROR("Struct type not supported.  Please find another way to do this if you can!");
     return nsnull;
   }
 
@@ -284,25 +314,18 @@ nsStyleContext::SetStyle(nsStyleStructID aSID, void* aStruct)
   // See the comments there (in nsRuleNode.h) for more details about
   // what this is doing and why.
 
-  const nsCachedStyleData::StyleStructInfo& info =
-      nsCachedStyleData::gInfo[aSID];
-  char* resetOrInheritSlot = reinterpret_cast<char*>(&mCachedStyleData) +
-                             info.mCachedStyleDataOffset;
-  char* resetOrInherit = reinterpret_cast<char*>
-                                         (*reinterpret_cast<void**>(resetOrInheritSlot));
-  if (!resetOrInherit) {
-    nsPresContext *presContext = mRuleNode->GetPresContext();
-    if (mCachedStyleData.IsReset(aSID)) {
-      mCachedStyleData.mResetData = new (presContext) nsResetStyleData;
-      resetOrInherit = reinterpret_cast<char*>(mCachedStyleData.mResetData);
-    } else {
-      mCachedStyleData.mInheritedData =
-          new (presContext) nsInheritedStyleData;
-      resetOrInherit =
-          reinterpret_cast<char*>(mCachedStyleData.mInheritedData);
+  char* dataSlot;
+  if (nsCachedStyleData::IsReset(aSID)) {
+    if (!mCachedResetData) {
+      mCachedResetData = new (mRuleNode->GetPresContext()) nsResetStyleData;
+      // XXXbz And if that fails?
     }
+    dataSlot = reinterpret_cast<char*>(mCachedResetData) +
+               nsCachedStyleData::gInfo[aSID].mInheritResetOffset;
+  } else {
+    dataSlot = reinterpret_cast<char*>(&mCachedInheritedData) +
+               nsCachedStyleData::gInfo[aSID].mInheritResetOffset;
   }
-  char* dataSlot = resetOrInherit + info.mInheritResetOffset;
   NS_ASSERTION(!*reinterpret_cast<void**>(dataSlot) ||
                (mBits & nsCachedStyleData::GetBitForSID(aSID)),
                "Going to leak style data");
@@ -389,16 +412,11 @@ nsStyleContext::CalcStyleDifference(nsStyleContext* aOther)
   // our position in the rule node tree is also the same.
   PRBool compare = mRuleNode != aOther->mRuleNode;
 
-  nsChangeHint maxHint = nsChangeHint(NS_STYLE_HINT_FRAMECHANGE |
-      nsChangeHint_UpdateCursor);
-  
 #define DO_STRUCT_DIFFERENCE(struct_)                                         \
   PR_BEGIN_MACRO                                                              \
     NS_ASSERTION(NS_IsHintSubset(nsStyle##struct_::MaxDifference(), maxHint), \
                  "Struct placed in the wrong maxHint section");               \
-    const nsStyle##struct_* this##struct_ =                                   \
-        static_cast<const nsStyle##struct_*>(                                 \
-                       PeekStyleData(eStyleStruct_##struct_));                \
+    const nsStyle##struct_* this##struct_ = PeekStyle##struct_();             \
     if (this##struct_) {                                                      \
       const nsStyle##struct_* other##struct_ = aOther->GetStyle##struct_();   \
       if ((compare || nsStyle##struct_::ForceCompare()) &&                    \
@@ -417,7 +435,12 @@ nsStyleContext::CalcStyleDifference(nsStyleContext* aOther)
   // causing the maximal difference, a FRAMECHANGE.
   // FRAMECHANGE Structs: Display, XUL, Content, UserInterface,
   // Visibility, Outline, TableBorder, Table, Text, UIReset, Quotes
+  nsChangeHint maxHint = nsChangeHint(NS_STYLE_HINT_FRAMECHANGE |
+      nsChangeHint_UpdateTransformLayer | nsChangeHint_UpdateOpacityLayer);
   DO_STRUCT_DIFFERENCE(Display);
+
+  maxHint = nsChangeHint(NS_STYLE_HINT_FRAMECHANGE |
+      nsChangeHint_UpdateCursor);
   DO_STRUCT_DIFFERENCE(XUL);
   DO_STRUCT_DIFFERENCE(Column);
   DO_STRUCT_DIFFERENCE(Content);
@@ -433,11 +456,9 @@ nsStyleContext::CalcStyleDifference(nsStyleContext* aOther)
   // a framechange here and a reflow should be sufficient.  See bug 35768.
   DO_STRUCT_DIFFERENCE(Quotes);
 
-#ifdef MOZ_SVG
   maxHint = nsChangeHint(NS_STYLE_HINT_REFLOW | nsChangeHint_UpdateEffects);
   DO_STRUCT_DIFFERENCE(SVGReset);
   DO_STRUCT_DIFFERENCE(SVG);
-#endif
 
   // At this point, we know that the worst kind of damage we could do is
   // a reflow.
@@ -453,17 +474,120 @@ nsStyleContext::CalcStyleDifference(nsStyleContext* aOther)
   DO_STRUCT_DIFFERENCE(Position);
   DO_STRUCT_DIFFERENCE(TextReset);
 
-  // At this point, we know that the worst kind of damage we could do is
-  // a re-render (i.e., a VISUAL change).
-  maxHint = NS_STYLE_HINT_VISUAL;
-
-  // The following structs cause (as their maximal difference) a
-  // re-render to occur.  VISUAL Structs: Color, Background
-  DO_STRUCT_DIFFERENCE(Color);
+  // Most backgrounds only require a re-render (i.e., a VISUAL change), but
+  // backgrounds using -moz-element need to reset SVG effects, too.
+  maxHint = nsChangeHint(NS_STYLE_HINT_VISUAL | nsChangeHint_UpdateEffects);
   DO_STRUCT_DIFFERENCE(Background);
 
+  // Color only needs a repaint.
+  maxHint = NS_STYLE_HINT_VISUAL;
+  DO_STRUCT_DIFFERENCE(Color);
+
 #undef DO_STRUCT_DIFFERENCE
 
+  // Note that we do not check whether this->RelevantLinkVisited() !=
+  // aOther->RelevantLinkVisited(); we don't need to since
+  // nsCSSFrameConstructor::DoContentStateChanged always adds
+  // nsChangeHint_RepaintFrame for NS_EVENT_STATE_VISITED changes (and
+  // needs to, since HasStateDependentStyle probably doesn't work right
+  // for NS_EVENT_STATE_VISITED).  Hopefully this doesn't actually
+  // expose whether links are visited to performance tests since all
+  // link coloring happens asynchronously at a time when it's hard for
+  // the page to measure.
+  // However, we do need to compute the larger of the changes that can
+  // happen depending on whether the link is visited or unvisited, since
+  // doing only the one that's currently appropriate would expose which
+  // links are in history to easy performance measurement.  Therefore,
+  // here, we add nsChangeHint_RepaintFrame hints (the maximum for
+  // things that can depend on :visited) for the properties on which we
+  // call GetVisitedDependentColor.
+  nsStyleContext *thisVis = GetStyleIfVisited(),
+                *otherVis = aOther->GetStyleIfVisited();
+  if (!thisVis != !otherVis) {
+    // One style context has a style-if-visited and the other doesn't.
+    // Presume a difference.
+    NS_UpdateHint(hint, nsChangeHint_RepaintFrame);
+  } else if (thisVis && !NS_IsHintSubset(nsChangeHint_RepaintFrame, hint)) {
+    // Both style contexts have a style-if-visited.
+    PRBool change = PR_FALSE;
+
+    // NB: Calling Peek on |this|, not |thisVis|, since callers may look
+    // at a struct on |this| without looking at the same struct on
+    // |thisVis| (including this function if we skip one of these checks
+    // due to change being true already or due to the old style context
+    // not having a style-if-visited), but not the other way around.
+    if (PeekStyleColor()) {
+      if (thisVis->GetStyleColor()->mColor !=
+          otherVis->GetStyleColor()->mColor) {
+        change = PR_TRUE;
+      }
+    }
+
+    // NB: Calling Peek on |this|, not |thisVis| (see above).
+    if (!change && PeekStyleBackground()) {
+      if (thisVis->GetStyleBackground()->mBackgroundColor !=
+          otherVis->GetStyleBackground()->mBackgroundColor) {
+        change = PR_TRUE;
+      }
+    }
+
+    // NB: Calling Peek on |this|, not |thisVis| (see above).
+    if (!change && PeekStyleBorder()) {
+      const nsStyleBorder *thisVisBorder = thisVis->GetStyleBorder();
+      const nsStyleBorder *otherVisBorder = otherVis->GetStyleBorder();
+      NS_FOR_CSS_SIDES(side) {
+        PRBool thisFG, otherFG;
+        nscolor thisColor, otherColor;
+        thisVisBorder->GetBorderColor(side, thisColor, thisFG);
+        otherVisBorder->GetBorderColor(side, otherColor, otherFG);
+        if (thisFG != otherFG || (!thisFG && thisColor != otherColor)) {
+          change = PR_TRUE;
+          break;
+        }
+      }
+    }
+
+    // NB: Calling Peek on |this|, not |thisVis| (see above).
+    if (!change && PeekStyleOutline()) {
+      const nsStyleOutline *thisVisOutline = thisVis->GetStyleOutline();
+      const nsStyleOutline *otherVisOutline = otherVis->GetStyleOutline();
+      PRBool haveColor;
+      nscolor thisColor, otherColor;
+      if (thisVisOutline->GetOutlineInitialColor() != 
+            otherVisOutline->GetOutlineInitialColor() ||
+          (haveColor = thisVisOutline->GetOutlineColor(thisColor)) != 
+            otherVisOutline->GetOutlineColor(otherColor) ||
+          (haveColor && thisColor != otherColor)) {
+        change = PR_TRUE;
+      }
+    }
+
+    // NB: Calling Peek on |this|, not |thisVis| (see above).
+    if (!change && PeekStyleColumn()) {
+      const nsStyleColumn *thisVisColumn = thisVis->GetStyleColumn();
+      const nsStyleColumn *otherVisColumn = otherVis->GetStyleColumn();
+      if (thisVisColumn->mColumnRuleColor != otherVisColumn->mColumnRuleColor ||
+          thisVisColumn->mColumnRuleColorIsForeground !=
+            otherVisColumn->mColumnRuleColorIsForeground) {
+        change = PR_TRUE;
+      }
+    }
+
+    // NB: Calling Peek on |this|, not |thisVis| (see above).
+    if (!change && PeekStyleSVG()) {
+      const nsStyleSVG *thisVisSVG = thisVis->GetStyleSVG();
+      const nsStyleSVG *otherVisSVG = otherVis->GetStyleSVG();
+      if (thisVisSVG->mFill != otherVisSVG->mFill ||
+          thisVisSVG->mStroke != otherVisSVG->mStroke) {
+        change = PR_TRUE;
+      }
+    }
+
+    if (change) {
+      NS_UpdateHint(hint, nsChangeHint_RepaintFrame);
+    }
+  }
+
   return hint;
 }
 
@@ -555,7 +679,7 @@ void
 nsStyleContext::Destroy()
 {
   // Get the pres context from our rule node.
-  nsCOMPtr<nsPresContext> presContext = mRuleNode->GetPresContext();
+  nsRefPtr<nsPresContext> presContext = mRuleNode->GetPresContext();
 
   // Call our destructor.
   this->~nsStyleContext();
@@ -568,13 +692,112 @@ nsStyleContext::Destroy()
 already_AddRefed<nsStyleContext>
 NS_NewStyleContext(nsStyleContext* aParentContext,
                    nsIAtom* aPseudoTag,
+                   nsCSSPseudoElements::Type aPseudoType,
                    nsRuleNode* aRuleNode,
                    nsPresContext* aPresContext)
 {
-  nsStyleContext* context = new (aPresContext) nsStyleContext(aParentContext, aPseudoTag, 
-                                                              aRuleNode, aPresContext);
+  nsStyleContext* context =
+    new (aPresContext) nsStyleContext(aParentContext, aPseudoTag, aPseudoType,
+                                      aRuleNode, aPresContext);
   if (context)
     context->AddRef();
   return context;
 }
 
+static nscolor ExtractColor(nsCSSProperty aProperty,
+                            nsStyleContext *aStyleContext)
+{
+  nsStyleAnimation::Value val;
+#ifdef DEBUG
+  PRBool success =
+#endif
+    nsStyleAnimation::ExtractComputedValue(aProperty, aStyleContext, val);
+  NS_ABORT_IF_FALSE(success,
+                    "aProperty must be extractable by nsStyleAnimation");
+  return val.GetColorValue();
+}
+
+struct ColorIndexSet {
+  PRUint8 colorIndex, alphaIndex;
+};
+
+static const ColorIndexSet gVisitedIndices[2] = { { 0, 0 }, { 1, 0 } };
+
+nscolor
+nsStyleContext::GetVisitedDependentColor(nsCSSProperty aProperty)
+{
+  NS_ASSERTION(aProperty == eCSSProperty_color ||
+               aProperty == eCSSProperty_background_color ||
+               aProperty == eCSSProperty_border_top_color ||
+               aProperty == eCSSProperty_border_right_color_value ||
+               aProperty == eCSSProperty_border_bottom_color ||
+               aProperty == eCSSProperty_border_left_color_value ||
+               aProperty == eCSSProperty_outline_color ||
+               aProperty == eCSSProperty__moz_column_rule_color ||
+               aProperty == eCSSProperty_fill ||
+               aProperty == eCSSProperty_stroke,
+               "we need to add to nsStyleContext::CalcStyleDifference");
+
+  nscolor colors[2];
+  colors[0] = ExtractColor(aProperty, this);
+
+  nsStyleContext *visitedStyle = this->GetStyleIfVisited();
+  if (!visitedStyle) {
+    return colors[0];
+  }
+
+  colors[1] = ExtractColor(aProperty, visitedStyle);
+
+  return nsStyleContext::CombineVisitedColors(colors,
+                                              this->RelevantLinkVisited());
+}
+
+/* static */ nscolor
+nsStyleContext::CombineVisitedColors(nscolor *aColors, PRBool aLinkIsVisited)
+{
+  if (NS_GET_A(aColors[1]) == 0) {
+    // If the style-if-visited is transparent, then just use the
+    // unvisited style rather than using the (meaningless) color
+    // components of the visited style along with a potentially
+    // non-transparent alpha value.
+    aLinkIsVisited = PR_FALSE;
+  }
+
+  // NOTE: We want this code to have as little timing dependence as
+  // possible on whether this->RelevantLinkVisited() is true.
+  const ColorIndexSet &set =
+    gVisitedIndices[aLinkIsVisited ? 1 : 0];
+
+  nscolor colorColor = aColors[set.colorIndex];
+  nscolor alphaColor = aColors[set.alphaIndex];
+  return NS_RGBA(NS_GET_R(colorColor), NS_GET_G(colorColor),
+                 NS_GET_B(colorColor), NS_GET_A(alphaColor));
+}
+
+void*
+nsStyleContext::Alloc(size_t aSize)
+{
+  nsIPresShell *shell = PresContext()->PresShell();
+
+  aSize += offsetof(AllocationHeader, mStorageStart);
+  AllocationHeader *alloc =
+    static_cast<AllocationHeader*>(shell->AllocateMisc(aSize));
+
+  alloc->mSize = aSize; // NOTE: inflated by header
+
+  alloc->mNext = mAllocations;
+  mAllocations = alloc;
+
+  return static_cast<void*>(&alloc->mStorageStart);
+}
+
+void
+nsStyleContext::FreeAllocations(nsPresContext *aPresContext)
+{
+  nsIPresShell *shell = aPresContext->PresShell();
+
+  for (AllocationHeader *alloc = mAllocations, *next; alloc; alloc = next) {
+    next = alloc->mNext;
+    shell->FreeMisc(alloc->mSize, alloc);
+  }
+}
