@@ -43,7 +43,6 @@
 
 #include "nsCOMPtr.h"
 #include "nsGridRowGroupLayout.h"
-#include "nsISupportsArray.h"
 #include "nsIServiceManager.h"
 #include "nsGkAtoms.h"
 #include "nsIContent.h"
@@ -56,7 +55,6 @@
 #include "nsCSSFrameConstructor.h"
 #include "nsIScrollableFrame.h"
 #include "nsIScrollbarFrame.h"
-#include "nsIScrollableView.h"
 #include "nsIView.h"
 #include "nsIViewManager.h"
 #include "nsStyleContext.h"
@@ -72,6 +70,7 @@
 #include "nsLayoutUtils.h"
 #include "nsPIListBoxObject.h"
 #include "nsContentUtils.h"
+#include "nsChildIterator.h"
 
 /////////////// nsListScrollSmoother //////////////////
 
@@ -172,25 +171,24 @@ NS_IMPL_ISUPPORTS1(nsListScrollSmoother, nsITimerCallback)
 
 nsListBoxBodyFrame::nsListBoxBodyFrame(nsIPresShell* aPresShell,
                                        nsStyleContext* aContext,
-                                       PRBool aIsRoot,
                                        nsIBoxLayout* aLayoutManager)
-  : nsBoxFrame(aPresShell, aContext, aIsRoot, aLayoutManager),
-    mRowCount(-1),
-    mRowHeight(0),
-    mRowHeightWasSet(PR_FALSE),
-    mAvailableHeight(0),
-    mStringWidth(-1),
+  : nsBoxFrame(aPresShell, aContext, PR_FALSE, aLayoutManager),
     mTopFrame(nsnull),
     mBottomFrame(nsnull),
     mLinkupFrame(nsnull),
+    mScrollSmoother(nsnull),
     mRowsToPrepend(0),
+    mRowCount(-1),
+    mRowHeight(0),
+    mAvailableHeight(0),
+    mStringWidth(-1),
     mCurrentIndex(0),
     mOldIndex(0),
+    mYPosition(0),
+    mTimePerRow(TIME_PER_ROW_INITAL),
+    mRowHeightWasSet(PR_FALSE),
     mScrolling(PR_FALSE),
     mAdjustScroll(PR_FALSE),
-    mYPosition(0),
-    mScrollSmoother(nsnull),
-    mTimePerRow(TIME_PER_ROW_INITAL),
     mReflowCallbackPosted(PR_FALSE)
 {
 }
@@ -206,32 +204,10 @@ nsListBoxBodyFrame::~nsListBoxBodyFrame()
 
 }
 
-////////// nsISupports /////////////////
-
-NS_IMETHODIMP_(nsrefcnt) 
-nsListBoxBodyFrame::AddRef(void)
-{
-  return 2;
-}
-
-NS_IMETHODIMP_(nsrefcnt)
-nsListBoxBodyFrame::Release(void)
-{
-  return 1;
-}
-
-//
-// QueryInterface
-//
-NS_INTERFACE_MAP_BEGIN(nsListBoxBodyFrame)
-  NS_INTERFACE_MAP_ENTRY(nsIScrollbarMediator)
-  if (aIID.Equals(NS_GET_IID(nsListBoxBodyFrame))) {
-    *aInstancePtr = this;
-    return NS_OK;
-  }
-  else
-NS_INTERFACE_MAP_END_INHERITING(nsBoxFrame)
-
+NS_QUERYFRAME_HEAD(nsListBoxBodyFrame)
+  NS_QUERYFRAME_ENTRY(nsIScrollbarMediator)
+  NS_QUERYFRAME_ENTRY(nsListBoxBodyFrame)
+NS_QUERYFRAME_TAIL_INHERITING(nsBoxFrame)
 
 ////////// nsIFrame /////////////////
 
@@ -244,12 +220,9 @@ nsListBoxBodyFrame::Init(nsIContent*     aContent,
   NS_ENSURE_SUCCESS(rv, rv);
   nsIScrollableFrame* scrollFrame = nsLayoutUtils::GetScrollableFrameFor(this);
   if (scrollFrame) {
-    nsIScrollableView* scrollableView = scrollFrame->GetScrollableView();
-    scrollableView->SetScrollProperties(NS_SCROLL_PROPERTY_ALWAYS_BLIT);
     nsIBox* verticalScrollbar = scrollFrame->GetScrollbarBox(PR_TRUE);
     if (verticalScrollbar) {
-      nsIScrollbarFrame* scrollbarFrame = nsnull;
-      CallQueryInterface(verticalScrollbar, &scrollbarFrame);
+      nsIScrollbarFrame* scrollbarFrame = do_QueryFrame(verticalScrollbar);
       scrollbarFrame->SetScrollbarMediatorContent(GetContent());
     }
   }
@@ -261,7 +234,7 @@ nsListBoxBodyFrame::Init(nsIContent*     aContent,
 }
 
 void
-nsListBoxBodyFrame::Destroy()
+nsListBoxBodyFrame::DestroyFrom(nsIFrame* aDestructRoot)
 {
   // make sure we cancel any posted callbacks.
   if (mReflowCallbackPosted)
@@ -277,7 +250,7 @@ nsListBoxBodyFrame::Destroy()
     mBoxObject->ClearCachedValues();
   }
 
-  nsBoxFrame::Destroy();
+  nsBoxFrame::DestroyFrom(aDestructRoot);
 }
 
 NS_IMETHODIMP
@@ -288,21 +261,8 @@ nsListBoxBodyFrame::AttributeChanged(PRInt32 aNameSpaceID,
   nsresult rv = NS_OK;
 
   if (aAttribute == nsGkAtoms::rows) {
-    nsAutoString rows;
-    mContent->GetAttr(kNameSpaceID_None, nsGkAtoms::rows, rows);
-    
-    if (!rows.IsEmpty()) {
-      PRInt32 dummy;
-      PRInt32 count = rows.ToInteger(&dummy);
-      PRInt32 rowHeight = GetRowHeightAppUnits();
-      rowHeight = nsPresContext::AppUnitsToIntCSSPixels(rowHeight);
-      nsAutoString value;
-      value.AppendInt(rowHeight*count);
-      mContent->SetAttr(kNameSpaceID_None, nsGkAtoms::minheight, value, PR_FALSE);
-
-      PresContext()->PresShell()->
-        FrameNeedsReflow(this, nsIPresShell::eStyleChange, NS_FRAME_IS_DIRTY);
-    }
+    PresContext()->PresShell()->
+      FrameNeedsReflow(this, nsIPresShell::eStyleChange, NS_FRAME_IS_DIRTY);
   }
   else
     rv = nsBoxFrame::AttributeChanged(aNameSpaceID, aAttribute, aModType);
@@ -330,6 +290,25 @@ nsListBoxBodyFrame::DoLayout(nsBoxLayoutState& aBoxLayoutState)
 
   nsresult rv = nsBoxFrame::DoLayout(aBoxLayoutState);
 
+  // determine the real height for the scrollable area from the total number
+  // of rows, since non-visible rows don't yet have frames
+  nsRect rect(nsPoint(0, 0), GetSize());
+  nsOverflowAreas overflow(rect, rect);
+  if (mLayoutManager) {
+    nsIFrame* childFrame = mFrames.FirstChild();
+    while (childFrame) {
+      ConsiderChildOverflow(overflow, childFrame);
+      childFrame = childFrame->GetNextSibling();
+    }
+
+    nsSize prefSize = mLayoutManager->GetPrefSize(this, aBoxLayoutState);
+    NS_FOR_FRAME_OVERFLOW_TYPES(otype) {
+      nsRect& o = overflow.Overflow(otype);
+      o.height = NS_MAX(o.height, prefSize.height);
+    }
+  }
+  FinishAndStoreOverflow(overflow, GetSize());
+
   if (mScrolling)
     aBoxLayoutState.SetPaintingDisabled(PR_FALSE);
 
@@ -381,7 +360,7 @@ nsListBoxBodyFrame::GetPrefSize(nsBoxLayoutState& aBoxLayoutState)
 ///////////// nsIScrollbarMediator ///////////////
 
 NS_IMETHODIMP
-nsListBoxBodyFrame::PositionChanged(nsISupports* aScrollbar, PRInt32 aOldIndex, PRInt32& aNewIndex)
+nsListBoxBodyFrame::PositionChanged(nsIScrollbarFrame* aScrollbar, PRInt32 aOldIndex, PRInt32& aNewIndex)
 { 
   if (mScrolling || mRowHeight == 0)
     return NS_OK;
@@ -433,7 +412,7 @@ nsListBoxBodyFrame::PositionChanged(nsISupports* aScrollbar, PRInt32 aOldIndex,
 }
 
 NS_IMETHODIMP
-nsListBoxBodyFrame::VisibilityChanged(nsISupports* aScrollbar, PRBool aVisible)
+nsListBoxBodyFrame::VisibilityChanged(PRBool aVisible)
 {
   if (mRowHeight == 0)
     return NS_OK;
@@ -451,7 +430,7 @@ nsListBoxBodyFrame::VisibilityChanged(nsISupports* aScrollbar, PRBool aVisible)
 }
 
 NS_IMETHODIMP
-nsListBoxBodyFrame::ScrollbarButtonPressed(nsISupports* aScrollbar, PRInt32 aOldIndex, PRInt32 aNewIndex)
+nsListBoxBodyFrame::ScrollbarButtonPressed(nsIScrollbarFrame* aScrollbar, PRInt32 aOldIndex, PRInt32 aNewIndex)
 {
   if (aOldIndex == aNewIndex)
     return NS_OK;
@@ -529,7 +508,6 @@ nsListBoxBodyFrame::GetIndexOfFirstVisibleRow(PRInt32 *aResult)
 nsresult
 nsListBoxBodyFrame::EnsureIndexIsVisible(PRInt32 aRowIndex)
 {
-  NS_ASSERTION(aRowIndex >= 0, "Ensure row is visible called with a negative number!");
   if (aRowIndex < 0)
     return NS_ERROR_ILLEGAL_VALUE;
 
@@ -544,7 +522,6 @@ nsListBoxBodyFrame::EnsureIndexIsVisible(PRInt32 aRowIndex)
   if (mCurrentIndex <= aRowIndex && aRowIndex < bottomIndex)
     return NS_OK;
 
-  // Check to be sure we're not scrolling off the bottom of the tree
   PRInt32 delta;
 
   PRBool up = aRowIndex < mCurrentIndex;
@@ -553,6 +530,10 @@ nsListBoxBodyFrame::EnsureIndexIsVisible(PRInt32 aRowIndex)
     mCurrentIndex = aRowIndex;
   }
   else {
+    // Check to be sure we're not scrolling off the bottom of the tree
+    if (aRowIndex >= GetRowCount())
+      return NS_ERROR_ILLEGAL_VALUE;
+
     // Bring it just into view.
     delta = 1 + (aRowIndex-bottomIndex);
     mCurrentIndex += delta; 
@@ -591,7 +572,7 @@ nsListBoxBodyFrame::ScrollByLines(PRInt32 aNumLines)
     
   // I'd use Composite here, but it doesn't always work.
   // vm->Composite();
-  PresContext()->GetViewManager()->ForceUpdate();
+  PresContext()->GetPresShell()->GetViewManager()->ForceUpdate();
 
   return NS_OK;
 }
@@ -604,13 +585,11 @@ nsListBoxBodyFrame::GetIndexOfItem(nsIDOMElement* aItem, PRInt32* _retval)
     *_retval = 0;
     nsCOMPtr<nsIContent> itemContent(do_QueryInterface(aItem));
 
-    nsIContent* listbox = mContent->GetBindingParent();
-    NS_ENSURE_STATE(listbox);
-
-    PRUint32 childCount = listbox->GetChildCount();
-    for (PRUint32 childIndex = 0; childIndex < childCount; childIndex++) {
-      nsIContent *child = listbox->GetChildAt(childIndex);
-
+    ChildIterator iter, last;
+    for (ChildIterator::Init(mContent, &iter, &last);
+         iter != last;
+         ++iter) {
+      nsIContent *child = (*iter);
       // we hit a list row, count it
       if (child->Tag() == nsGkAtoms::listitem) {
         // is this it?
@@ -634,15 +613,12 @@ nsListBoxBodyFrame::GetItemAtIndex(PRInt32 aIndex, nsIDOMElement** aItem)
   if (aIndex < 0)
     return NS_OK;
   
-  nsIContent* listbox = mContent->GetBindingParent();
-  NS_ENSURE_STATE(listbox);
-
-  PRUint32 childCount = listbox->GetChildCount();
-
   PRInt32 itemCount = 0;
-  for (PRUint32 childIndex = 0; childIndex < childCount; childIndex++) {
-    nsIContent *child = listbox->GetChildAt(childIndex);
-
+  ChildIterator iter, last;
+  for (ChildIterator::Init(mContent, &iter, &last);
+       iter != last;
+       ++iter) {
+    nsIContent *child = (*iter);
     // we hit a list row, check if it is the one we are looking for
     if (child->Tag() == nsGkAtoms::listitem) {
       // is this it?
@@ -690,20 +666,6 @@ nsListBoxBodyFrame::SetRowHeight(nscoord aRowHeight)
 { 
   if (aRowHeight > mRowHeight) { 
     mRowHeight = aRowHeight;
-    
-    nsAutoString rows;
-    mContent->GetAttr(kNameSpaceID_None, nsGkAtoms::rows, rows);
-    if (rows.IsEmpty())
-      mContent->GetAttr(kNameSpaceID_None, nsGkAtoms::size, rows);
-    
-    if (!rows.IsEmpty()) {
-      PRInt32 dummy;
-      PRInt32 count = rows.ToInteger(&dummy);
-      PRInt32 rowHeight = nsPresContext::AppUnitsToIntCSSPixels(aRowHeight);
-      nsAutoString value;
-      value.AppendInt(rowHeight*count);
-      mContent->SetAttr(kNameSpaceID_None, nsGkAtoms::minheight, value, PR_FALSE);
-    }
 
     // signal we need to dirty everything 
     // and we want to be notified after reflow
@@ -716,11 +678,10 @@ nsListBoxBodyFrame::SetRowHeight(nscoord aRowHeight)
 nscoord
 nsListBoxBodyFrame::GetAvailableHeight()
 {
-  nsIScrollableFrame* scrollFrame
-    = nsLayoutUtils::GetScrollableFrameFor(this);
+  nsIScrollableFrame* scrollFrame =
+    nsLayoutUtils::GetScrollableFrameFor(this);
   if (scrollFrame) {
-    nsIScrollableView* scrollView = scrollFrame->GetScrollableView();
-    return scrollView->View()->GetBounds().height;
+    return scrollFrame->GetScrollPortRect().height;
   }
   return 0;
 }
@@ -748,7 +709,7 @@ nsListBoxBodyFrame::ComputeIntrinsicWidth(nsBoxLayoutState& aBoxLayoutState)
     nsRefPtr<nsStyleContext> styleContext;
     nsPresContext *presContext = aBoxLayoutState.PresContext();
     styleContext = presContext->StyleSet()->
-      ResolveStyleFor(firstRowContent, nsnull);
+      ResolveStyleFor(firstRowContent->AsElement(), nsnull);
 
     nscoord width = 0;
     nsMargin margin(0,0,0,0);
@@ -759,13 +720,13 @@ nsListBoxBodyFrame::ComputeIntrinsicWidth(nsBoxLayoutState& aBoxLayoutState)
     if (styleContext->GetStyleMargin()->GetMargin(margin))
       width += margin.LeftRight();
 
-    nsIContent* listbox = mContent->GetBindingParent();
-    NS_ENSURE_TRUE(listbox, largestWidth);
 
-    PRUint32 childCount = listbox->GetChildCount();
-
-    for (PRUint32 i = 0; i < childCount && i < 100; ++i) {
-      nsIContent *child = listbox->GetChildAt(i);
+    ChildIterator iter, last;
+    PRUint32 i = 0;
+    for (ChildIterator::Init(mContent, &iter, &last);
+         iter != last && i < 100;
+         ++iter, ++i) {
+      nsIContent *child = (*iter);
 
       if (child->Tag() == nsGkAtoms::listitem) {
         nsIRenderingContext* rendContext = aBoxLayoutState.GetRenderingContext();
@@ -800,12 +761,12 @@ void
 nsListBoxBodyFrame::ComputeTotalRowCount()
 {
   mRowCount = 0;
-  nsIContent* listbox = mContent->GetBindingParent();
-  ENSURE_TRUE(listbox);
 
-  PRUint32 childCount = listbox->GetChildCount();
-  for (PRUint32 i = 0; i < childCount; i++) {
-    if (listbox->GetChildAt(i)->Tag() == nsGkAtoms::listitem)
+  ChildIterator iter, last;
+  for (ChildIterator::Init(mContent, &iter, &last);
+       iter != last;
+       ++iter) {
+    if ((*iter)->Tag() == nsGkAtoms::listitem)
       ++mRowCount;
   }
 }
@@ -841,9 +802,15 @@ nsListBoxBodyFrame::ScrollToIndex(PRInt32 aRowIndex)
 
   mCurrentIndex = newIndex;
 
+  nsWeakFrame weak(this);
+
   // Since we're going to flush anyway, we need to not do this off an event
   DoInternalPositionChangedSync(up, delta);
 
+  if (!weak.IsAlive()) {
+    return NS_OK;
+  }
+
   // This change has to happen immediately.
   // Flush any pending reflow commands.
   // XXXbz why, exactly?
@@ -894,7 +861,9 @@ nsListBoxBodyFrame::DoInternalPositionChangedSync(PRBool aUp, PRInt32 aDelta)
   nsTArray< nsRefPtr<nsPositionChangedEvent> > temp;
   temp.SwapElements(mPendingPositionChangeEvents);
   for (PRUint32 i = 0; i < temp.Length(); ++i) {
-    temp[i]->Run();
+    if (weak.IsAlive()) {
+      temp[i]->Run();
+    }
     temp[i]->Revoke();
   }
 
@@ -911,7 +880,7 @@ nsListBoxBodyFrame::DoInternalPositionChanged(PRBool aUp, PRInt32 aDelta)
   if (aDelta == 0)
     return NS_OK;
 
-  nsPresContext *presContext = PresContext();
+  nsRefPtr<nsPresContext> presContext(PresContext());
   nsBoxLayoutState state(presContext);
 
   // begin timing how long it takes to scroll a row
@@ -923,45 +892,52 @@ nsListBoxBodyFrame::DoInternalPositionChanged(PRBool aUp, PRInt32 aDelta)
     return NS_OK;
   }
 
-  PRInt32 visibleRows = 0;
-  if (mRowHeight)
-    visibleRows = GetAvailableHeight()/mRowHeight;
+  {
+    nsAutoScriptBlocker scriptBlocker;
+
+    PRInt32 visibleRows = 0;
+    if (mRowHeight)
+      visibleRows = GetAvailableHeight()/mRowHeight;
   
-  if (aDelta < visibleRows) {
-    PRInt32 loseRows = aDelta;
-    if (aUp) {
-      // scrolling up, destroy rows from the bottom downwards
-      ReverseDestroyRows(loseRows);
-      mRowsToPrepend += aDelta;
-      mLinkupFrame = nsnull;
+    if (aDelta < visibleRows) {
+      PRInt32 loseRows = aDelta;
+      if (aUp) {
+        // scrolling up, destroy rows from the bottom downwards
+        ReverseDestroyRows(loseRows);
+        mRowsToPrepend += aDelta;
+        mLinkupFrame = nsnull;
+      }
+      else {
+        // scrolling down, destroy rows from the top upwards
+        DestroyRows(loseRows);
+        mRowsToPrepend = 0;
+      }
     }
     else {
-      // scrolling down, destroy rows from the top upwards
-      DestroyRows(loseRows);
-      mRowsToPrepend = 0;
+      // We have scrolled so much that all of our current frames will
+      // go off screen, so blow them all away. Weeee!
+      nsIFrame *currBox = mFrames.FirstChild();
+      nsCSSFrameConstructor* fc = presContext->PresShell()->FrameConstructor();
+      fc->BeginUpdate();
+      while (currBox) {
+        nsIFrame *nextBox = currBox->GetNextSibling();
+        RemoveChildFrame(state, currBox);
+        currBox = nextBox;
+      }
+      fc->EndUpdate();
     }
-  }
-  else {
-    // We have scrolled so much that all of our current frames will
-    // go off screen, so blow them all away. Weeee!
-    nsIFrame *currBox = mFrames.FirstChild();
-    nsCSSFrameConstructor* fc = PresContext()->PresShell()->FrameConstructor();
-    fc->BeginUpdate();
-    while (currBox) {
-      nsIFrame *nextBox = currBox->GetNextSibling();
-      RemoveChildFrame(state, currBox);
-      currBox = nextBox;
-    }
-    fc->EndUpdate();
-  }
 
-  // clear frame markers so that CreateRows will re-create
-  mTopFrame = mBottomFrame = nsnull; 
+    // clear frame markers so that CreateRows will re-create
+    mTopFrame = mBottomFrame = nsnull; 
   
-  mYPosition = mCurrentIndex*mRowHeight;
-  mScrolling = PR_TRUE;
-  PresContext()->PresShell()->
-    FrameNeedsReflow(this, nsIPresShell::eResize, NS_FRAME_HAS_DIRTY_CHILDREN);
+    mYPosition = mCurrentIndex*mRowHeight;
+    mScrolling = PR_TRUE;
+    presContext->PresShell()->
+      FrameNeedsReflow(this, nsIPresShell::eResize, NS_FRAME_HAS_DIRTY_CHILDREN);
+  }
+  if (!weakThis.IsAlive()) {
+    return NS_OK;
+  }
   // Flush calls CreateRows
   // XXXbz there has to be a better way to do this than flushing!
   presContext->PresShell()->FlushPendingNotifications(Flush_Layout);
@@ -1011,7 +987,8 @@ nsListBoxBodyFrame::VerticalScroll(PRInt32 aPosition)
 
   nsPoint scrollPosition = scrollFrame->GetScrollPosition();
  
-  scrollFrame->ScrollTo(nsPoint(scrollPosition.x, aPosition));
+  scrollFrame->ScrollTo(nsPoint(scrollPosition.x, aPosition),
+                        nsIScrollableFrame::INSTANT);
 
   mYPosition = aPosition;
 }
@@ -1124,7 +1101,7 @@ nsListBoxBodyFrame::ReverseDestroyRows(PRInt32& aRowsToLose)
     --aRowsToLose;
     
     nsIFrame* prevFrame;
-    prevFrame = mFrames.GetPrevSiblingFor(childFrame);
+    prevFrame = childFrame->GetPrevSibling();
     RemoveChildFrame(state, childFrame);
 
     mBottomFrame = childFrame = prevFrame;
@@ -1233,13 +1210,15 @@ nsListBoxBodyFrame::GetNextItemBox(nsIBox* aBox, PRInt32 aOffset,
       // There is a content node that wants a frame.
       nsIContent *nextContent = parentContent->GetChildAt(i + aOffset + 1);
 
-      if (!nextContent->IsNodeOfType(nsINode::eXUL) ||
+      if (!nextContent->IsXUL() ||
           nextContent->Tag() != nsGkAtoms::listitem)
         return GetNextItemBox(aBox, ++aOffset, aCreated);
 
       nsPresContext* presContext = PresContext();
-      nsIFrame* existingFrame =
-        presContext->GetPresShell()->GetPrimaryFrameFor(nextContent);
+      nsIFrame* existingFrame = nextContent->GetPrimaryFrame();
+
+      if (existingFrame && existingFrame->GetParent() != this)
+        return GetNextItemBox(aBox, ++aOffset, aCreated);
 
       if (!existingFrame) {
         // Either append the new frame, or insert it after the current frame
@@ -1319,13 +1298,13 @@ nsListBoxBodyFrame::ContinueReflow(nscoord height)
 }
 
 NS_IMETHODIMP
-nsListBoxBodyFrame::ListBoxAppendFrames(nsIFrame* aFrameList)
+nsListBoxBodyFrame::ListBoxAppendFrames(nsFrameList& aFrameList)
 {
   // append them after
   nsBoxLayoutState state(PresContext());
-  mFrames.AppendFrames(nsnull, aFrameList);
+  const nsFrameList::Slice& newFrames = mFrames.AppendFrames(nsnull, aFrameList);
   if (mLayoutManager)
-    mLayoutManager->ChildrenAppended(this, state, aFrameList);
+    mLayoutManager->ChildrenAppended(this, state, newFrames);
   PresContext()->PresShell()->
     FrameNeedsReflow(this, nsIPresShell::eTreeChange,
                      NS_FRAME_HAS_DIRTY_CHILDREN);
@@ -1334,13 +1313,15 @@ nsListBoxBodyFrame::ListBoxAppendFrames(nsIFrame* aFrameList)
 }
 
 NS_IMETHODIMP
-nsListBoxBodyFrame::ListBoxInsertFrames(nsIFrame* aPrevFrame, nsIFrame* aFrameList)
+nsListBoxBodyFrame::ListBoxInsertFrames(nsIFrame* aPrevFrame,
+                                        nsFrameList& aFrameList)
 {
   // insert the frames to our info list
   nsBoxLayoutState state(PresContext());
-  mFrames.InsertFrames(nsnull, aPrevFrame, aFrameList);
+  const nsFrameList::Slice& newFrames =
+    mFrames.InsertFrames(nsnull, aPrevFrame, aFrameList);
   if (mLayoutManager)
-    mLayoutManager->ChildrenInserted(this, state, aPrevFrame, aFrameList);
+    mLayoutManager->ChildrenInserted(this, state, aPrevFrame, newFrames);
   PresContext()->PresShell()->
     FrameNeedsReflow(this, nsIPresShell::eTreeChange,
                      NS_FRAME_HAS_DIRTY_CHILDREN);
@@ -1357,13 +1338,12 @@ nsListBoxBodyFrame::OnContentInserted(nsPresContext* aPresContext, nsIContent* a
   if (mRowCount >= 0)
     ++mRowCount;
 
-  nsIPresShell *shell = aPresContext->PresShell();
   // The RDF content builder will build content nodes such that they are all 
   // ready when OnContentInserted is first called, meaning the first call
   // to CreateRows will create all the frames, but OnContentInserted will
   // still be called again for each content node - so we need to make sure
   // that the frame for each content node hasn't already been created.
-  nsIFrame* childFrame = shell->GetPrimaryFrameFor(aChildContent);
+  nsIFrame* childFrame = aChildContent->GetPrimaryFrame();
   if (childFrame)
     return;
 
@@ -1378,7 +1358,7 @@ nsListBoxBodyFrame::OnContentInserted(nsPresContext* aPresContext, nsIContent* a
     mRowsToPrepend = 1;
   } else if (nextSiblingContent) {
     // we may be inserting before a frame that is on screen
-    nsIFrame* nextSiblingFrame = shell->GetPrimaryFrameFor(nextSiblingContent);
+    nsIFrame* nextSiblingFrame = nextSiblingContent->GetPrimaryFrame();
     mLinkupFrame = nextSiblingFrame;
   }
   
@@ -1392,7 +1372,10 @@ nsListBoxBodyFrame::OnContentInserted(nsPresContext* aPresContext, nsIContent* a
 // Called by nsCSSFrameConstructor when listitem content is removed.
 //
 void
-nsListBoxBodyFrame::OnContentRemoved(nsPresContext* aPresContext, nsIFrame* aChildFrame, PRInt32 aIndex)
+nsListBoxBodyFrame::OnContentRemoved(nsPresContext* aPresContext,
+                                     nsIContent* aContainer,
+                                     nsIFrame* aChildFrame,
+                                     nsIContent* aOldNextSibling)
 {
   NS_ASSERTION(!aChildFrame || aChildFrame->GetParent() == this,
                "Removing frame that's not our child... Not good");
@@ -1400,17 +1383,16 @@ nsListBoxBodyFrame::OnContentRemoved(nsPresContext* aPresContext, nsIFrame* aChi
   if (mRowCount >= 0)
     --mRowCount;
 
-  nsIContent* listBoxContent = mContent->GetBindingParent();
-  if (listBoxContent) {
+  if (aContainer) {
     if (!aChildFrame) {
       // The row we are removing is out of view, so we need to try to
       // determine the index of its next sibling.
-      nsIContent *oldNextSiblingContent = listBoxContent->GetChildAt(aIndex);
-  
       PRInt32 siblingIndex = -1;
-      if (oldNextSiblingContent) {
+      if (aOldNextSibling) {
         nsCOMPtr<nsIContent> nextSiblingContent;
-        GetListItemNextSibling(oldNextSiblingContent, getter_AddRefs(nextSiblingContent), siblingIndex);
+        GetListItemNextSibling(aOldNextSibling,
+                               getter_AddRefs(nextSiblingContent),
+                               siblingIndex);
       }
     
       // if the row being removed is off-screen and above the top frame, we need to
@@ -1429,11 +1411,13 @@ nsListBoxBodyFrame::OnContentRemoved(nsPresContext* aPresContext, nsIFrame* aChi
       // down by one, and we will have to insert a new frame at the top.
       
       // if the last content node has a frame, we are scrolled to the bottom
-      PRUint32 childCount = listBoxContent->GetChildCount();
-      if (childCount > 0) {
-        nsIContent *lastChild = listBoxContent->GetChildAt(childCount - 1);
-        nsIFrame* lastChildFrame = 
-          aPresContext->PresShell()->GetPrimaryFrameFor(lastChild);
+      ChildIterator iter, last;
+      ChildIterator::Init(mContent, &iter, &last);
+      if (iter != last) {
+        iter = last;
+        --iter;
+        nsIContent *lastChild = *iter;
+        nsIFrame* lastChildFrame = lastChild->GetPrimaryFrame();
       
         if (lastChildFrame) {
           mTopFrame = nsnull;
@@ -1465,14 +1449,13 @@ void
 nsListBoxBodyFrame::GetListItemContentAt(PRInt32 aIndex, nsIContent** aContent)
 {
   *aContent = nsnull;
-  nsIContent* listboxContent = mContent->GetBindingParent();
-  ENSURE_TRUE(listboxContent);
 
-  PRUint32 childCount = listboxContent->GetChildCount();
   PRInt32 itemsFound = 0;
-  for (PRUint32 i = 0; i < childCount; ++i) {
-    nsIContent *kid = listboxContent->GetChildAt(i);
-
+  ChildIterator iter, last;
+  for (ChildIterator::Init(mContent, &iter, &last);
+       iter != last;
+       ++iter) {
+    nsIContent *kid = (*iter);
     if (kid->Tag() == nsGkAtoms::listitem) {
       ++itemsFound;
       if (itemsFound-1 == aIndex) {
@@ -1480,7 +1463,6 @@ nsListBoxBodyFrame::GetListItemContentAt(PRInt32 aIndex, nsIContent** aContent)
         NS_IF_ADDREF(*aContent);
         return;
       }
-        
     }
   }
 }
@@ -1490,13 +1472,12 @@ nsListBoxBodyFrame::GetListItemNextSibling(nsIContent* aListItem, nsIContent** a
 {
   *aContent = nsnull;
   aSiblingIndex = -1;
-  nsIContent* listboxContent = mContent->GetBindingParent();
-  ENSURE_TRUE(listboxContent);
-
-  PRUint32 childCount = listboxContent->GetChildCount();
   nsIContent *prevKid = nsnull;
-  for (PRUint32 i = 0; i < childCount; ++i) {
-    nsIContent *kid = listboxContent->GetChildAt(i);
+  ChildIterator iter, last;
+  for (ChildIterator::Init(mContent, &iter, &last);
+       iter != last;
+       ++iter) {
+    nsIContent *kid = (*iter);
 
     if (kid->Tag() == nsGkAtoms::listitem) {
       ++aSiblingIndex;
@@ -1505,7 +1486,6 @@ nsListBoxBodyFrame::GetListItemNextSibling(nsIContent* aListItem, nsIContent** a
         NS_IF_ADDREF(*aContent);
         return;
       }
-        
     }
     prevKid = kid;
   }
@@ -1517,16 +1497,17 @@ void
 nsListBoxBodyFrame::RemoveChildFrame(nsBoxLayoutState &aState,
                                      nsIFrame         *aFrame)
 {
-  nsPresContext* presContext = PresContext();
-  nsCSSFrameConstructor* fc = presContext->PresShell()->FrameConstructor();
-  fc->RemoveMappingsForFrameSubtree(aFrame);
+  if (!mFrames.ContainsFrame(aFrame)) {
+    NS_ERROR("tried to remove a child frame which isn't our child");
+    return;
+  }
 
-#ifdef DEBUG
-  PRBool removed =
-#endif
-    mFrames.RemoveFrame(aFrame);
-  NS_ASSERTION(removed,
-               "Going to destroy a frame we didn't remove.  Prepare to crash");
+  if (aFrame == GetContentInsertionFrame()) {
+    // Don't touch that one
+    return;
+  }
+
+  mFrames.RemoveFrame(aFrame);
   if (mLayoutManager)
     mLayoutManager->ChildrenRemoved(this, aState, aFrame);
   aFrame->Destroy();
@@ -1534,10 +1515,17 @@ nsListBoxBodyFrame::RemoveChildFrame(nsBoxLayoutState &aState,
 
 // Creation Routines ///////////////////////////////////////////////////////////////////////
 
+already_AddRefed<nsIBoxLayout> NS_NewListBoxLayout();
+
 nsIFrame*
-NS_NewListBoxBodyFrame(nsIPresShell* aPresShell, nsStyleContext* aContext,
-                       PRBool aIsRoot, nsIBoxLayout* aLayoutManager)
+NS_NewListBoxBodyFrame(nsIPresShell* aPresShell, nsStyleContext* aContext)
 {
-  return 
-    new (aPresShell) nsListBoxBodyFrame(aPresShell, aContext, aIsRoot, aLayoutManager);
+  nsCOMPtr<nsIBoxLayout> layout = NS_NewListBoxLayout();
+  if (!layout) {
+    return nsnull;
+  }
+
+  return new (aPresShell) nsListBoxBodyFrame(aPresShell, aContext, layout);
 }
+
+NS_IMPL_FRAMEARENA_HELPERS(nsListBoxBodyFrame)
