@@ -37,7 +37,7 @@
 /*
  * RSA key generation, public key op, private key op.
  *
- * $Id: rsa.c,v 1.39 2009/02/03 05:34:41 julien.pierre.boogz%sun.com Exp $
+ * $Id: rsa.c,v 1.39.22.1 2010/11/16 19:06:38 rrelyea%redhat.com Exp $
  */
 #ifdef FREEBL_NO_DEPEND
 #include "stubs.h"
@@ -116,21 +116,21 @@ static struct RSABlindingParamsListStr blindingParamsList = { 0 };
 static PRBool nssRSAUseBlinding = PR_TRUE;
 
 static SECStatus
-rsa_keygen_from_primes(mp_int *p, mp_int *q, mp_int *e, RSAPrivateKey *key,
-                       unsigned int keySizeInBits)
+rsa_build_from_primes(mp_int *p, mp_int *q, 
+		mp_int *e, PRBool needPublicExponent, 
+		mp_int *d, PRBool needPrivateExponent,
+		RSAPrivateKey *key, unsigned int keySizeInBits)
 {
-    mp_int n, d, phi;
+    mp_int n, phi;
     mp_int psub1, qsub1, tmp;
     mp_err   err = MP_OKAY;
     SECStatus rv = SECSuccess;
     MP_DIGITS(&n)     = 0;
-    MP_DIGITS(&d)     = 0;
     MP_DIGITS(&phi)   = 0;
     MP_DIGITS(&psub1) = 0;
     MP_DIGITS(&qsub1) = 0;
     MP_DIGITS(&tmp)   = 0;
     CHECK_MPI_OK( mp_init(&n)     );
-    CHECK_MPI_OK( mp_init(&d)     );
     CHECK_MPI_OK( mp_init(&phi)   );
     CHECK_MPI_OK( mp_init(&psub1) );
     CHECK_MPI_OK( mp_init(&qsub1) );
@@ -143,12 +143,25 @@ rsa_keygen_from_primes(mp_int *p, mp_int *q, mp_int *e, RSAPrivateKey *key,
 	rv = SECFailure;
 	goto cleanup;
     }
+
+    /* at least one exponent must be given */
+    PORT_Assert(!(needPublicExponent && needPrivateExponent));
+
     /* 2.  Compute phi = (p-1)*(q-1) */
     CHECK_MPI_OK( mp_sub_d(p, 1, &psub1) );
     CHECK_MPI_OK( mp_sub_d(q, 1, &qsub1) );
-    CHECK_MPI_OK( mp_mul(&psub1, &qsub1, &phi) );
-    /* 3.  Compute d = e**-1 mod(phi) */
-    err = mp_invmod(e, &phi, &d);
+    if (needPublicExponent || needPrivateExponent) {
+	CHECK_MPI_OK( mp_mul(&psub1, &qsub1, &phi) );
+	/* 3.  Compute d = e**-1 mod(phi) */
+	/*     or      e = d**-1 mod(phi) as necessary */
+	if (needPublicExponent) {
+	    err = mp_invmod(d, &phi, e);
+	} else {
+	    err = mp_invmod(e, &phi, d);
+	}
+    } else {
+	err = MP_OKAY;
+    }
     /*     Verify that phi(n) and e have no common divisors */
     if (err != MP_OKAY) {
 	if (err == MP_UNDEF) {
@@ -158,20 +171,30 @@ rsa_keygen_from_primes(mp_int *p, mp_int *q, mp_int *e, RSAPrivateKey *key,
 	}
 	goto cleanup;
     }
-    MPINT_TO_SECITEM(&n, &key->modulus, key->arena);
-    MPINT_TO_SECITEM(&d, &key->privateExponent, key->arena);
+
     /* 4.  Compute exponent1 = d mod (p-1) */
-    CHECK_MPI_OK( mp_mod(&d, &psub1, &tmp) );
+    CHECK_MPI_OK( mp_mod(d, &psub1, &tmp) );
     MPINT_TO_SECITEM(&tmp, &key->exponent1, key->arena);
     /* 5.  Compute exponent2 = d mod (q-1) */
-    CHECK_MPI_OK( mp_mod(&d, &qsub1, &tmp) );
+    CHECK_MPI_OK( mp_mod(d, &qsub1, &tmp) );
     MPINT_TO_SECITEM(&tmp, &key->exponent2, key->arena);
     /* 6.  Compute coefficient = q**-1 mod p */
     CHECK_MPI_OK( mp_invmod(q, p, &tmp) );
     MPINT_TO_SECITEM(&tmp, &key->coefficient, key->arena);
+
+    /* copy our calculated results, overwrite what is there */
+    key->modulus.data = NULL;
+    MPINT_TO_SECITEM(&n, &key->modulus, key->arena);
+    key->privateExponent.data = NULL;
+    MPINT_TO_SECITEM(d, &key->privateExponent, key->arena);
+    key->publicExponent.data = NULL;
+    MPINT_TO_SECITEM(e, &key->publicExponent, key->arena);
+    key->prime1.data = NULL;
+    MPINT_TO_SECITEM(p, &key->prime1, key->arena);
+    key->prime2.data = NULL;
+    MPINT_TO_SECITEM(q, &key->prime2, key->arena);
 cleanup:
     mp_clear(&n);
-    mp_clear(&d);
     mp_clear(&phi);
     mp_clear(&psub1);
     mp_clear(&qsub1);
@@ -229,7 +252,7 @@ RSAPrivateKey *
 RSA_NewKey(int keySizeInBits, SECItem *publicExponent)
 {
     unsigned int primeLen;
-    mp_int p, q, e;
+    mp_int p, q, e, d;
     int kiter;
     mp_err   err = MP_OKAY;
     SECStatus rv = SECSuccess;
@@ -260,14 +283,15 @@ RSA_NewKey(int keySizeInBits, SECItem *publicExponent)
     MP_DIGITS(&p) = 0;
     MP_DIGITS(&q) = 0;
     MP_DIGITS(&e) = 0;
+    MP_DIGITS(&d) = 0;
     CHECK_MPI_OK( mp_init(&p) );
     CHECK_MPI_OK( mp_init(&q) );
     CHECK_MPI_OK( mp_init(&e) );
+    CHECK_MPI_OK( mp_init(&d) );
     /* 2.  Set the version number (PKCS1 v1.5 says it should be zero) */
     SECITEM_AllocItem(arena, &key->version, 1);
     key->version.data[0] = 0;
     /* 3.  Set the public exponent */
-    SECITEM_CopyItem(arena, &key->publicExponent, publicExponent);
     SECITEM_TO_MPINT(*publicExponent, &e);
     kiter = 0;
     do {
@@ -279,7 +303,10 @@ RSA_NewKey(int keySizeInBits, SECItem *publicExponent)
 	if (mp_cmp(&p, &q) < 0)
 	    mp_exch(&p, &q);
 	/* Attempt to use these primes to generate a key */
-	rv = rsa_keygen_from_primes(&p, &q, &e, key, keySizeInBits);
+	rv = rsa_build_from_primes(&p, &q, 
+			&e, PR_FALSE,  /* needPublicExponent=false */
+			&d, PR_TRUE,   /* needPrivateExponent=true */
+			key, keySizeInBits);
 	if (rv == SECSuccess)
 	    break; /* generated two good primes */
 	prerr = PORT_GetError();
@@ -288,12 +315,11 @@ RSA_NewKey(int keySizeInBits, SECItem *publicExponent)
     } while (prerr == SEC_ERROR_NEED_RANDOM && kiter < MAX_KEY_GEN_ATTEMPTS);
     if (prerr)
 	goto cleanup;
-    MPINT_TO_SECITEM(&p, &key->prime1, arena);
-    MPINT_TO_SECITEM(&q, &key->prime2, arena);
 cleanup:
     mp_clear(&p);
     mp_clear(&q);
     mp_clear(&e);
+    mp_clear(&d);
     if (err) {
 	MP_TO_SEC_ERROR(err);
 	rv = SECFailure;
@@ -305,6 +331,484 @@ cleanup:
     return key;
 }
 
+mp_err
+rsa_is_prime(mp_int *p) {
+    int res;
+
+    /* run a Fermat test */
+    res = mpp_fermat(p, 2);
+    if (res != MP_OKAY) {
+	return res;
+    }
+
+    /* If that passed, run some Miller-Rabin tests */
+    res = mpp_pprime(p, 2);
+    return res;
+}
+
+/*
+ * Try to find the two primes based on 2 exponents plus either a prime
+ *   or a modulus.
+ *
+ * In: e, d and either p or n (depending on the setting of hasModulus).
+ * Out: p,q.
+ * 
+ * Step 1, Since d = e**-1 mod phi, we know that d*e == 1 mod phi, or
+ *	d*e = 1+k*phi, or d*e-1 = k*phi. since d is less than phi and e is
+ *	usually less than d, then k must be an integer between e-1 and 1 
+ *	(probably on the order of e).
+ * Step 1a, If we were passed just a prime, we can divide k*phi by that
+ *      prime-1 and get k*(q-1). This will reduce the size of our division
+ *      through the rest of the loop.
+ * Step 2, Loop through the values k=e-1 to 1 looking for k. k should be on
+ *	the order or e, and e is typically small. This may take a while for
+ *	a large random e. We are looking for a k that divides kphi
+ *	evenly. Once we find a k that divides kphi evenly, we assume it 
+ *	is the true k. It's possible this k is not the 'true' k but has 
+ *	swapped factors of p-1 and/or q-1. Because of this, we 
+ *	tentatively continue Steps 3-6 inside this loop, and may return looking
+ *	for another k on failure.
+ * Step 3, Calculate are tentative phi=kphi/k. Note: real phi is (p-1)*(q-1).
+ * Step 4a, if we have a prime, kphi is already k*(q-1), so phi is or tenative
+ *      q-1. q = phi+1. If k is correct, q should be the right length and 
+ *      prime.
+ * Step 4b, It's possible q-1 and k could have swapped factors. We now have a
+ * 	possible solution that meets our criteria. It may not be the only 
+ *      solution, however, so we keep looking. If we find more than one, 
+ *      we will fail since we cannot determine which is the correct
+ *      solution, and returning the wrong modulus will compromise both
+ *      moduli. If no other solution is found, we return the unique solution.
+ * Step 5a, If we have the modulus (n=pq), then use the following formula to 
+ * 	calculate  s=(p+q): , phi = (p-1)(q-1) = pq  -p-q +1 = n-s+1. so
+ *	s=n-phi+1.
+ * Step 5b, Use n=pq and s=p+q to solve for p and q as follows:
+ *	since q=s-p, then n=p*(s-p)= sp - p^2, rearranging p^2-s*p+n = 0.
+ *	from the quadratic equation we have p=1/2*(s+sqrt(s*s-4*n)) and
+ *	q=1/2*(s-sqrt(s*s-4*n)) if s*s-4*n is a perfect square, we are DONE.
+ *	If it is not, continue in our look looking for another k. NOTE: the
+ *	code actually distributes the 1/2 and results in the equations:
+ *	sqrt = sqrt(s/2*s/2-n), p=s/2+sqrt, q=s/2-sqrt. The algebra saves us
+ *	and extra divide by 2 and a multiply by 4.
+ * 
+ * This will return p & q. q may be larger than p in the case that p was given
+ * and it was the smaller prime.
+ */
+static mp_err
+rsa_get_primes_from_exponents(mp_int *e, mp_int *d, mp_int *p, mp_int *q,
+			      mp_int *n, PRBool hasModulus, 
+			      unsigned int keySizeInBits)
+{
+    mp_int kphi; /* k*phi */
+    mp_int k;    /* current guess at 'k' */
+    mp_int phi;  /* (p-1)(q-1) */
+    mp_int s;    /* p+q/2 (s/2 in the algebra) */
+    mp_int r;    /* remainder */
+    mp_int tmp; /* p-1 if p is given, n+1 is modulus is given */
+    mp_int sqrt; /* sqrt(s/2*s/2-n) */
+    mp_err err = MP_OKAY;
+    unsigned int order_k;
+
+    MP_DIGITS(&kphi) = 0;
+    MP_DIGITS(&phi) = 0;
+    MP_DIGITS(&s) = 0;
+    MP_DIGITS(&k) = 0;
+    MP_DIGITS(&r) = 0;
+    MP_DIGITS(&tmp) = 0;
+    MP_DIGITS(&sqrt) = 0;
+    CHECK_MPI_OK( mp_init(&kphi) );
+    CHECK_MPI_OK( mp_init(&phi) );
+    CHECK_MPI_OK( mp_init(&s) );
+    CHECK_MPI_OK( mp_init(&k) );
+    CHECK_MPI_OK( mp_init(&r) );
+    CHECK_MPI_OK( mp_init(&tmp) );
+    CHECK_MPI_OK( mp_init(&sqrt) );
+
+    /* our algorithm looks for a factor k whose maximum size is dependent
+     * on the size of our smallest exponent, which had better be the public
+     * exponent (if it's the private, the key is vulnerable to a brute force
+     * attack).
+     * 
+     * since our factor search is linear, we need to limit the maximum
+     * size of the public key. this should not be a problem normally, since 
+     * public keys are usually small. 
+     *
+     * if we want to handle larger public key sizes, we should have
+     * a version which tries to 'completely' factor k*phi (where completely
+     * means 'factor into primes, or composites with which are products of
+     * large primes). Once we have all the factors, we can sort them out and
+     * try different combinations to form our phi. The risk is if (p-1)/2,
+     * (q-1)/2, and k are all large primes. In any case if the public key
+     * is small (order of 20 some bits), then a linear search for k is 
+     * manageable.
+     */
+    if (mpl_significant_bits(e) > 23) {
+	err=MP_RANGE;
+	goto cleanup;
+    }
+
+    /* calculate k*phi = e*d - 1 */
+    CHECK_MPI_OK( mp_mul(e, d, &kphi) );
+    CHECK_MPI_OK( mp_sub_d(&kphi, 1, &kphi) );
+
+
+    /* kphi is (e*d)-1, which is the same as k*(p-1)(q-1)
+     * d < (p-1)(q-1), therefor k must be less than e-1
+     * We can narrow down k even more, though. Since p and q are odd and both 
+     * have their high bit set, then we know that phi must be on order of 
+     * keySizeBits.
+     */
+    order_k = (unsigned)mpl_significant_bits(&kphi) - keySizeInBits;
+
+    /* for (k=kinit; order(k) >= order_k; k--) { */
+    /* k=kinit: k can't be bigger than  kphi/2^(keySizeInBits -1) */
+    CHECK_MPI_OK( mp_2expt(&k,keySizeInBits-1) );
+    CHECK_MPI_OK( mp_div(&kphi, &k, &k, NULL));
+    if (mp_cmp(&k,e) >= 0) {
+	/* also can't be bigger then e-1 */
+        CHECK_MPI_OK( mp_sub_d(e, 1, &k) );
+    }
+
+    /* calculate our temp value */
+    /* This saves recalculating this value when the k guess is wrong, which
+     * is reasonably frequent. */
+    /* for the modulus case, tmp = n+1 (used to calculate p+q = tmp - phi) */
+    /* for the prime case, tmp = p-1 (used to calculate q-1= phi/tmp) */
+    if (hasModulus) {
+	CHECK_MPI_OK( mp_add_d(n, 1, &tmp) );
+    } else {
+	CHECK_MPI_OK( mp_sub_d(p, 1, &tmp) );
+	CHECK_MPI_OK(mp_div(&kphi,&tmp,&kphi,&r));
+	if (mp_cmp_z(&r) != 0) {
+	    /* p-1 doesn't divide kphi, some parameter wasn't correct */
+	    err=MP_RANGE;
+	    goto cleanup;
+	}
+	mp_zero(q);
+	/* kphi is now k*(q-1) */
+    }
+
+    /* rest of the for loop */
+    for (; (err == MP_OKAY) && (mpl_significant_bits(&k) >= order_k); 
+						err = mp_sub_d(&k, 1, &k)) {
+	/* looking for k as a factor of kphi */
+	CHECK_MPI_OK(mp_div(&kphi,&k,&phi,&r));
+	if (mp_cmp_z(&r) != 0) {
+	    /* not a factor, try the next one */
+	    continue;
+	}
+	/* we have a possible phi, see if it works */
+	if (!hasModulus) {
+	    if ((unsigned)mpl_significant_bits(&phi) != keySizeInBits/2) {
+		/* phi is not the right size */
+		continue;
+	    }
+	    /* phi should be divisible by 2, since
+	     * q is odd and phi=(q-1). */
+	    if (mpp_divis_d(&phi,2) == MP_NO) {
+		/* phi is not divisible by 4 */
+		continue;
+	    }
+	    /* we now have a candidate for the second prime */
+	    CHECK_MPI_OK(mp_add_d(&phi, 1, &tmp));
+	    
+	    /* check to make sure it is prime */
+	    err = rsa_is_prime(&tmp);
+	    if (err != MP_OKAY) {
+		if (err == MP_NO) {
+		    /* No, then we still have the wrong phi */
+		    err = MP_OKAY;
+        	    continue;
+		}
+		goto cleanup;
+	    }
+	    /*
+	     * It is possible that we have the wrong phi if 
+	     * k_guess*(q_guess-1) = k*(q-1) (k and q-1 have swapped factors).
+	     * since our q_quess is prime, however. We have found a valid
+	     * rsa key because:
+	     *   q is the correct order of magnitude.
+	     *   phi = (p-1)(q-1) where p and q are both primes.
+	     *   e*d mod phi = 1.
+	     * There is no way to know from the info given if this is the 
+	     * original key. We never want to return the wrong key because if
+	     * two moduli with the same factor is known, then euclid's gcd
+	     * algorithm can be used to find that factor. Even though the 
+	     * caller didn't pass the original modulus, it doesn't mean the
+	     * modulus wasn't known or isn't available somewhere. So to be safe
+	     * if we can't be sure we have the right q, we don't return any.
+	     * 
+	     * So to make sure we continue looking for other valid q's. If none
+	     * are found, then we can safely return this one, otherwise we just
+	     * fail */
+	    if (mp_cmp_z(q) != 0) {
+		/* this is the second valid q, don't return either, 
+		 * just fail */
+		err = MP_RANGE;
+		break;
+	    }
+	    /* we only have one q so far, save it and if no others are found,
+	     * it's safe to return it */
+	    CHECK_MPI_OK(mp_copy(&tmp, q));
+	    continue;
+	}
+	/* test our tentative phi */
+	/* phi should be the correct order */
+	if ((unsigned)mpl_significant_bits(&phi) != keySizeInBits) {
+	    /* phi is not the right size */
+	    continue;
+	}
+	/* phi should be divisible by 4, since
+	 * p and q are odd and phi=(p-1)(q-1). */
+	if (mpp_divis_d(&phi,4) == MP_NO) {
+	    /* phi is not divisible by 4 */
+	    continue;
+	}
+	/* n was given, calculate s/2=(p+q)/2 */
+	CHECK_MPI_OK( mp_sub(&tmp, &phi, &s) );
+	CHECK_MPI_OK( mp_div_2(&s, &s) );
+
+	/* calculate sqrt(s/2*s/2-n) */
+	CHECK_MPI_OK(mp_sqr(&s,&sqrt));
+	CHECK_MPI_OK(mp_sub(&sqrt,n,&r));  /* r as a tmp */
+	CHECK_MPI_OK(mp_sqrt(&r,&sqrt));
+	/* make sure it's a perfect square */
+	/* r is our original value we took the square root of */
+	/* q is the square of our tentative square root. They should be equal*/
+	CHECK_MPI_OK(mp_sqr(&sqrt,q)); /* q as a tmp */
+	if (mp_cmp(&r,q) != 0) {
+	    /* sigh according to the doc, mp_sqrt could return sqrt-1 */
+	   CHECK_MPI_OK(mp_add_d(&sqrt,1,&sqrt));
+	   CHECK_MPI_OK(mp_sqr(&sqrt,q));
+	   if (mp_cmp(&r,q) != 0) {
+		/* s*s-n not a perfect square, this phi isn't valid, find 			 * another.*/
+		continue;
+	    }
+	}
+
+	/* NOTE: In this case we know we have the one and only answer.
+	 * "Why?", you ask. Because:
+	 *    1) n is a composite of two large primes (or it wasn't a
+	 *       valid RSA modulus).
+	 *    2) If we know any number such that x^2-n is a perfect square 
+	 *       and x is not (n+1)/2, then we can calculate 2 non-trivial
+	 *       factors of n.
+	 *    3) Since we know that n has only 2 non-trivial prime factors, 
+	 *       we know the two factors we have are the only possible factors.
+	 */
+
+	/* Now we are home free to calculate p and q */
+	/* p = s/2 + sqrt, q= s/2 - sqrt */
+	CHECK_MPI_OK(mp_add(&s,&sqrt,p));
+	CHECK_MPI_OK(mp_sub(&s,&sqrt,q));
+	break;
+    }
+    if ((unsigned)mpl_significant_bits(&k) < order_k) {
+	if (hasModulus || (mp_cmp_z(q) == 0)) {
+	    /* If we get here, something was wrong with the parameters we 
+	     * were given */
+	    err = MP_RANGE; 
+	}
+    }
+cleanup:
+    mp_clear(&kphi);
+    mp_clear(&phi);
+    mp_clear(&s);
+    mp_clear(&k);
+    mp_clear(&r);
+    mp_clear(&tmp);
+    mp_clear(&sqrt);
+    return err;
+}
+     
+/*
+ * take a private key with only a few elements and fill out the missing pieces.
+ *
+ * All the entries will be overwritten with data allocated out of the arena
+ * If no arena is supplied, one will be created.
+ *
+ * The following fields must be supplied in order for this function
+ * to succeed:
+ *   one of either publicExponent or privateExponent
+ *   two more of the following 5 parameters.
+ *      modulus (n)
+ *      prime1  (p)
+ *      prime2  (q)
+ *      publicExponent (e)
+ *      privateExponent (d)
+ *
+ * NOTE: if only the publicExponent, privateExponent, and one prime is given,
+ * then there may be more than one RSA key that matches that combination.
+ *
+ * All parameters will be replaced in the key structure with new parameters
+ * Allocated out of the arena. There is no attempt to free the old structures.
+ * Prime1 will always be greater than prime2 (even if the caller supplies the
+ * smaller prime as prime1 or the larger prime as prime2). The parameters are
+ * not overwritten on failure.
+ *
+ *  How it works:
+ *     We can generate all the parameters from:
+ *        one of the exponents, plus the two primes. (rsa_build_key_from_primes) *
+ *     If we are given one of the exponents and both primes, we are done.
+ *     If we are given one of the exponents, the modulus and one prime, we 
+ *        caclulate the second prime by dividing the modulus by the given 
+ *        prime, giving us and exponent and 2 primes.
+ *     If we are given 2 exponents and either the modulus or one of the primes
+ *        we calculate k*phi = d*e-1, where k is an integer less than d which 
+ *        divides d*e-1. We find factor k so we can isolate phi.
+ *            phi = (p-1)(q-1)
+ *       If one of the primes are given, we can use phi to find the other prime
+ *        as follows: q = (phi/(p-1)) + 1. We now have 2 primes and an 
+ *        exponent. (NOTE: if more then one prime meets this condition, the
+ *        operation will fail. See comments elsewhere in this file about this).
+ *       If the modulus is given, then we can calculate the sum of the primes
+ *        as follows: s := (p+q), phi = (p-1)(q-1) = pq -p - q +1, pq = n ->
+ *        phi = n - s + 1, s = n - phi +1.  Now that we have s = p+q and n=pq,
+ *	  we can solve our 2 equations and 2 unknowns as follows: q=s-p ->
+ *        n=p*(s-p)= sp -p^2 -> p^2-sp+n = 0. Using the quadratic to solve for
+ *        p, p=1/2*(s+ sqrt(s*s-4*n)) [q=1/2*(s-sqrt(s*s-4*n)]. We again have
+ *        2 primes and an exponent.
+ *
+ */
+SECStatus
+RSA_PopulatePrivateKey(RSAPrivateKey *key)
+{
+    PRArenaPool *arena = NULL;
+    PRBool needPublicExponent = PR_TRUE;
+    PRBool needPrivateExponent = PR_TRUE;
+    PRBool hasModulus = PR_FALSE;
+    unsigned int keySizeInBits = 0;
+    int prime_count = 0;
+    /* standard RSA nominclature */
+    mp_int p, q, e, d, n;
+    /* remainder */
+    mp_int r;
+    mp_err err = 0;
+    SECStatus rv = SECFailure;
+
+    MP_DIGITS(&p) = 0;
+    MP_DIGITS(&q) = 0;
+    MP_DIGITS(&e) = 0;
+    MP_DIGITS(&d) = 0;
+    MP_DIGITS(&n) = 0;
+    MP_DIGITS(&r) = 0;
+    CHECK_MPI_OK( mp_init(&p) );
+    CHECK_MPI_OK( mp_init(&q) );
+    CHECK_MPI_OK( mp_init(&e) );
+    CHECK_MPI_OK( mp_init(&d) );
+    CHECK_MPI_OK( mp_init(&n) );
+    CHECK_MPI_OK( mp_init(&r) );
+ 
+    /* if the key didn't already have an arena, create one. */
+    if (key->arena == NULL) {
+	arena = PORT_NewArena(NSS_FREEBL_DEFAULT_CHUNKSIZE);
+	if (!arena) {
+	    goto cleanup;
+	}
+	key->arena = arena;
+    }
+
+    /* load up the known exponents */
+    if (key->publicExponent.data) {
+        SECITEM_TO_MPINT(key->publicExponent, &e);
+	needPublicExponent = PR_FALSE;
+    } 
+    if (key->privateExponent.data) {
+        SECITEM_TO_MPINT(key->privateExponent, &d);
+	needPrivateExponent = PR_FALSE;
+    }
+    if (needPrivateExponent && needPublicExponent) {
+	/* Not enough information, we need at least one exponent */
+	err = MP_BADARG;
+	goto cleanup;
+    }
+
+    /* load up the known primes. If only one prime is given, it will be
+     * assigned 'p'. Once we have both primes, well make sure p is the larger.
+     * The value prime_count tells us howe many we have acquired.
+     */
+    if (key->prime1.data) {
+	int primeLen = key->prime1.len;
+	if (key->prime1.data[0] == 0) {
+	   primeLen--;
+	}
+	keySizeInBits = primeLen * 2 * BITS_PER_BYTE;
+        SECITEM_TO_MPINT(key->prime1, &p);
+	prime_count++;
+    }
+    if (key->prime2.data) {
+	int primeLen = key->prime2.len;
+	if (key->prime2.data[0] == 0) {
+	   primeLen--;
+	}
+	keySizeInBits = primeLen * 2 * BITS_PER_BYTE;
+        SECITEM_TO_MPINT(key->prime2, prime_count ? &q : &p);
+	prime_count++;
+    }
+    /* load up the modulus */
+    if (key->modulus.data) {
+	int modLen = key->modulus.len;
+	if (key->modulus.data[0] == 0) {
+	   modLen--;
+	}
+	keySizeInBits = modLen * BITS_PER_BYTE;
+	SECITEM_TO_MPINT(key->modulus, &n);
+	hasModulus = PR_TRUE;
+    }
+    /* if we have the modulus and one prime, calculate the second. */
+    if ((prime_count == 1) && (hasModulus)) {
+	mp_div(&n,&p,&q,&r);
+	if (mp_cmp_z(&r) != 0) {
+	   /* p is not a factor or n, fail */
+	   err = MP_BADARG;
+	   goto cleanup;
+	}
+	prime_count++;
+    }
+
+    /* If we didn't have enough primes try to calculate the primes from
+     * the exponents */
+    if (prime_count < 2) {
+	/* if we don't have at least 2 primes at this point, then we need both
+	 * exponents and one prime or a modulus*/
+	if (!needPublicExponent && !needPrivateExponent &&
+		((prime_count > 0) || hasModulus)) {
+	    CHECK_MPI_OK(rsa_get_primes_from_exponents(&e,&d,&p,&q,
+			&n,hasModulus,keySizeInBits));
+	} else {
+	    /* not enough given parameters to get both primes */
+	    err = MP_BADARG;
+	    goto cleanup;
+	}
+     }
+
+     /* force p to the the larger prime */
+     if (mp_cmp(&p, &q) < 0)
+	mp_exch(&p, &q);
+
+     /* we now have our 2 primes and at least one exponent, we can fill
+      * in the key */
+     rv = rsa_build_from_primes(&p, &q, 
+			&e, needPublicExponent,
+			&d, needPrivateExponent,
+			key, keySizeInBits);
+cleanup:
+    mp_clear(&p);
+    mp_clear(&q);
+    mp_clear(&e);
+    mp_clear(&d);
+    mp_clear(&n);
+    mp_clear(&r);
+    if (err) {
+	MP_TO_SEC_ERROR(err);
+	rv = SECFailure;
+    }
+    if (rv && arena) {
+	PORT_FreeArena(arena, PR_TRUE);
+	key->arena = NULL;
+    }
+    return rv;
+}
+
 static unsigned int
 rsa_modulusLen(SECItem *modulus)
 {
