@@ -51,11 +51,14 @@
 #include <string>
 #include <vector>
 #include <algorithm>
+#include <stdarg.h>
 #include "prinit.h"
 #include "prerror.h"
+#include "prenv.h"
 #include "prio.h"
 #include "prnetdb.h"
 #include "prtpool.h"
+#include "prtypes.h"
 #include "nss.h"
 #include "pk11func.h"
 #include "key.h"
@@ -66,6 +69,110 @@
 using std::string;
 using std::vector;
 
+#define IS_DELIM(m, c)          ((m)[(c) >> 3] & (1 << ((c) & 7)))
+#define SET_DELIM(m, c)         ((m)[(c) >> 3] |= (1 << ((c) & 7)))
+#define DELIM_TABLE_SIZE        32
+
+// You can set the level of logging by env var SSLTUNNEL_LOG_LEVEL=n, where n
+// is 0 through 3.  The default is 1, INFO level logging.
+enum LogLevel {
+  LEVEL_DEBUG = 0,
+  LEVEL_INFO = 1,
+  LEVEL_ERROR = 2,
+  LEVEL_SILENT = 3
+} gLogLevel, gLastLogLevel;
+
+#define _LOG_OUTPUT(level, func, params) \
+PR_BEGIN_MACRO \
+  if (level >= gLogLevel) { \
+    gLastLogLevel = level; \
+    func params;\
+  } \
+PR_END_MACRO
+
+// The most verbose output
+#define LOG_DEBUG(params) \
+  _LOG_OUTPUT(LEVEL_DEBUG, printf, params)
+
+// Top level informative messages
+#define LOG_INFO(params) \
+  _LOG_OUTPUT(LEVEL_INFO, printf, params)
+
+// Serious errors that must be logged always until completely gag
+#define LOG_ERROR(params) \
+  _LOG_OUTPUT(LEVEL_ERROR, eprintf, params)
+
+// Same as LOG_ERROR, but when logging is set to LEVEL_DEBUG, the message 
+// will be put to the stdout instead of stderr to keep continuity with other 
+// LOG_DEBUG message output
+#define LOG_ERRORD(params) \
+PR_BEGIN_MACRO \
+  if (gLogLevel == LEVEL_DEBUG) \
+    _LOG_OUTPUT(LEVEL_ERROR, printf, params); \
+  else \
+    _LOG_OUTPUT(LEVEL_ERROR, eprintf, params); \
+PR_END_MACRO
+
+// If there is any output written between LOG_BEGIN_BLOCK() and
+// LOG_END_BLOCK() then a new line will be put to the proper output (out/err)
+#define LOG_BEGIN_BLOCK() \
+  gLastLogLevel = LEVEL_SILENT;
+
+#define LOG_END_BLOCK() \
+PR_BEGIN_MACRO \
+  if (gLastLogLevel == LEVEL_ERROR) \
+    LOG_ERROR(("\n")); \
+  if (gLastLogLevel < LEVEL_ERROR) \
+    _LOG_OUTPUT(gLastLogLevel, printf, ("\n")); \
+PR_END_MACRO
+
+int eprintf(const char* str, ...)
+{
+  va_list ap;
+  va_start(ap, str);
+  int result = vfprintf(stderr, str, ap);
+  va_end(ap);
+  return result;
+}
+
+// Copied from nsCRT
+char* strtok2(char* string, const char* delims, char* *newStr)
+{
+  PR_ASSERT(string);
+  
+  char delimTable[DELIM_TABLE_SIZE];
+  PRUint32 i;
+  char* result;
+  char* str = string;
+  
+  for (i = 0; i < DELIM_TABLE_SIZE; i++)
+    delimTable[i] = '\0';
+  
+  for (i = 0; delims[i]; i++) {
+    SET_DELIM(delimTable, static_cast<PRUint8>(delims[i]));
+  }
+  
+  // skip to beginning
+  while (*str && IS_DELIM(delimTable, static_cast<PRUint8>(*str))) {
+    str++;
+  }
+  result = str;
+  
+  // fix up the end of the token
+  while (*str) {
+    if (IS_DELIM(delimTable, static_cast<PRUint8>(*str))) {
+      *str++ = '\0';
+      break;
+    }
+    str++;
+  }
+  *newStr = str;
+  
+  return str == result ? NULL : result;
+}
+
+
+
 enum client_auth_option {
   caNone = 0,
   caRequire = 1,
@@ -84,8 +191,20 @@ typedef struct {
   PRFileDesc* client_sock;
   PRNetAddr client_addr;
   server_info_t* server_info;
+  // the original host in the Host: header for this connection is
+  // stored here, for proxied connections
+  string original_host;
+  // true if no SSL should be used for this connection
+  bool http_proxy_only;
+  // true if this connection is for a WebSocket
+  bool iswebsocket;
 } connection_info_t;
 
+typedef struct {
+  string fullHost;
+  bool matched;
+} server_match_t;
+
 const PRInt32 BUF_SIZE = 16384;
 const PRInt32 BUF_MARGIN = 1024;
 const PRInt32 BUF_TOTAL = BUF_SIZE + BUF_MARGIN;
@@ -157,14 +276,15 @@ private:
 
 // These are suggestions. If the number of ports to proxy on * 2
 // is greater than either of these, then we'll use that value instead.
-const PRInt32 INITIAL_THREADS = 1;
-const PRInt32 MAX_THREADS = 5;
-const PRInt32 DEFAULT_STACKSIZE = (512 * 1024);
+const PRUint32 INITIAL_THREADS = 1;
+const PRUint32 MAX_THREADS = 5;
+const PRUint32 DEFAULT_STACKSIZE = (512 * 1024);
 
 // global data
 string nssconfigdir;
 vector<server_info_t> servers;
 PRNetAddr remote_addr;
+PRNetAddr websocket_server;
 PRThreadPool* threads = NULL;
 PRLock* shutdown_lock = NULL;
 PRCondVar* shutdown_condvar = NULL;
@@ -185,6 +305,14 @@ PR_CALLBACK PRIntn ClientAuthValueComparator(const void *v1, const void *v2)
     return -1;
 }
 
+static PRIntn match_hostname(PLHashEntry *he, PRIntn index, void* arg)
+{
+  server_match_t *match = (server_match_t*)arg;
+  if (match->fullHost.find((char*)he->key) != string::npos)
+    match->matched = true;
+  return HT_ENUMERATE_NEXT;
+}
+
 /*
  * Signal the main thread that the application should shut down.
  */
@@ -199,21 +327,36 @@ bool ReadConnectRequest(server_info_t* server_info,
     relayBuffer& buffer, PRInt32* result, string& certificate,
     client_auth_option* clientauth, string& host)
 {
-  if (buffer.present() < 4)
+  if (buffer.present() < 4) {
+    LOG_DEBUG((" !! only %d bytes present in the buffer", (int)buffer.present()));
     return false;
-  if (strncmp(buffer.buffertail-4, "\r\n\r\n", 4))
+  }
+  if (strncmp(buffer.buffertail-4, "\r\n\r\n", 4)) {
+    LOG_ERRORD((" !! request is not tailed with CRLFCRLF but with %x %x %x %x", 
+               *(buffer.buffertail-4),
+               *(buffer.buffertail-3),
+               *(buffer.buffertail-2),
+               *(buffer.buffertail-1)));
     return false;
+  }
+
+  LOG_DEBUG((" parsing initial connect request, dump:\n%.*s\n", (int)buffer.present(), buffer.bufferhead));
 
   *result = 400;
 
   char* token;
-  token = strtok(buffer.bufferhead, " ");
-  if (!token) 
+  char* _caret;
+  token = strtok2(buffer.bufferhead, " ", &_caret);
+  if (!token) {
+    LOG_ERRORD((" no space found"));
     return true;
-  if (strcmp(token, "CONNECT")) 
+  }
+  if (strcmp(token, "CONNECT")) {
+    LOG_ERRORD((" not CONNECT request but %s", token));
     return true;
+  }
 
-  token = strtok(NULL, " ");
+  token = strtok2(_caret, " ", &_caret);
   void* c = PL_HashTableLookup(server_info->host_cert_table, token);
   if (c)
     certificate = static_cast<char*>(c);
@@ -227,9 +370,11 @@ bool ReadConnectRequest(server_info_t* server_info,
   else
     *clientauth = caNone;
 
-  token = strtok(NULL, "/");
-  if (strcmp(token, "HTTP"))
+  token = strtok2(_caret, "/", &_caret);
+  if (strcmp(token, "HTTP")) {  
+    LOG_ERRORD((" not tailed with HTTP but with %s", token));
     return true;
+  }
 
   *result = 200;
   return true;
@@ -243,26 +388,26 @@ bool ConfigureSSLServerSocket(PRFileDesc* socket, server_info_t* si, string &cer
   AutoCert cert(PK11_FindCertFromNickname(
       certnick, NULL));
   if (!cert) {
-    fprintf(stderr, "Failed to find cert %s\n", si->cert_nickname.c_str());
+    LOG_ERROR(("Failed to find cert %s\n", certnick));
     return false;
   }
 
   AutoKey privKey(PK11_FindKeyByAnyCert(cert, NULL));
   if (!privKey) {
-    fprintf(stderr, "Failed to find private key\n");
+    LOG_ERROR(("Failed to find private key\n"));
     return false;
   }
 
   PRFileDesc* ssl_socket = SSL_ImportFD(NULL, socket);
   if (!ssl_socket) {
-    fprintf(stderr, "Error importing SSL socket\n");
+    LOG_ERROR(("Error importing SSL socket\n"));
     return false;
   }
 
   SSLKEAType certKEA = NSS_FindCertKEAType(cert);
   if (SSL_ConfigSecureServer(ssl_socket, cert, privKey, certKEA)
       != SECSuccess) {
-    fprintf(stderr, "Error configuring SSL server socket\n");
+    LOG_ERROR(("Error configuring SSL server socket\n"));
     return false;
   }
 
@@ -281,6 +426,113 @@ bool ConfigureSSLServerSocket(PRFileDesc* socket, server_info_t* si, string &cer
   return true;
 }
 
+/**
+ * This function examines the buffer for a S5ec-WebSocket-Location: field, 
+ * and if it's present, it replaces the hostname in that field with the
+ * value in the server's original_host field.  This function works
+ * in the reverse direction as AdjustWebSocketHost(), replacing the real
+ * hostname of a response with the potentially fake hostname that is expected
+ * by the browser (e.g., mochi.test).
+ *
+ * @return true if the header was adjusted successfully, or not found, false
+ * if the header is present but the url is not, which should indicate
+ * that more data needs to be read from the socket
+ */
+bool AdjustWebSocketLocation(relayBuffer& buffer, connection_info_t *ci)
+{
+  assert(buffer.margin());
+  buffer.buffertail[1] = '\0';
+
+  char* wsloc = strstr(buffer.bufferhead, "Sec-WebSocket-Location:");
+  if (!wsloc)
+    return true;
+  // advance pointer to the start of the hostname
+  wsloc = strstr(wsloc, "ws://");
+  if (!wsloc)
+    return false;
+  wsloc += 5;
+  // find the end of the hostname
+  char* wslocend = strchr(wsloc + 1, '/');
+  if (!wslocend)
+    return false;
+  char *crlf = strstr(wsloc, "\r\n");
+  if (!crlf)
+    return false;
+  if (ci->original_host.empty())
+    return true;
+
+  int diff = ci->original_host.length() - (wslocend-wsloc);
+  if (diff > 0)
+    assert(size_t(diff) <= buffer.margin());
+  memmove(wslocend + diff, wslocend, buffer.buffertail - wsloc - diff);
+  buffer.buffertail += diff;
+
+  memcpy(wsloc, ci->original_host.c_str(), ci->original_host.length());
+  return true;
+}
+
+/**
+ * This function examines the buffer for a Host: field, and if it's present,
+ * it replaces the hostname in that field with the hostname in the server's
+ * remote_addr field.  This is needed because proxy requests may be coming
+ * from mochitest with fake hosts, like mochi.test, and these need to be
+ * replaced with the host that the destination server is actually running
+ * on.
+ */
+bool AdjustWebSocketHost(relayBuffer& buffer, connection_info_t *ci)
+{
+  const char HEADER_UPGRADE[] = "Upgrade:";
+  const char HEADER_HOST[] = "Host:";
+
+  PRNetAddr inet_addr = (websocket_server.inet.port ? websocket_server :
+    remote_addr);
+
+  assert(buffer.margin());
+
+  // Cannot use strnchr so add a null char at the end. There is always some
+  // space left because we preserve a margin.
+  buffer.buffertail[1] = '\0';
+
+  // Verify this is a WebSocket header.
+  char* h1 = strstr(buffer.bufferhead, HEADER_UPGRADE);
+  if (!h1)
+    return false;
+  h1 += strlen(HEADER_UPGRADE);
+  h1 += strspn(h1, " \t");
+  char* h2 = strstr(h1, "WebSocket\r\n");
+  if (!h2)
+    return false;
+
+  char* host = strstr(buffer.bufferhead, HEADER_HOST);
+  if (!host)
+    return false;
+  // advance pointer to beginning of hostname
+  host += strlen(HEADER_HOST);
+  host += strspn(host, " \t");
+
+  char* endhost = strstr(host, "\r\n");
+  if (!endhost)
+    return false;
+
+  // Save the original host, so we can use it later on responses from the
+  // server.
+  ci->original_host.assign(host, endhost-host);
+
+  char newhost[40];
+  PR_NetAddrToString(&inet_addr, newhost, sizeof(newhost));
+  assert(strlen(newhost) < sizeof(newhost) - 7);
+  sprintf(newhost, "%s:%d", newhost, PR_ntohs(inet_addr.inet.port));
+
+  int diff = strlen(newhost) - (endhost-host);
+  if (diff > 0)
+    assert(size_t(diff) <= buffer.margin());
+  memmove(endhost + diff, endhost, buffer.buffertail - host - diff);
+  buffer.buffertail += diff;
+
+  memcpy(host, newhost, strlen(newhost));
+  return true;
+}
+
 /**
  * This function prefixes Request-URI path with a full scheme-host-port
  * string.
@@ -292,6 +544,7 @@ bool AdjustRequestURI(relayBuffer& buffer, string *host)
   // Cannot use strnchr so add a null char at the end. There is always some space left
   // because we preserve a margin.
   buffer.buffertail[1] = '\0';
+  LOG_DEBUG((" incoming request to adjust:\n%s\n", buffer.bufferhead));
 
   char *token, *path;
   path = strchr(buffer.bufferhead, ' ') + 1;
@@ -344,7 +597,7 @@ bool ConnectSocket(PRFileDesc *fd, const PRNetAddr *addr, PRIntervalTime timeout
 void HandleConnection(void* data)
 {
   connection_info_t* ci = static_cast<connection_info_t*>(data);
-  PRIntervalTime connect_timeout = PR_SecondsToInterval(2);
+  PRIntervalTime connect_timeout = PR_SecondsToInterval(30);
 
   AutoFD other_sock(PR_NewTCPSocket());
   bool client_done = false;
@@ -356,6 +609,10 @@ void HandleConnection(void* data)
   client_auth_option clientAuth;
   string fullHost;
 
+  LOG_DEBUG(("SSLTUNNEL(%p)): incoming connection csock(0)=%p, ssock(1)=%p\n",
+         static_cast<void*>(data),
+         static_cast<void*>(ci->client_sock),
+         static_cast<void*>(other_sock)));
   if (other_sock) 
   {
     PRInt32 numberOfSockets = 1;
@@ -364,7 +621,8 @@ void HandleConnection(void* data)
 
     if (!do_http_proxy)
     {
-      if (!ConfigureSSLServerSocket(ci->client_sock, ci->server_info, certificateToUse, caNone))
+      if (!ci->http_proxy_only && 
+          !ConfigureSSLServerSocket(ci->client_sock, ci->server_info, certificateToUse, caNone))
         client_error = true;
       else if (!ConnectSocket(other_sock, &remote_addr, connect_timeout))
         client_error = true;
@@ -377,21 +635,34 @@ void HandleConnection(void* data)
       {ci->client_sock, PR_POLL_READ, 0},
       {other_sock, PR_POLL_READ, 0}
     };
+    PRBool socketErrorState[2] = {PR_FALSE, PR_FALSE};
 
     while (!((client_error||client_done) && buffers[0].empty() && buffers[1].empty()))
     {
       sockets[0].in_flags |= PR_POLL_EXCEPT;
       sockets[1].in_flags |= PR_POLL_EXCEPT;
+      LOG_DEBUG(("SSLTUNNEL(%p)): polling flags csock(0)=%c%c, ssock(1)=%c%c\n",
+                 static_cast<void*>(data),
+                 sockets[0].in_flags & PR_POLL_READ  ? 'R' : '-',
+                 sockets[0].in_flags & PR_POLL_WRITE ? 'W' : '-',
+                 sockets[1].in_flags & PR_POLL_READ  ? 'R' : '-',
+                 sockets[1].in_flags & PR_POLL_WRITE ? 'W' : '-'));
       PRInt32 pollStatus = PR_Poll(sockets, numberOfSockets, PR_MillisecondsToInterval(1000));
       if (pollStatus < 0)
       {
+        LOG_DEBUG(("SSLTUNNEL(%p)): pollStatus=%d, exiting\n",
+                   static_cast<void*>(data), pollStatus));
         client_error = true;
         break;
       }
 
       if (pollStatus == 0)
+      {
         // timeout
+        LOG_DEBUG(("SSLTUNNEL(%p)): poll timeout, looping\n",
+                   static_cast<void*>(data)));
         continue;
+      }
 
       for (PRInt32 s = 0; s < numberOfSockets; ++s)
       {
@@ -401,9 +672,18 @@ void HandleConnection(void* data)
         PRInt16 &in_flags2 = sockets[s2].in_flags;
         sockets[s].out_flags = 0;
 
+        LOG_BEGIN_BLOCK();
+        LOG_DEBUG(("SSLTUNNEL(%p)): %csock(%d)=%p out_flags=%d",
+                   static_cast<void*>(data),
+                   s == 0 ? 'c' : 's',
+                   s,
+                   static_cast<void*>(sockets[s].fd),
+                   out_flags));
         if (out_flags & (PR_POLL_EXCEPT | PR_POLL_ERR | PR_POLL_HUP))
         {
+          LOG_DEBUG((" :exception\n"));
           client_error = true;
+          socketErrorState[s] = PR_TRUE;
           // We got a fatal error state on the socket. Clear the output buffer
           // for this socket to break the main loop, we will never more be able
           // to send those data anyway.
@@ -411,36 +691,87 @@ void HandleConnection(void* data)
           continue;
         } // PR_POLL_EXCEPT, PR_POLL_ERR, PR_POLL_HUP handling
 
+        if (out_flags & PR_POLL_READ && !buffers[s].free())
+        {
+           LOG_DEBUG((" no place in read buffer but got read flag, dropping it now!"));
+           in_flags &= ~PR_POLL_READ;
+        }
+
         if (out_flags & PR_POLL_READ && buffers[s].free())
         {
+          LOG_DEBUG((" :reading"));
           PRInt32 bytesRead = PR_Recv(sockets[s].fd, buffers[s].buffertail, 
               buffers[s].free(), 0, PR_INTERVAL_NO_TIMEOUT);
 
           if (bytesRead == 0)
           {
+            LOG_DEBUG((" socket gracefully closed"));
             client_done = true;
             in_flags &= ~PR_POLL_READ;
           }
           else if (bytesRead < 0)
           {
             if (PR_GetError() != PR_WOULD_BLOCK_ERROR)
+            {
+              LOG_DEBUG((" error=%d", PR_GetError()));
+              // We are in error state, indicate that the connection was 
+              // not closed gracefully
               client_error = true;
+              socketErrorState[s] = PR_TRUE;
+              // Wipe out our send buffer, we cannot send it anyway.
+              buffers[s2].bufferhead = buffers[s2].buffertail = buffers[s2].buffer;
+            }
+            else
+              LOG_DEBUG((" would block"));
           }
           else
           {
+            // If the other socket is in error state (unable to send/receive)
+            // throw this data away and continue loop
+            if (socketErrorState[s2])
+            {
+              LOG_DEBUG((" have read but other socket is in error state\n"));
+              continue;
+            }
+
             buffers[s].buffertail += bytesRead;
+            LOG_DEBUG((", read %d bytes", bytesRead));
 
             // We have to accept and handle the initial CONNECT request here
             PRInt32 response;
             if (!connect_accepted && ReadConnectRequest(ci->server_info, buffers[s],
                 &response, certificateToUse, &clientAuth, fullHost))
             {
+              // Mark this as a proxy-only connection (no SSL) if the CONNECT
+              // request didn't come for port 443 or from any of the server's
+              // cert or clientauth hostnames.
+              if (fullHost.find(":443") == string::npos)
+              {
+                server_match_t match;
+                match.fullHost = fullHost;
+                match.matched = false;
+                PL_HashTableEnumerateEntries(ci->server_info->host_cert_table, 
+                                             match_hostname, 
+                                             &match);
+                PL_HashTableEnumerateEntries(ci->server_info->host_clientauth_table, 
+                                             match_hostname, 
+                                             &match);
+                ci->http_proxy_only = !match.matched;
+              }
+              else
+              {
+                ci->http_proxy_only = false;
+              }
+
               // Clean the request as it would be read
               buffers[s].bufferhead = buffers[s].buffertail = buffers[s].buffer;
+              in_flags |= PR_POLL_WRITE;
+              connect_accepted = true;
 
               // Store response to the oposite buffer
               if (response != 200)
               {
+                LOG_ERRORD((" could not read the connect request, closing connection with %d", response));
                 client_done = true;
                 sprintf(buffers[s2].buffer, "HTTP/1.1 %d ERROR\r\nConnection: close\r\n\r\n", response);
                 buffers[s2].buffertail = buffers[s2].buffer + strlen(buffers[s2].buffer);
@@ -450,77 +781,126 @@ void HandleConnection(void* data)
               strcpy(buffers[s2].buffer, "HTTP/1.1 200 Connected\r\nConnection: keep-alive\r\n\r\n");
               buffers[s2].buffertail = buffers[s2].buffer + strlen(buffers[s2].buffer);
 
-              if (!ConnectSocket(other_sock, &remote_addr, connect_timeout))
-              {
-                client_error = true;
-                break;
-              }
-
+              LOG_DEBUG((" accepted CONNECT request, connected to the server, sending OK to the client\n"));
               // Send the response to the client socket
-              in_flags |= PR_POLL_WRITE;
-              connect_accepted = true;
               break;
             } // end of CONNECT handling
 
-            if (!buffers[s].free()) // Do not poll for read when the buffer is full
+            if (!buffers[s].free())
+            {
+              // Do not poll for read when the buffer is full
+              LOG_DEBUG((" no place in our read buffer, stop reading"));
               in_flags &= ~PR_POLL_READ;
+            }
 
             if (ssl_updated)
             {
-              if (s == 0 && expect_request_start)
-                expect_request_start = !AdjustRequestURI(buffers[s], &fullHost);
+              if (s == 0 && expect_request_start) 
+              {
+                if (!strstr(buffers[s].bufferhead, "\r\n\r\n"))
+                {
+                  // We haven't received the complete header yet, so wait.
+                  continue;
+                }
+                else
+                {
+                  ci->iswebsocket = AdjustWebSocketHost(buffers[s], ci);
+                  expect_request_start = !(ci->iswebsocket || 
+                                           AdjustRequestURI(buffers[s], &fullHost));
+                  PRNetAddr* addr = &remote_addr;
+                  if (ci->iswebsocket && websocket_server.inet.port)
+                    addr = &websocket_server;
+                  if (!ConnectSocket(other_sock, addr, connect_timeout))
+                  {
+                    LOG_ERRORD((" could not open connection to the real server\n"));
+                    client_error = true;
+                    break;
+                  }
+                  LOG_DEBUG(("\n connected to remote server\n"));
+                  numberOfSockets = 2;
+                }
+              }
+              else if (s == 1 && ci->iswebsocket)
+              {
+                if (!AdjustWebSocketLocation(buffers[s], ci))
+                  continue;
+              }
 
               in_flags2 |= PR_POLL_WRITE;
+              LOG_DEBUG((" telling the other socket to write"));
             }
+            else
+              LOG_DEBUG((" we have something for the other socket to write, but ssl has not been administered on it"));
           }
         } // PR_POLL_READ handling
 
         if (out_flags & PR_POLL_WRITE)
         {
+          LOG_DEBUG((" :writing"));
           PRInt32 bytesWrite = PR_Send(sockets[s].fd, buffers[s2].bufferhead, 
               buffers[s2].present(), 0, PR_INTERVAL_NO_TIMEOUT);
 
           if (bytesWrite < 0)
           {
             if (PR_GetError() != PR_WOULD_BLOCK_ERROR) {
+              LOG_DEBUG((" error=%d", PR_GetError()));
               client_error = true;
+              socketErrorState[s] = PR_TRUE;
               // We got a fatal error while writting the buffer. Clear it to break
               // the main loop, we will never more be able to send it.
               buffers[s2].bufferhead = buffers[s2].buffertail = buffers[s2].buffer;
             }
+            else
+              LOG_DEBUG((" would block"));
           }
           else
           {
+            LOG_DEBUG((", written %d bytes", bytesWrite));
+            buffers[s2].buffertail[1] = '\0';
+            LOG_DEBUG((" dump:\n%.*s\n", bytesWrite, buffers[s2].bufferhead));
+            
             buffers[s2].bufferhead += bytesWrite;
             if (buffers[s2].present())
-              in_flags |= PR_POLL_WRITE;              
+            {
+              LOG_DEBUG((" still have to write %d bytes", (int)buffers[s2].present()));
+              in_flags |= PR_POLL_WRITE;
+            }              
             else
             {
               if (!ssl_updated)
               {
+                LOG_DEBUG((" proxy response sent to the client"));
                 // Proxy response has just been writen, update to ssl
                 ssl_updated = true;
-                if (!ConfigureSSLServerSocket(ci->client_sock, ci->server_info, certificateToUse, clientAuth))
+                if (!ci->http_proxy_only && 
+                    !ConfigureSSLServerSocket(ci->client_sock, ci->server_info, certificateToUse, clientAuth))
                 {
+                  LOG_ERRORD((" failed to config server socket\n"));
                   client_error = true;
                   break;
                 }
 
-                numberOfSockets = 2;
+                LOG_DEBUG((" client socket updated to SSL"));
               } // sslUpdate
 
-              in_flags &= ~PR_POLL_WRITE;              
+              LOG_DEBUG((" dropping our write flag and setting other socket read flag"));
+              in_flags &= ~PR_POLL_WRITE;
               in_flags2 |= PR_POLL_READ;
               buffers[s2].compact();
             }
           }
         } // PR_POLL_WRITE handling
+        LOG_END_BLOCK(); // end the log
       } // for...
     } // while, poll
   }
   else
     client_error = true;
 
+  LOG_DEBUG(("SSLTUNNEL(%p)): exiting root function for csock=%p, ssock=%p\n",
+             static_cast<void*>(data),
+             static_cast<void*>(ci->client_sock),
+             static_cast<void*>(other_sock)));
   if (!client_error)
     PR_Shutdown(ci->client_sock, PR_SHUTDOWN_SEND);
   PR_Close(ci->client_sock);
@@ -541,27 +921,34 @@ void StartServer(void* data)
   //TODO: select ciphers?
   AutoFD listen_socket(PR_NewTCPSocket());
   if (!listen_socket) {
-    fprintf(stderr, "failed to create socket\n");
+    LOG_ERROR(("failed to create socket\n"));
     SignalShutdown();
     return;
   }
 
+  // In case the socket is still open in the TIME_WAIT state from a previous
+  // instance of ssltunnel we ask to reuse the port.
+  PRSocketOptionData socket_option;
+  socket_option.option = PR_SockOpt_Reuseaddr;
+  socket_option.value.reuse_addr = PR_TRUE;
+  PR_SetSocketOption(listen_socket, &socket_option);
+
   PRNetAddr server_addr;
   PR_InitializeNetAddr(PR_IpAddrAny, si->listen_port, &server_addr);
   if (PR_Bind(listen_socket, &server_addr) != PR_SUCCESS) {
-    fprintf(stderr, "failed to bind socket\n");
+    LOG_ERROR(("failed to bind socket\n"));
     SignalShutdown();
     return;
   }
 
   if (PR_Listen(listen_socket, 1) != PR_SUCCESS) {
-    fprintf(stderr, "failed to listen on socket\n");
+    LOG_ERROR(("failed to listen on socket\n"));
     SignalShutdown();
     return;
   }
 
-  printf("Server listening on port %d with cert %s\n", si->listen_port,
-         si->cert_nickname.c_str());
+  LOG_INFO(("Server listening on port %d with cert %s\n", si->listen_port,
+         si->cert_nickname.c_str()));
 
   while (!shutdown_server) {
     connection_info_t* ci = new connection_info_t();
@@ -590,7 +977,7 @@ char* password_func(PK11SlotInfo* slot, PRBool retry, void* arg)
   if (retry)
     return NULL;
 
-  return "";
+  return PL_strdup("");
 }
 
 server_info_t* findServerInfo(int portnumber)
@@ -610,30 +997,48 @@ int processConfigLine(char* configLine)
   if (*configLine == 0 || *configLine == '#')
     return 0;
 
-  char* keyword = strtok(configLine, ":");
+  char* _caret;
+  char* keyword = strtok2(configLine, ":", &_caret);
 
   // Configure usage of http/ssl tunneling proxy behavior
   if (!strcmp(keyword, "httpproxy"))
   {
-    char* value = strtok(NULL, ":");
+    char* value = strtok2(_caret, ":", &_caret);
     if (!strcmp(value, "1"))
       do_http_proxy = true;
 
     return 0;
   }
 
+  if (!strcmp(keyword, "websocketserver"))
+  {
+    char* ipstring = strtok2(_caret, ":", &_caret);
+    if (PR_StringToNetAddr(ipstring, &websocket_server) != PR_SUCCESS) {
+      LOG_ERROR(("Invalid IP address in proxy config: %s\n", ipstring));
+      return 1;
+    }
+    char* remoteport = strtok2(_caret, ":", &_caret);
+    int port = atoi(remoteport);
+    if (port <= 0) {
+      LOG_ERROR(("Invalid remote port in proxy config: %s\n", remoteport));
+      return 1;
+    }
+    websocket_server.inet.port = PR_htons(port);
+    return 0;
+  }
+
   // Configure the forward address of the target server
   if (!strcmp(keyword, "forward"))
   {
-    char* ipstring = strtok(NULL, ":");
+    char* ipstring = strtok2(_caret, ":", &_caret);
     if (PR_StringToNetAddr(ipstring, &remote_addr) != PR_SUCCESS) {
-      fprintf(stderr, "Invalid remote IP address: %s\n", ipstring);
+      LOG_ERROR(("Invalid remote IP address: %s\n", ipstring));
       return 1;
     }
-    char* serverportstring = strtok(NULL, ":");
+    char* serverportstring = strtok2(_caret, ":", &_caret);
     int port = atoi(serverportstring);
     if (port <= 0) {
-      fprintf(stderr, "Invalid remote port: %s\n", serverportstring);
+      LOG_ERROR(("Invalid remote port: %s\n", serverportstring));
       return 1;
     }
     remote_addr.inet.port = PR_htons(port);
@@ -644,20 +1049,20 @@ int processConfigLine(char* configLine)
   // Configure all listen sockets and port+certificate bindings
   if (!strcmp(keyword, "listen"))
   {
-    char* hostname = strtok(NULL, ":");
+    char* hostname = strtok2(_caret, ":", &_caret);
     char* hostportstring = NULL;
     if (strcmp(hostname, "*"))
     {
       any_host_spec_config = true;
-      hostportstring = strtok(NULL, ":");
+      hostportstring = strtok2(_caret, ":", &_caret);
     }
 
-    char* serverportstring = strtok(NULL, ":");
-    char* certnick = strtok(NULL, ":");
+    char* serverportstring = strtok2(_caret, ":", &_caret);
+    char* certnick = strtok2(_caret, ":", &_caret);
 
     int port = atoi(serverportstring);
     if (port <= 0) {
-      fprintf(stderr, "Invalid port specified: %s\n", serverportstring);
+      LOG_ERROR(("Invalid port specified: %s\n", serverportstring));
       return 1;
     }
 
@@ -673,7 +1078,7 @@ int processConfigLine(char* configLine)
 
       PLHashEntry* entry = PL_HashTableAdd(existingServer->host_cert_table, hostname_copy, certnick_copy);
       if (!entry) {
-        fprintf(stderr, "Out of memory");
+        LOG_ERROR(("Out of memory"));
         return 1;
       }
     }
@@ -685,13 +1090,13 @@ int processConfigLine(char* configLine)
       server.host_cert_table = PL_NewHashTable(0, PL_HashString, PL_CompareStrings, PL_CompareStrings, NULL, NULL);
       if (!server.host_cert_table)
       {
-        fprintf(stderr, "Internal, could not create hash table\n");
+        LOG_ERROR(("Internal, could not create hash table\n"));
         return 1;
       }
       server.host_clientauth_table = PL_NewHashTable(0, PL_HashString, PL_CompareStrings, ClientAuthValueComparator, NULL, NULL);
       if (!server.host_clientauth_table)
       {
-        fprintf(stderr, "Internal, could not create hash table\n");
+        LOG_ERROR(("Internal, could not create hash table\n"));
         return 1;
       }
       servers.push_back(server);
@@ -702,22 +1107,22 @@ int processConfigLine(char* configLine)
   
   if (!strcmp(keyword, "clientauth"))
   {
-    char* hostname = strtok(NULL, ":");
-    char* hostportstring = strtok(NULL, ":");
-    char* serverportstring = strtok(NULL, ":");
+    char* hostname = strtok2(_caret, ":", &_caret);
+    char* hostportstring = strtok2(_caret, ":", &_caret);
+    char* serverportstring = strtok2(_caret, ":", &_caret);
 
     int port = atoi(serverportstring);
     if (port <= 0) {
-      fprintf(stderr, "Invalid port specified: %s\n", serverportstring);
+      LOG_ERROR(("Invalid port specified: %s\n", serverportstring));
       return 1;
     }
 
     if (server_info_t* existingServer = findServerInfo(port))
     {
-      char* authoptionstring = strtok(NULL, ":");
+      char* authoptionstring = strtok2(_caret, ":", &_caret);
       client_auth_option* authoption = new client_auth_option;
       if (!authoption) {
-        fprintf(stderr, "Out of memory");
+        LOG_ERROR(("Out of memory"));
         return 1;
       }
 
@@ -729,7 +1134,7 @@ int processConfigLine(char* configLine)
         *authoption = caNone;
       else
       {
-        fprintf(stderr, "Incorrect client auth option modifier for host '%s'", hostname);
+        LOG_ERROR(("Incorrect client auth option modifier for host '%s'", hostname));
         return 1;
       }
 
@@ -737,7 +1142,7 @@ int processConfigLine(char* configLine)
 
       char *hostname_copy = new char[strlen(hostname)+strlen(hostportstring)+2];
       if (!hostname_copy) {
-        fprintf(stderr, "Out of memory");
+        LOG_ERROR(("Out of memory"));
         return 1;
       }
 
@@ -747,13 +1152,13 @@ int processConfigLine(char* configLine)
 
       PLHashEntry* entry = PL_HashTableAdd(existingServer->host_clientauth_table, hostname_copy, authoption);
       if (!entry) {
-        fprintf(stderr, "Out of memory");
+        LOG_ERROR(("Out of memory"));
         return 1;
       }
     }
     else
     {
-      fprintf(stderr, "Server on port %d for client authentication option is not defined, use 'listen' option first", port);
+      LOG_ERROR(("Server on port %d for client authentication option is not defined, use 'listen' option first", port));
       return 1;
     }
 
@@ -763,11 +1168,11 @@ int processConfigLine(char* configLine)
   // Configure the NSS certificate database directory
   if (!strcmp(keyword, "certdbdir"))
   {
-    nssconfigdir = strtok(NULL, "\n");
+    nssconfigdir = strtok2(_caret, "\n", &_caret);
     return 0;
   }
 
-  printf("Error: keyword \"%s\" unexpected\n", keyword);
+  LOG_ERROR(("Error: keyword \"%s\" unexpected\n", keyword));
   return 1;
 }
 
@@ -801,13 +1206,13 @@ int parseConfigFile(const char* filePath)
   // Check mandatory items
   if (nssconfigdir.empty())
   {
-    printf("Error: missing path to NSS certification database\n,use certdbdir:<path> in the config file\n");
+    LOG_ERROR(("Error: missing path to NSS certification database\n,use certdbdir:<path> in the config file\n"));
     return 1;
   }
 
   if (any_host_spec_config && !do_http_proxy)
   {
-    printf("Warning: any host-specific configurations are ignored, add httpproxy:1 to allow them\n");
+    LOG_ERROR(("Warning: any host-specific configurations are ignored, add httpproxy:1 to allow them\n"));
   }
 
   return 0;
@@ -829,14 +1234,20 @@ PRIntn freeClientAuthHashItems(PLHashEntry *he, PRIntn i, void *arg)
 
 int main(int argc, char** argv)
 {
-  char* configFilePath;
+  const char* configFilePath;
+  
+  const char* logLevelEnv = PR_GetEnv("SSLTUNNEL_LOG_LEVEL");
+  gLogLevel = logLevelEnv ? (LogLevel)atoi(logLevelEnv) : LEVEL_INFO;
+  
   if (argc == 1)
     configFilePath = "ssltunnel.cfg";
   else
     configFilePath = argv[1];
 
+  memset(&websocket_server, 0, sizeof(PRNetAddr));
+
   if (parseConfigFile(configFilePath)) {
-    fprintf(stderr, "Error: config file \"%s\" missing or formating incorrect\n"
+    LOG_ERROR(("Error: config file \"%s\" missing or formating incorrect\n"
       "Specify path to the config file as parameter to ssltunnel or \n"
       "create ssltunnel.cfg in the working directory.\n\n"
       "Example format of the config file:\n\n"
@@ -861,31 +1272,32 @@ int main(int argc, char** argv)
       "       # in httpproxy mode and only after the 'listen' option has been\n"
       "       # specified. You also have to specify the tunnel listen port.\n"
       "       clientauth:requesting-client-cert.host.com:443:4443:request\n"
-      "       clientauth:requiring-client-cert.host.com:443:4443:require\n",
-      configFilePath);
+      "       clientauth:requiring-client-cert.host.com:443:4443:require\n"
+      "       # Proxy WebSocket traffic to the server at 127.0.0.1:9999,\n"
+      "       # instead of the server specified in the 'forward' option.\n"
+      "       websocketserver:127.0.0.1:9999\n",
+      configFilePath));
     return 1;
   }
 
   // create a thread pool to handle connections
-  threads = PR_CreateThreadPool(std::max<PRInt32>(INITIAL_THREADS,
-                                                  servers.size()*2),
-                                std::max<PRInt32>(MAX_THREADS,
-                                                  servers.size()*2),
+  threads = PR_CreateThreadPool(PR_MAX(INITIAL_THREADS, servers.size()*2),
+                                PR_MAX(MAX_THREADS, servers.size()*2),
                                 DEFAULT_STACKSIZE);
   if (!threads) {
-    fprintf(stderr, "Failed to create thread pool\n");
+    LOG_ERROR(("Failed to create thread pool\n"));
     return 1;
   }
 
   shutdown_lock = PR_NewLock();
   if (!shutdown_lock) {
-    fprintf(stderr, "Failed to create lock\n");
+    LOG_ERROR(("Failed to create lock\n"));
     PR_ShutdownThreadPool(threads);
     return 1;
   }
   shutdown_condvar = PR_NewCondVar(shutdown_lock);
   if (!shutdown_condvar) {
-    fprintf(stderr, "Failed to create condvar\n");
+    LOG_ERROR(("Failed to create condvar\n"));
     PR_ShutdownThreadPool(threads);
     PR_DestroyLock(shutdown_lock);
     return 1;
@@ -898,7 +1310,7 @@ int main(int argc, char** argv)
     PRInt32 errorlen = PR_GetErrorTextLength();
     char* err = new char[errorlen+1];
     PR_GetErrorText(err);
-    fprintf(stderr, "Failed to init NSS: %s", err);
+    LOG_ERROR(("Failed to init NSS: %s", err));
     delete[] err;
     PR_ShutdownThreadPool(threads);
     PR_DestroyCondVar(shutdown_condvar);
@@ -907,7 +1319,7 @@ int main(int argc, char** argv)
   }
 
   if (NSS_SetDomesticPolicy() != SECSuccess) {
-    fprintf(stderr, "NSS_SetDomesticPolicy failed\n");
+    LOG_ERROR(("NSS_SetDomesticPolicy failed\n"));
     PR_ShutdownThreadPool(threads);
     PR_DestroyCondVar(shutdown_condvar);
     PR_DestroyLock(shutdown_lock);
@@ -917,7 +1329,7 @@ int main(int argc, char** argv)
 
   // these values should make NSS use the defaults
   if (SSL_ConfigServerSessionIDCache(0, 0, 0, NULL) != SECSuccess) {
-    fprintf(stderr, "SSL_ConfigServerSessionIDCache failed\n");
+    LOG_ERROR(("SSL_ConfigServerSessionIDCache failed\n"));
     PR_ShutdownThreadPool(threads);
     PR_DestroyCondVar(shutdown_condvar);
     PR_DestroyLock(shutdown_lock);
@@ -936,14 +1348,14 @@ int main(int argc, char** argv)
   PR_WaitCondVar(shutdown_condvar, PR_INTERVAL_NO_TIMEOUT);
   PR_Unlock(shutdown_lock);
   shutdown_server = true;
-  printf("Shutting down...\n");
+  LOG_INFO(("Shutting down...\n"));
   // cleanup
   PR_ShutdownThreadPool(threads);
   PR_JoinThreadPool(threads);
   PR_DestroyCondVar(shutdown_condvar);
   PR_DestroyLock(shutdown_lock);
   if (NSS_Shutdown() == SECFailure) {
-    fprintf(stderr, "Leaked NSS objects!\n");
+    LOG_DEBUG(("Leaked NSS objects!\n"));
   }
   
   for (vector<server_info_t>::iterator it = servers.begin();
