@@ -55,8 +55,6 @@
 
 #include "nsIDOMWindow.h"
 
-#include "nsIPresShell.h"
-#include "nsPresContext.h"
 #include "nsIStringBundle.h"
 #include "nsIScriptSecurityManager.h"
 
@@ -65,6 +63,8 @@
 
 #include "nsIDOMDocument.h"
 #include "nsIDocument.h"
+#include "nsPresContext.h"
+#include "nsIAsyncVerifyRedirectCallback.h"
 
 static NS_DEFINE_CID(kThisImplCID, NS_THIS_DOCLOADER_IMPL_CID);
 
@@ -98,13 +98,22 @@ struct nsRequestInfo : public PLDHashEntryHdr
 {
   nsRequestInfo(const void *key)
     : mKey(key), mCurrentProgress(0), mMaxProgress(0), mUploading(PR_FALSE)
+   , mIsDone(PR_FALSE)
   {
   }
 
+  nsIRequest* Request() {
+    return static_cast<nsIRequest*>(const_cast<void*>(mKey));
+  }
+
   const void* mKey; // Must be first for the pldhash stubs to work
   nsInt64 mCurrentProgress;
   nsInt64 mMaxProgress;
   PRBool mUploading;
+
+  PRBool mIsDone;
+  nsString mLastStatus;
+  nsresult mLastStatusCode;
 };
 
 
@@ -117,6 +126,12 @@ RequestInfoHashInitEntry(PLDHashTable *table, PLDHashEntryHdr *entry,
   return PR_TRUE;
 }
 
+static void
+RequestInfoHashClearEntry(PLDHashTable *table, PLDHashEntryHdr *entry)
+{
+  nsRequestInfo* info = static_cast<nsRequestInfo *>(entry);
+  info->~nsRequestInfo();
+}
 
 struct nsListenerInfo {
   nsListenerInfo(nsIWeakReference *aListener, unsigned long aNotifyMask) 
@@ -138,6 +153,7 @@ nsDocLoader::nsDocLoader()
     mListenerInfoList(8),
     mIsLoadingDocument(PR_FALSE),
     mIsRestoringDocument(PR_FALSE),
+    mDontFlushLayout(PR_FALSE),
     mIsFlushingLayout(PR_FALSE)
 {
 #if defined(PR_LOGGING)
@@ -153,7 +169,7 @@ nsDocLoader::nsDocLoader()
     PL_DHashVoidPtrKeyStub,
     PL_DHashMatchEntryStub,
     PL_DHashMoveEntryStub,
-    PL_DHashClearEntryStub,
+    RequestInfoHashClearEntry,
     PL_DHashFinalizeStub,
     RequestInfoHashInitEntry
   };
@@ -312,6 +328,10 @@ nsDocLoader::Stop(void)
   if (mLoadGroup)
     rv = mLoadGroup->Cancel(NS_BINDING_ABORTED);
 
+  // Don't report that we're flushing layout so IsBusy returns false after a
+  // Stop call.
+  mIsFlushingLayout = PR_FALSE;
+
   // Clear out mChildrenInOnload.  We want to make sure to fire our
   // onload at this point, and there's no issue with mChildrenInOnload
   // after this, since mDocumentRequest will be null after the
@@ -576,6 +596,8 @@ nsDocLoader::OnStopRequest(nsIRequest *aRequest,
   //
   nsRequestInfo *info = GetRequestInfo(aRequest);
   if (info) {
+    info->mIsDone = PR_TRUE;
+
     nsInt64 oldMax = info->mMaxProgress;
 
     info->mMaxProgress = info->mCurrentProgress;
@@ -729,15 +751,26 @@ void nsDocLoader::DocLoaderIsEmpty(PRBool aFlushLayout)
 
     NS_ASSERTION(!mIsFlushingLayout, "Someone screwed up");
 
-    // The load group for this DocumentLoader is idle.  Flush layout if we need
-    // to.
-    if (aFlushLayout) {
+    // The load group for this DocumentLoader is idle.  Flush if we need to.
+    if (aFlushLayout && !mDontFlushLayout) {
       nsCOMPtr<nsIDOMDocument> domDoc = do_GetInterface(GetAsSupports(this));
       nsCOMPtr<nsIDocument> doc = do_QueryInterface(domDoc);
       if (doc) {
-        mIsFlushingLayout = PR_TRUE;
-        doc->FlushPendingNotifications(Flush_Layout);
-        mIsFlushingLayout = PR_FALSE;
+        // We start loads from style resolution, so we need to flush out style
+        // no matter what.  If we have user fonts, we also need to flush layout,
+        // since the reflow is what starts font loads.
+        mozFlushType flushType = Flush_Style;
+        nsIPresShell* shell = doc->GetShell();
+        if (shell) {
+          // Be safe in case this presshell is in teardown now
+          nsPresContext* presContext = shell->GetPresContext();
+          if (presContext && presContext->GetUserFontSet()) {
+            flushType = Flush_Layout;
+          }
+        }
+        mDontFlushLayout = mIsFlushingLayout = PR_TRUE;
+        doc->FlushPendingNotifications(flushType);
+        mDontFlushLayout = mIsFlushingLayout = PR_FALSE;
       }
     }
 
@@ -837,6 +870,24 @@ void nsDocLoader::doStartURLLoad(nsIRequest *request)
                     NS_OK);
 }
 
+// PLDHashTable enumeration callback that finds a RequestInfo that's not done
+// yet.
+static PLDHashOperator
+FindUnfinishedRequestCallback(PLDHashTable *table, PLDHashEntryHdr *hdr,
+                              PRUint32 number, void *arg)
+{
+  nsRequestInfo* info = static_cast<nsRequestInfo *>(hdr);
+  nsRequestInfo** retval = static_cast<nsRequestInfo**>(arg);
+
+  if (!info->mIsDone && !info->mLastStatus.IsEmpty()) {
+    *retval = info;
+    return PL_DHASH_STOP;
+  }
+
+  return PL_DHASH_NEXT;
+}
+
+
 void nsDocLoader::doStopURLLoad(nsIRequest *request, nsresult aStatus)
 {
 #if defined(DEBUG)
@@ -854,6 +905,17 @@ void nsDocLoader::doStopURLLoad(nsIRequest *request, nsresult aStatus)
                     nsIWebProgressListener::STATE_STOP |
                     nsIWebProgressListener::STATE_IS_REQUEST,
                     aStatus);
+
+  // Fire a status change message for a random unfinished request to make sure
+  // that the displayed status is not outdated.
+  nsRequestInfo* unfinishedRequest = nsnull;
+  PL_DHashTableEnumerate(&mRequestInfoHash, FindUnfinishedRequestCallback,
+                         &unfinishedRequest);
+  if (unfinishedRequest) {
+    FireOnStatusChange(this, unfinishedRequest->Request(),
+                       unfinishedRequest->mLastStatusCode,
+                       unfinishedRequest->mLastStatus.get());
+  }
 }
 
 void nsDocLoader::doStopDocumentLoad(nsIRequest *request,
@@ -1022,7 +1084,7 @@ NS_IMETHODIMP nsDocLoader::OnProgress(nsIRequest *aRequest, nsISupports* ctxt,
       // so update mMaxSelfProgress...  Otherwise, set it to -1 to indicate
       // that the content-length is no longer known.
       //
-      if (nsUint64(aProgressMax) != LL_MAXUINT) {
+      if (PRUint64(aProgressMax) != LL_MAXUINT) {
         mMaxSelfProgress  += PRInt64(aProgressMax);
         info->mMaxProgress = PRInt64(aProgressMax);
       } else {
@@ -1106,13 +1168,25 @@ NS_IMETHODIMP nsDocLoader::OnStatus(nsIRequest* aRequest, nsISupports* ctxt,
         info->mMaxProgress = LL_ZERO;
       }
     }
-    
-    nsresult rv;
-    nsCOMPtr<nsIStringBundleService> sbs = do_GetService(NS_STRINGBUNDLE_CONTRACTID, &rv);
-    if (NS_FAILED(rv)) return rv;
+
+    nsCOMPtr<nsIStringBundleService> sbs =
+      mozilla::services::GetStringBundleService();
+    if (!sbs)
+      return NS_ERROR_FAILURE;
     nsXPIDLString msg;
-    rv = sbs->FormatStatusMessage(aStatus, aStatusArg, getter_Copies(msg));
-    if (NS_FAILED(rv)) return rv;
+    nsresult rv = sbs->FormatStatusMessage(aStatus, aStatusArg,
+                                           getter_Copies(msg));
+    if (NS_FAILED(rv))
+      return rv;
+
+    // Keep around the message. In case a request finishes, we need to make sure
+    // to send the status message of another request to our user to that we
+    // don't display, for example, "Transferring" messages for requests that are
+    // already done.
+    if (info) {
+      info->mLastStatus = msg;
+      info->mLastStatusCode = aStatus;
+    }
     FireOnStatusChange(this, aRequest, aStatus, msg);
   }
   return NS_OK;
@@ -1239,12 +1313,13 @@ void nsDocLoader::FireOnStateChange(nsIWebProgress *aProgress,
    */
   nsCOMPtr<nsIWebProgressListener> listener;
   PRInt32 count = mListenerInfoList.Count();
+  PRInt32 notifyMask = (aStateFlags >> 16) & nsIWebProgress::NOTIFY_STATE_ALL;
 
   while (--count >= 0) {
     nsListenerInfo *info;
 
     info = static_cast<nsListenerInfo*>(mListenerInfoList.SafeElementAt(count));
-    if (!info || !(info->mNotifyMask & (aStateFlags >>16))) {
+    if (!info || !(info->mNotifyMask & notifyMask)) {
       continue;
     }
 
@@ -1510,9 +1585,10 @@ PRInt64 nsDocLoader::CalculateMaxProgress()
   return max;
 }
 
-NS_IMETHODIMP nsDocLoader::OnChannelRedirect(nsIChannel *aOldChannel,
-                                             nsIChannel *aNewChannel,
-                                             PRUint32    aFlags)
+NS_IMETHODIMP nsDocLoader::AsyncOnChannelRedirect(nsIChannel *aOldChannel,
+                                                  nsIChannel *aNewChannel,
+                                                  PRUint32 aFlags,
+                                                  nsIAsyncVerifyRedirectCallback *cb)
 {
   if (aOldChannel)
   {
@@ -1537,6 +1613,7 @@ NS_IMETHODIMP nsDocLoader::OnChannelRedirect(nsIChannel *aOldChannel,
     FireOnStateChange(this, aOldChannel, stateFlags, NS_OK);
   }
 
+  cb->OnRedirectVerifyCallback(NS_OK);
   return NS_OK;
 }
 
@@ -1696,4 +1773,3 @@ void nsDocLoader::DumpChannelInfo()
   printf("\nCurrent=%d   Total=%d\n====\n", current, max);
 }
 #endif /* 0 */
-
