@@ -41,6 +41,11 @@
 #include <sys/types.h>
 #include <sys/stat.h>
 
+#if defined(MOZ_ENABLE_CONTENTACTION)
+#include <contentaction/contentaction.h>
+#include <QString>
+#endif
+
 #include "nsOSHelperAppService.h"
 #include "nsMIMEInfoUnix.h"
 #ifdef MOZ_WIDGET_GTK2
@@ -1208,10 +1213,19 @@ nsresult nsOSHelperAppService::OSProtocolHandlerExists(const char * aProtocolSch
        aProtocolScheme));
   *aHandlerExists = PR_FALSE;
 
+#if defined(MOZ_ENABLE_CONTENTACTION)
+  // libcontentaction requires character ':' after scheme
+  ContentAction::Action action =
+    ContentAction::Action::defaultActionForScheme(QString(aProtocolScheme) + ':');
+
+  if (action.isValid())
+    *aHandlerExists = PR_TRUE;
+#endif
+
 #ifdef MOZ_WIDGET_GTK2
   // Check the GConf registry for a protocol handler
   *aHandlerExists = nsGNOMERegistry::HandlerExists(aProtocolScheme);
-#ifdef MOZ_PLATFORM_HILDON
+#if (MOZ_PLATFORM_MAEMO == 5) && defined (MOZ_ENABLE_GNOMEVFS)
   *aHandlerExists = nsMIMEInfoUnix::HandlerExists(aProtocolScheme);
 #endif
 #endif
@@ -1580,7 +1594,12 @@ nsOSHelperAppService::GetMIMEInfoFromOS(const nsACString& aType,
       return retval;
     }
 
-    // Copy the attributes of retval onto miByExt, to return it
+    // Copy the attributes of retval (mimeinfo from type) onto miByExt, to
+    // return it 
+    // but reset to just collected mDefaultAppDescription (from ext)
+    nsAutoString byExtDefault;
+    miByExt->GetDefaultDescription(byExtDefault);
+    retval->SetDefaultDescription(byExtDefault);
     retval->CopyBasicDataTo(miByExt);
 
     miByExt.swap(retval);
