@@ -126,6 +126,7 @@
 #endif
 
 #include "nsCycleCollectionParticipant.h"
+#include "nsCycleCollectorUtils.h"
 #include "nsIProgrammingLanguage.h"
 #include "nsBaseHashtable.h"
 #include "nsHashKeys.h"
@@ -138,12 +139,17 @@
 #include "prtime.h"
 #include "nsPrintfCString.h"
 #include "nsTArray.h"
+#include "mozilla/FunctionTimer.h"
 #include "nsIObserverService.h"
 #include "nsIConsoleService.h"
 #include "nsServiceManagerUtils.h"
 #include "nsThreadUtils.h"
 #include "nsTPtrArray.h"
 #include "nsTArray.h"
+#include "mozilla/Services.h"
+#include "nsICycleCollectorListener.h"
+#include "nsIXPConnect.h"
+#include "nsIJSRuntimeService.h"
 
 #include <stdio.h>
 #include <string.h>
@@ -152,6 +158,17 @@
 #include <process.h>
 #endif
 
+#ifdef XP_WIN
+#include <windows.h>
+#endif
+
+#include "mozilla/Mutex.h"
+#include "mozilla/CondVar.h"
+
+using namespace mozilla;
+
+//#define COLLECT_TIME_DEBUG
+
 #ifdef DEBUG_CC
 #define IF_DEBUG_CC_PARAM(_p) , _p
 #define IF_DEBUG_CC_ONLY_PARAM(_p) _p
@@ -167,6 +184,16 @@
 #define SHUTDOWN_COLLECTIONS(params) DEFAULT_SHUTDOWN_COLLECTIONS
 #endif
 
+#if defined(XP_WIN)
+// Defined in nsThreadManager.cpp.
+extern DWORD gTLSThreadIDIndex;
+#elif defined(NS_TLS)
+// Defined in nsThreadManager.cpp.
+extern NS_TLS mozilla::threads::ID gTLSThreadID;
+#else
+PRThread* gCycleCollectorThread = nsnull;
+#endif
+
 // Various parameters of this collector can be tuned using environment
 // variables.
 
@@ -647,11 +674,6 @@ struct GCGraph
 // Add/Remove/Has rather than PutEntry/RemoveEntry/GetEntry.
 typedef nsTHashtable<nsVoidPtrHashKey> PointerSet;
 
-#ifdef DEBUG_CC
-static void
-WriteGraph(FILE *stream, GCGraph &graph, const void *redPtr);
-#endif
-
 static inline void
 ToParticipant(nsISupports *s, nsXPCOMCycleCollectionParticipant **cp);
 
@@ -928,7 +950,13 @@ nsPurpleBuffer::SelectPointers(GCGraphBuilder &aBuilder)
 struct nsCycleCollectionXPCOMRuntime : 
     public nsCycleCollectionLanguageRuntime 
 {
-    nsresult BeginCycleCollection(nsCycleCollectionTraversalCallback &cb) 
+    nsresult BeginCycleCollection(nsCycleCollectionTraversalCallback &cb,
+                                  bool explainLiveExpectedGarbage)
+    {
+        return NS_OK;
+    }
+
+    nsresult FinishTraverse() 
     {
         return NS_OK;
     }
@@ -940,10 +968,6 @@ struct nsCycleCollectionXPCOMRuntime :
 
     inline nsCycleCollectionParticipant *ToParticipant(void *p);
 
-    void CommenceShutdown()
-    {
-    }
-
 #ifdef DEBUG_CC
     virtual void PrintAllReferencesTo(void *p) {}
 #endif
@@ -955,6 +979,8 @@ struct nsCycleCollector
     PRBool mScanInProgress;
     PRBool mFollowupCollection;
     PRUint32 mCollectedObjects;
+    PRBool mFirstCollection;
+    PRTime mCollectionStart;
 
     nsCycleCollectionLanguageRuntime *mRuntimes[nsIProgrammingLanguage::MAX+1];
     nsCycleCollectionXPCOMRuntime mXPCOMRuntime;
@@ -989,9 +1015,18 @@ struct nsCycleCollector
     nsPurpleBufferEntry* Suspect2(nsISupports *n);
     PRBool Forget2(nsPurpleBufferEntry *e);
 
-    PRUint32 Collect(PRUint32 aTryCollections = 1);
-    PRBool BeginCollection();
+    PRUint32 Collect(PRUint32 aTryCollections,
+                     nsICycleCollectorListener *aListener);
+
+    // Prepare for and cleanup after one or more collection(s).
+    PRBool PrepareForCollection(nsTPtrArray<PtrInfo> *aWhiteNodes);
+    void CleanupAfterCollection();
+
+    // Start and finish an individual collection.
+    PRBool BeginCollection(PRBool aForceGC,
+                           nsICycleCollectorListener *aListener);
     PRBool FinishCollection();
+
     PRUint32 SuspectedCount();
     void Shutdown();
 
@@ -1003,11 +1038,10 @@ struct nsCycleCollector
     }
 
 #ifdef DEBUG_CC
-    nsCycleCollectorStats mStats;    
+    nsCycleCollectorStats mStats;
 
     FILE *mPtrLog;
 
-    void MaybeDrawGraphs();
     void Allocated(void *n, size_t sz);
     void Freed(void *n);
 
@@ -1021,18 +1055,27 @@ struct nsCycleCollector
 };
 
 
+/**
+ * GraphWalker is templatized over a Visitor class that must provide
+ * the following two methods:
+ *
+ * PRBool ShouldVisitNode(PtrInfo const *pi);
+ * void VisitNode(PtrInfo *pi);
+ */
+template <class Visitor>
 class GraphWalker
 {
 private:
+    Visitor mVisitor;
+
     void DoWalk(nsDeque &aQueue);
 
 public:
     void Walk(PtrInfo *s0);
     void WalkFromRoots(GCGraph &aGraph);
-
-    // Provided by concrete walker subtypes.
-    virtual PRBool ShouldVisitNode(PtrInfo const *pi) = 0;
-    virtual void VisitNode(PtrInfo *pi) = 0;
+    // copy-constructing the visitor should be cheap, and less
+    // indirection than using a reference
+    GraphWalker(const Visitor aVisitor) : mVisitor(aVisitor) {}
 };
 
 
@@ -1090,20 +1133,6 @@ Fault(const char *msg, const void *ptr=nsnull)
         else
             printf("Fatal fault in cycle collector: %s\n", msg);
 
-     
-        if (sCollector->mGraph.mRootCount > 0) {
-            FILE *stream;
-#ifdef WIN32
-            const char fname[] = "c:\\fault-graph.dot";
-#else
-            const char fname[] = "/tmp/fault-graph.dot";
-#endif
-            printf("depositing faulting cycle-collection graph in %s\n", fname);
-            stream = fopen(fname, "w+");
-            WriteGraph(stream, sCollector->mGraph, ptr);
-            fclose(stream);
-        } 
-
         exit(1);
     }
 #endif
@@ -1158,23 +1187,23 @@ Fault(const char *msg, PtrInfo *pi)
 }
 #endif
 
-static inline bool
-CheckMainThreadIfFast()
+static inline void
+AbortIfOffMainThreadIfCheckFast()
 {
 #if defined(XP_WIN) || defined(NS_TLS)
-    return NS_IsMainThread();
-#else
-    return true;
+    if (!NS_IsMainThread() && !NS_IsCycleCollectorThread()) {
+        NS_RUNTIMEABORT("Main-thread-only object used off the main thread");
+    }
 #endif
 }
 
 static nsISupports *
 canonicalize(nsISupports *in)
 {
-    nsCOMPtr<nsISupports> child;
+    nsISupports* child;
     in->QueryInterface(NS_GET_IID(nsCycleCollectionISupports),
-                       getter_AddRefs(child));
-    return child.get();
+                       reinterpret_cast<void**>(&child));
+    return child;
 }
 
 static inline void
@@ -1202,16 +1231,18 @@ nsCycleCollectionXPCOMRuntime::ToParticipant(void *p)
 }
 
 
+template <class Visitor>
 void
-GraphWalker::Walk(PtrInfo *s0)
+GraphWalker<Visitor>::Walk(PtrInfo *s0)
 {
     nsDeque queue;
     queue.Push(s0);
     DoWalk(queue);
 }
 
+template <class Visitor>
 void
-GraphWalker::WalkFromRoots(GCGraph& aGraph)
+GraphWalker<Visitor>::WalkFromRoots(GCGraph& aGraph)
 {
     nsDeque queue;
     NodePool::Enumerator etor(aGraph.mNodes);
@@ -1221,16 +1252,17 @@ GraphWalker::WalkFromRoots(GCGraph& aGraph)
     DoWalk(queue);
 }
 
+template <class Visitor>
 void
-GraphWalker::DoWalk(nsDeque &aQueue)
+GraphWalker<Visitor>::DoWalk(nsDeque &aQueue)
 {
     // Use a aQueue to match the breadth-first traversal used when we
     // built the graph, for hopefully-better locality.
     while (aQueue.GetSize() > 0) {
         PtrInfo *pi = static_cast<PtrInfo*>(aQueue.PopFront());
 
-        if (this->ShouldVisitNode(pi)) {
-            this->VisitNode(pi);
+        if (mVisitor.ShouldVisitNode(pi)) {
+            mVisitor.VisitNode(pi);
             for (EdgePool::Iterator child = pi->mFirstChild,
                                 child_end = pi->mLastChild;
                  child != child_end; ++child) {
@@ -1245,6 +1277,95 @@ GraphWalker::DoWalk(nsDeque &aQueue)
 }
 
 
+class nsCycleCollectorLogger : public nsICycleCollectorListener
+{
+public:
+    nsCycleCollectorLogger() : mStream(nsnull)
+    {
+    }
+    ~nsCycleCollectorLogger()
+    {
+        if (mStream) {
+            fclose(mStream);
+        }
+    }
+    NS_DECL_ISUPPORTS
+
+    NS_IMETHOD Begin()
+    {
+        char name[255];
+        sprintf(name, "cc-edges-%d.log", ++gLogCounter);
+        mStream = fopen(name, "w");
+
+        return mStream ? NS_OK : NS_ERROR_FAILURE;
+    }
+    NS_IMETHOD NoteObject(PRUint64 aAddress, const char *aObjectDescription)
+    {
+        fprintf(mStream, "%p %s\n", (void*)aAddress, aObjectDescription);
+
+        return NS_OK;
+    }
+    NS_IMETHOD NoteEdge(PRUint64 aFromAddress, PRUint64 aToAddress,
+                        const char *aEdgeName)
+    {
+        fprintf(mStream, "> %p %s\n", (void*)aToAddress, aEdgeName);
+
+        return NS_OK;
+    }
+    NS_IMETHOD BeginDescriptions()
+    {
+        fputs("==========\n", mStream);
+
+        return NS_OK;
+    }
+    NS_IMETHOD DescribeRefcountedObject(PRUint64 aAddress, PRUint32 aKnownEdges,
+                                        PRUint32 aTotalEdges)
+    {
+        PRBool root = aKnownEdges != aTotalEdges;
+        fprintf(mStream, "%p", (void*)aAddress);
+        if (root) {
+            fprintf(mStream, " [root] [%u/%u]", aKnownEdges, aTotalEdges);
+        }
+        fputc('\n', mStream);
+
+        return NS_OK;
+    }
+    NS_IMETHOD DescribeGCedObject(PRUint64 aAddress, PRBool aMarked)
+    {
+        fprintf(mStream, "%p%s\n", (void*)aAddress, aMarked ? " [root]" : "");
+
+        return NS_OK;
+    }
+    NS_IMETHOD End()
+    {
+        fclose(mStream);
+        mStream = nsnull;
+
+        return NS_OK;
+    }
+
+private:
+    FILE *mStream;
+
+    static PRUint32 gLogCounter;
+};
+
+NS_IMPL_ISUPPORTS1(nsCycleCollectorLogger, nsICycleCollectorListener)
+
+PRUint32 nsCycleCollectorLogger::gLogCounter = 0;
+
+nsresult
+nsCycleCollectorLoggerConstructor(nsISupports* aOuter,
+                                  const nsIID& aIID,
+                                  void* *aInstancePtr)
+{
+    NS_ENSURE_TRUE(!aOuter, NS_ERROR_NO_AGGREGATION);
+
+    nsISupports *logger = new nsCycleCollectorLogger();
+
+    return logger->QueryInterface(aIID, aInstancePtr);
+}
+
 ////////////////////////////////////////////////////////////////////////
 // Bacon & Rajan's |MarkRoots| routine.
 ////////////////////////////////////////////////////////////////////////
@@ -1283,14 +1404,15 @@ private:
     PLDHashTable mPtrToNodeMap;
     PtrInfo *mCurrPi;
     nsCycleCollectionLanguageRuntime **mRuntimes; // weak, from nsCycleCollector
-#ifdef DEBUG_CC
     nsCString mNextEdgeName;
-#endif
+    nsICycleCollectorListener *mListener;
 
 public:
     GCGraphBuilder(GCGraph &aGraph,
-                   nsCycleCollectionLanguageRuntime **aRuntimes);
+                   nsCycleCollectionLanguageRuntime **aRuntimes,
+                   nsICycleCollectorListener *aListener);
     ~GCGraphBuilder();
+    bool Initialized();
 
     PRUint32 Count() const { return mPtrToNodeMap.entryCount; }
 
@@ -1323,19 +1445,25 @@ private:
 };
 
 GCGraphBuilder::GCGraphBuilder(GCGraph &aGraph,
-                               nsCycleCollectionLanguageRuntime **aRuntimes)
+                               nsCycleCollectionLanguageRuntime **aRuntimes,
+                               nsICycleCollectorListener *aListener)
     : mNodeBuilder(aGraph.mNodes),
       mEdgeBuilder(aGraph.mEdges),
-      mRuntimes(aRuntimes)
+      mRuntimes(aRuntimes),
+      mListener(aListener)
 {
     if (!PL_DHashTableInit(&mPtrToNodeMap, &PtrNodeOps, nsnull,
                            sizeof(PtrToNodeEntry), 32768))
         mPtrToNodeMap.ops = nsnull;
-#ifdef DEBUG_CC
-    // Do we need to set these all the time?
-    mFlags |= nsCycleCollectionTraversalCallback::WANT_DEBUG_INFO |
-              nsCycleCollectionTraversalCallback::WANT_ALL_TRACES;
+    // We want all edges and all info if DEBUG_CC is set or if we have a
+    // listener. Do we want them all the time?
+#ifndef DEBUG_CC
+    if (mListener)
 #endif
+    {
+        mFlags |= nsCycleCollectionTraversalCallback::WANT_DEBUG_INFO |
+                  nsCycleCollectionTraversalCallback::WANT_ALL_TRACES;
+    }
 }
 
 GCGraphBuilder::~GCGraphBuilder()
@@ -1344,6 +1472,12 @@ GCGraphBuilder::~GCGraphBuilder()
         PL_DHashTableFinish(&mPtrToNodeMap);
 }
 
+bool
+GCGraphBuilder::Initialized()
+{
+    return !!mPtrToNodeMap.ops;
+}
+
 PtrInfo*
 GCGraphBuilder::AddNode(void *s, nsCycleCollectionParticipant *aParticipant
                         IF_DEBUG_CC_PARAM(PRUint32 aLangID)
@@ -1436,6 +1570,10 @@ GCGraphBuilder::DescribeNode(CCNodeType type, nsrefcnt refCount,
     mCurrPi->mName = PL_strdup(objName);
 #endif
 
+    if (mListener) {
+        mListener->NoteObject((PRUint64)mCurrPi->mPointer, objName);
+    }
+
     if (type == RefCounted) {
         if (refCount == 0 || refCount == PR_UINT32_MAX)
             Fault("zero or overflowing refcount", mCurrPi);
@@ -1453,10 +1591,11 @@ GCGraphBuilder::DescribeNode(CCNodeType type, nsrefcnt refCount,
 NS_IMETHODIMP_(void)
 GCGraphBuilder::NoteXPCOMChild(nsISupports *child) 
 {
-#ifdef DEBUG_CC
-    nsCString edgeName(mNextEdgeName);
-    mNextEdgeName.Truncate();
-#endif
+    nsCString edgeName;
+    if (WantDebugInfo()) {
+        edgeName.Assign(mNextEdgeName);
+        mNextEdgeName.Truncate();
+    }
     if (!child || !(child = canonicalize(child)))
         return; 
 
@@ -1475,6 +1614,10 @@ GCGraphBuilder::NoteXPCOMChild(nsISupports *child)
 #ifdef DEBUG_CC
         mCurrPi->mEdgeNames.AppendElement(edgeName);
 #endif
+        if (mListener) {
+            mListener->NoteEdge((PRUint64)mCurrPi->mPointer, (PRUint64)child,
+                                edgeName.get());
+        }
         ++childPi->mInternalRefs;
     }
 }
@@ -1483,10 +1626,11 @@ NS_IMETHODIMP_(void)
 GCGraphBuilder::NoteNativeChild(void *child,
                                 nsCycleCollectionParticipant *participant)
 {
-#ifdef DEBUG_CC
-    nsCString edgeName(mNextEdgeName);
-    mNextEdgeName.Truncate();
-#endif
+    nsCString edgeName;
+    if (WantDebugInfo()) {
+        edgeName.Assign(mNextEdgeName);
+        mNextEdgeName.Truncate();
+    }
     if (!child)
         return;
 
@@ -1499,16 +1643,21 @@ GCGraphBuilder::NoteNativeChild(void *child,
 #ifdef DEBUG_CC
     mCurrPi->mEdgeNames.AppendElement(edgeName);
 #endif
+    if (mListener) {
+        mListener->NoteEdge((PRUint64)mCurrPi->mPointer, (PRUint64)child,
+                            edgeName.get());
+    }
     ++childPi->mInternalRefs;
 }
 
 NS_IMETHODIMP_(void)
 GCGraphBuilder::NoteScriptChild(PRUint32 langID, void *child) 
 {
-#ifdef DEBUG_CC
-    nsCString edgeName(mNextEdgeName);
-    mNextEdgeName.Truncate();
-#endif
+    nsCString edgeName;
+    if (WantDebugInfo()) {
+        edgeName.Assign(mNextEdgeName);
+        mNextEdgeName.Truncate();
+    }
     if (!child)
         return;
 
@@ -1534,15 +1683,19 @@ GCGraphBuilder::NoteScriptChild(PRUint32 langID, void *child)
 #ifdef DEBUG_CC
     mCurrPi->mEdgeNames.AppendElement(edgeName);
 #endif
+    if (mListener) {
+        mListener->NoteEdge((PRUint64)mCurrPi->mPointer, (PRUint64)child,
+                            edgeName.get());
+    }
     ++childPi->mInternalRefs;
 }
 
 NS_IMETHODIMP_(void)
 GCGraphBuilder::NoteNextEdgeName(const char* name)
 {
-#ifdef DEBUG_CC
-    mNextEdgeName = name;
-#endif
+    if (WantDebugInfo()) {
+        mNextEdgeName = name;
+    }
 }
 
 static PRBool
@@ -1618,9 +1771,10 @@ nsCycleCollector::MarkRoots(GCGraphBuilder &builder)
 ////////////////////////////////////////////////////////////////////////
 
 
-struct ScanBlackWalker : public GraphWalker
+struct ScanBlackVisitor
 {
-    ScanBlackWalker(PRUint32 &aWhiteNodeCount) : mWhiteNodeCount(aWhiteNodeCount)
+    ScanBlackVisitor(PRUint32 &aWhiteNodeCount)
+        : mWhiteNodeCount(aWhiteNodeCount)
     {
     }
 
@@ -1643,9 +1797,9 @@ struct ScanBlackWalker : public GraphWalker
 };
 
 
-struct scanWalker : public GraphWalker
+struct scanVisitor
 {
-    scanWalker(PRUint32 &aWhiteNodeCount) : mWhiteNodeCount(aWhiteNodeCount)
+    scanVisitor(PRUint32 &aWhiteNodeCount) : mWhiteNodeCount(aWhiteNodeCount)
     {
     }
 
@@ -1666,9 +1820,9 @@ struct scanWalker : public GraphWalker
             sCollector->mStats.mSetColorWhite++;
 #endif
         } else {
-            ScanBlackWalker(mWhiteNodeCount).Walk(pi);
+            GraphWalker<ScanBlackVisitor>(ScanBlackVisitor(mWhiteNodeCount)).Walk(pi);
             NS_ASSERTION(pi->mColor == black,
-                         "Why didn't ScanBlackWalker make pi black?");
+                         "Why didn't ScanBlackVisitor make pi black?");
         }
     }
 
@@ -1683,7 +1837,7 @@ nsCycleCollector::ScanRoots()
     // On the assumption that most nodes will be black, it's
     // probably faster to use a GraphWalker than a
     // NodePool::Enumerator.
-    scanWalker(mWhiteNodeCount).WalkFromRoots(mGraph); 
+    GraphWalker<scanVisitor>(scanVisitor(mWhiteNodeCount)).WalkFromRoots(mGraph); 
 
 #ifdef DEBUG_CC
     // Sanity check: scan should have colored all grey nodes black or
@@ -1989,6 +2143,7 @@ nsCycleCollector::nsCycleCollector() :
     mCollectionInProgress(PR_FALSE),
     mScanInProgress(PR_FALSE),
     mCollectedObjects(0),
+    mFirstCollection(PR_TRUE),
     mWhiteNodes(nsnull),
     mWhiteNodeCount(0),
 #ifdef DEBUG_CC
@@ -2052,91 +2207,7 @@ nsCycleCollector::ForgetRuntime(PRUint32 langID)
     mRuntimes[langID] = nsnull;
 }
 
-
 #ifdef DEBUG_CC
-static void
-WriteGraph(FILE *stream, GCGraph &graph, const void *redPtr)
-{
-    fprintf(stream, 
-            "digraph collection {\n"
-            "rankdir=LR\n"
-            "node [fontname=fixed, fontsize=10, style=filled, shape=box]\n"
-            );
-    
-    NodePool::Enumerator etor(graph.mNodes);
-    while (!etor.IsDone()) {
-        PtrInfo *pi = etor.GetNext();
-        const void *p = pi->mPointer;
-        fprintf(stream, 
-                "n%p [label=\"%s\\n%p\\n",
-                p,
-                pi->mName,
-                p);
-        if (pi->mRefCount != 0 && pi->mRefCount != PR_UINT32_MAX) {
-            fprintf(stream, 
-                    "%u/%u refs found",
-                    pi->mInternalRefs, pi->mRefCount);
-        }
-        fprintf(stream, 
-                "\", fillcolor=%s, fontcolor=%s]\n", 
-                (redPtr && redPtr == p ? "red" : (pi->mColor == black ? "black" : "white")),
-                (pi->mColor == black ? "white" : "black"));
-        for (EdgePool::Iterator child = pi->mFirstChild,
-                 child_end = pi->mLastChild;
-             child != child_end; ++child) {
-            fprintf(stream, "n%p -> n%p\n", p, (*child)->mPointer);
-        }
-    }
-    
-    fprintf(stream, "\n}\n");    
-}
-
-
-void 
-nsCycleCollector::MaybeDrawGraphs()
-{
-    if (mParams.mDrawGraphs) {
-        // We draw graphs only if there were any white nodes.
-        PRBool anyWhites = PR_FALSE;
-        NodePool::Enumerator fwetor(mGraph.mNodes);
-        while (!fwetor.IsDone())
-        {
-            PtrInfo *pinfo = fwetor.GetNext();
-            if (pinfo->mColor == white) {
-                anyWhites = PR_TRUE;
-                break;
-            }
-        }
-
-        if (anyWhites) {
-            // We can't just use _popen here because graphviz-for-windows
-            // doesn't set up its stdin stream properly, sigh.
-            FILE *stream;
-#ifdef WIN32
-            stream = fopen("c:\\cycle-graph.dot", "w+");
-#else
-            stream = popen("dotty -", "w");
-#endif
-            WriteGraph(stream, mGraph, nsnull);
-#ifdef WIN32
-            fclose(stream);
-            // Even dotty doesn't work terribly well on windows, since
-            // they execute lefty asynchronously. So we'll just run 
-            // lefty ourselves.
-            _spawnlp(_P_WAIT, 
-                     "lefty", 
-                     "lefty",
-                     "-e",
-                     "\"load('dotty.lefty');"
-                     "dotty.simple('c:\\cycle-graph.dot');\"",
-                     NULL);
-            unlink("c:\\cycle-graph.dot");
-#else
-            pclose(stream);
-#endif
-        }
-    }
-}
 
 class Suppressor :
     public nsCycleCollectionTraversalCallback
@@ -2215,8 +2286,7 @@ nsCycleCollector_isScanSafe(nsISupports *s)
 PRBool
 nsCycleCollector::Suspect(nsISupports *n)
 {
-    if (!CheckMainThreadIfFast())
-        return PR_FALSE;
+    AbortIfOffMainThreadIfCheckFast();
 
     // Re-entering ::Suspect during collection used to be a fault, but
     // we are canonicalizing nsISupports pointers using QI, so we will
@@ -2256,12 +2326,7 @@ nsCycleCollector::Suspect(nsISupports *n)
 PRBool
 nsCycleCollector::Forget(nsISupports *n)
 {
-    if (!CheckMainThreadIfFast()) {
-        if (!mParams.mDoNothing) {
-            Fault("Forget called off main thread");
-        }
-        return PR_TRUE; // it's as good as forgotten
-    }
+    AbortIfOffMainThreadIfCheckFast();
 
     // Re-entering ::Forget during collection used to be a fault, but
     // we are canonicalizing nsISupports pointers using QI, so we will
@@ -2295,8 +2360,7 @@ nsCycleCollector::Forget(nsISupports *n)
 nsPurpleBufferEntry*
 nsCycleCollector::Suspect2(nsISupports *n)
 {
-    if (!CheckMainThreadIfFast())
-        return nsnull;
+    AbortIfOffMainThreadIfCheckFast();
 
     // Re-entering ::Suspect during collection used to be a fault, but
     // we are canonicalizing nsISupports pointers using QI, so we will
@@ -2337,8 +2401,7 @@ nsCycleCollector::Suspect2(nsISupports *n)
 PRBool
 nsCycleCollector::Forget2(nsPurpleBufferEntry *e)
 {
-    if (!CheckMainThreadIfFast())
-        return PR_FALSE;
+    AbortIfOffMainThreadIfCheckFast();
 
     // Re-entering ::Forget during collection used to be a fault, but
     // we are canonicalizing nsISupports pointers using QI, so we will
@@ -2396,8 +2459,8 @@ nsCycleCollector::Freed(void *n)
 }
 #endif
 
-PRUint32
-nsCycleCollector::Collect(PRUint32 aTryCollections)
+PRBool
+nsCycleCollector::PrepareForCollection(nsTPtrArray<PtrInfo> *aWhiteNodes)
 {
 #if defined(DEBUG_CC) && !defined(__MINGW32__)
     if (!mParams.mDoNothing && mParams.mHookMalloc)
@@ -2406,68 +2469,34 @@ nsCycleCollector::Collect(PRUint32 aTryCollections)
 
     // This can legitimately happen in a few cases. See bug 383651.
     if (mCollectionInProgress)
-        return 0;
+        return PR_FALSE;
+
+    NS_TIME_FUNCTION;
 
 #ifdef COLLECT_TIME_DEBUG
-    printf("cc: Starting nsCycleCollector::Collect(%d)\n", aTryCollections);
-    PRTime start = PR_Now();
+    printf("cc: nsCycleCollector::PrepareForCollection()\n");
+    mCollectionStart = PR_Now();
 #endif
 
     mCollectionInProgress = PR_TRUE;
 
     nsCOMPtr<nsIObserverService> obs =
-      do_GetService("@mozilla.org/observer-service;1");
-    if (obs) {
+        mozilla::services::GetObserverService();
+    if (obs)
         obs->NotifyObservers(nsnull, "cycle-collector-begin", nsnull);
-    }
 
     mFollowupCollection = PR_FALSE;
     mCollectedObjects = 0;
-    nsAutoTPtrArray<PtrInfo, 4000> whiteNodes;
-    mWhiteNodes = &whiteNodes;
 
-    PRUint32 totalCollections = 0;
-    while (aTryCollections > totalCollections) {
-        PRBool collected;
-        if (mRuntimes[nsIProgrammingLanguage::JAVASCRIPT]) {
-            collected = static_cast<nsCycleCollectionJSRuntime*>
-                (mRuntimes[nsIProgrammingLanguage::JAVASCRIPT])->Collect();
-        }
-        else {
-            collected = BeginCollection() && FinishCollection();
-        }
+    mWhiteNodes = aWhiteNodes;
 
-#ifdef DEBUG_CC
-        // We wait until after FinishCollection to check the white nodes because
-        // some objects may outlive CollectWhite but then be freed by
-        // FinishCycleCollection (like XPConnect's deferred release of native
-        // objects).
-        PRUint32 i, count = mWhiteNodes->Length();
-        for (i = 0; i < count; ++i) {
-            PtrInfo *pinfo = mWhiteNodes->ElementAt(i);
-            if (pinfo->mLangID == nsIProgrammingLanguage::CPLUSPLUS &&
-                mPurpleBuf.Exists(pinfo->mPointer)) {
-                printf("nsCycleCollector: %s object @%p is still alive after\n"
-                       "  calling RootAndUnlinkJSObjects, Unlink, and Unroot on"
-                       " it!  This probably\n"
-                       "  means the Unlink implementation was insufficient.\n",
-                       pinfo->mName, pinfo->mPointer);
-            }
-        }
-#endif
-        mWhiteNodes->Clear();
-        ClearGraph();
-
-        mParams.mDoNothing = PR_FALSE;
-
-        if (!collected)
-            break;
-        
-        ++totalCollections;
-    }
+    return PR_TRUE;
+}
 
+void
+nsCycleCollector::CleanupAfterCollection()
+{
     mWhiteNodes = nsnull;
-
     mCollectionInProgress = PR_FALSE;
 
 #ifdef XP_OS2
@@ -2478,29 +2507,79 @@ nsCycleCollector::Collect(PRUint32 aTryCollections)
 #endif
 
 #ifdef COLLECT_TIME_DEBUG
-    printf("cc: Collect() took %lldms\n",
-           (PR_Now() - start) / PR_USEC_PER_MSEC);
+    printf("cc: CleanupAfterCollection(), total time %lldms\n",
+           (PR_Now() - mCollectionStart) / PR_USEC_PER_MSEC);
 #endif
 #ifdef DEBUG_CC
     ExplainLiveExpectedGarbage();
 #endif
+}
+
+PRUint32
+nsCycleCollector::Collect(PRUint32 aTryCollections,
+                          nsICycleCollectorListener *aListener)
+{
+    nsAutoTPtrArray<PtrInfo, 4000> whiteNodes;
+
+    if (!PrepareForCollection(&whiteNodes))
+        return 0;
+
+    PRUint32 totalCollections = 0;
+    while (aTryCollections > totalCollections) {
+        // Synchronous cycle collection. Always force a JS GC as well.
+        if (!(BeginCollection(PR_TRUE, aListener) && FinishCollection()))
+            break;
+
+        ++totalCollections;
+    }
+
+    CleanupAfterCollection();
+
     return mCollectedObjects;
 }
 
 PRBool
-nsCycleCollector::BeginCollection()
+nsCycleCollector::BeginCollection(PRBool aForceGC,
+                                  nsICycleCollectorListener *aListener)
 {
     if (mParams.mDoNothing)
         return PR_FALSE;
 
-    GCGraphBuilder builder(mGraph, mRuntimes);
+    // The cycle collector uses the mark bitmap to discover what JS objects
+    // were reachable only from XPConnect roots that might participate in
+    // cycles. If this is the first cycle collection after startup force
+    // a garbage collection, otherwise the GC might not have run yet and
+    // the bitmap is invalid.
+    if (mFirstCollection && mRuntimes[nsIProgrammingLanguage::JAVASCRIPT]) {
+        aForceGC = PR_TRUE;
+        mFirstCollection = PR_FALSE;
+    }
+
+    if (aForceGC) {
+#ifdef COLLECT_TIME_DEBUG
+        PRTime start = PR_Now();
+#endif
+        static_cast<nsCycleCollectionJSRuntime*>
+            (mRuntimes[nsIProgrammingLanguage::JAVASCRIPT])->Collect();
+#ifdef COLLECT_TIME_DEBUG
+        printf("cc: GC() took %lldms\n", (PR_Now() - start) / PR_USEC_PER_MSEC);
+#endif
+    }
+
+    if (aListener && NS_FAILED(aListener->Begin())) {
+        aListener = nsnull;
+    }
+
+    GCGraphBuilder builder(mGraph, mRuntimes, aListener);
+    if (!builder.Initialized())
+        return PR_FALSE;
 
 #ifdef COLLECT_TIME_DEBUG
     PRTime now = PR_Now();
 #endif
     for (PRUint32 i = 0; i <= nsIProgrammingLanguage::MAX; ++i) {
         if (mRuntimes[i])
-            mRuntimes[i]->BeginCycleCollection(builder);
+            mRuntimes[i]->BeginCycleCollection(builder, PR_FALSE);
     }
 
 #ifdef COLLECT_TIME_DEBUG
@@ -2569,12 +2648,29 @@ nsCycleCollector::BeginCollection()
                (PR_Now() - now) / PR_USEC_PER_MSEC);
 #endif
 
-#ifdef DEBUG_CC
-        MaybeDrawGraphs();
-#endif
-
         mScanInProgress = PR_FALSE;
 
+        if (aListener) {
+            aListener->BeginDescriptions();
+
+            NodePool::Enumerator etor(mGraph.mNodes);
+            while (!etor.IsDone()) {
+                PtrInfo *pi = etor.GetNext();
+                if (pi->mColor == black) {
+                    PRUint64 p = (PRUint64)pi->mPointer;
+                    if (pi->mRefCount > 0 && pi->mRefCount < PR_UINT32_MAX) {
+                        aListener->DescribeRefcountedObject(p, pi->mInternalRefs,
+                                                           pi->mRefCount);
+                    }
+                    else {
+                        aListener->DescribeGCedObject(p, pi->mRefCount != 0);
+                    }
+                }
+            }
+
+            aListener->End();
+        }
+
 #ifdef DEBUG_CC
         if (mFollowupCollection && purpleStart != purpleEnd) {
             PRUint32 i = 0;
@@ -2594,15 +2690,10 @@ nsCycleCollector::BeginCollection()
         }
 #endif
 
-#ifdef COLLECT_TIME_DEBUG
-        now = PR_Now();
-#endif
-        RootWhite();
-
-#ifdef COLLECT_TIME_DEBUG
-        printf("cc: RootWhite() took %lldms\n",
-               (PR_Now() - now) / PR_USEC_PER_MSEC);
-#endif
+        for (PRUint32 i = 0; i <= nsIProgrammingLanguage::MAX; ++i) {
+            if (mRuntimes[i])
+                mRuntimes[i]->FinishTraverse();
+        }
     }
     else {
         mScanInProgress = PR_FALSE;
@@ -2617,6 +2708,15 @@ nsCycleCollector::FinishCollection()
 #ifdef COLLECT_TIME_DEBUG
     PRTime now = PR_Now();
 #endif
+
+    RootWhite();
+
+#ifdef COLLECT_TIME_DEBUG
+    printf("cc: RootWhite() took %lldms\n",
+           (PR_Now() - now) / PR_USEC_PER_MSEC);
+    now = PR_Now();
+#endif
+
     PRBool collected = CollectWhite();
 
 #ifdef COLLECT_TIME_DEBUG
@@ -2637,6 +2737,30 @@ nsCycleCollector::FinishCollection()
 
     mFollowupCollection = PR_TRUE;
 
+#ifdef DEBUG_CC
+    // We wait until after FinishCollection to check the white nodes because
+    // some objects may outlive CollectWhite but then be freed by
+    // FinishCycleCollection (like XPConnect's deferred release of native
+    // objects).
+    PRUint32 i, count = mWhiteNodes->Length();
+    for (i = 0; i < count; ++i) {
+        PtrInfo *pinfo = mWhiteNodes->ElementAt(i);
+        if (pinfo->mLangID == nsIProgrammingLanguage::CPLUSPLUS &&
+            mPurpleBuf.Exists(pinfo->mPointer)) {
+            printf("nsCycleCollector: %s object @%p is still alive after\n"
+                   "  calling RootAndUnlinkJSObjects, Unlink, and Unroot on"
+                   " it!  This probably\n"
+                   "  means the Unlink implementation was insufficient.\n",
+                   pinfo->mName, pinfo->mPointer);
+        }
+    }
+#endif
+
+    mWhiteNodes->Clear();
+    ClearGraph();
+
+    mParams.mDoNothing = PR_FALSE;
+
     return collected;
 }
 
@@ -2652,15 +2776,10 @@ nsCycleCollector::Shutdown()
     // Here we want to run a final collection and then permanently
     // disable the collector because the program is shutting down.
 
-    for (PRUint32 i = 0; i <= nsIProgrammingLanguage::MAX; ++i) {
-        if (mRuntimes[i])
-            mRuntimes[i]->CommenceShutdown();
-    }
-
-    Collect(SHUTDOWN_COLLECTIONS(mParams));
+    Collect(SHUTDOWN_COLLECTIONS(mParams), nsnull);
 
 #ifdef DEBUG_CC
-    GCGraphBuilder builder(mGraph, mRuntimes);
+    GCGraphBuilder builder(mGraph, mRuntimes, nsnull);
     mScanInProgress = PR_TRUE;
     SelectPurple(builder);
     mScanInProgress = PR_FALSE;
@@ -2685,16 +2804,16 @@ AddExpectedGarbage(nsVoidPtrHashKey *p, void *arg)
     return PL_DHASH_NEXT;
 }
 
-struct SetSCCWalker : public GraphWalker
+struct SetSCCVisitor
 {
-    SetSCCWalker(PRUint32 aIndex) : mIndex(aIndex) {}
+    SetSCCVisitor(PRUint32 aIndex) : mIndex(aIndex) {}
     PRBool ShouldVisitNode(PtrInfo const *pi) { return pi->mSCCIndex == 0; }
     void VisitNode(PtrInfo *pi) { pi->mSCCIndex = mIndex; }
 private:
     PRUint32 mIndex;
 };
 
-struct SetNonRootGreyWalker : public GraphWalker
+struct SetNonRootGreyVisitor
 {
     PRBool ShouldVisitNode(PtrInfo const *pi) { return pi->mColor == white; }
     void VisitNode(PtrInfo *pi) { pi->mColor = grey; }
@@ -2735,7 +2854,7 @@ nsCycleCollector::ExplainLiveExpectedGarbage()
     mScanInProgress = PR_TRUE;
 
     {
-        GCGraphBuilder builder(mGraph, mRuntimes);
+        GCGraphBuilder builder(mGraph, mRuntimes, nsnull);
 
         // Instead of adding roots from the purple buffer, we add them
         // from the list of nodes we were expected to collect.
@@ -2749,7 +2868,7 @@ nsCycleCollector::ExplainLiveExpectedGarbage()
 
         for (PRUint32 i = 0; i <= nsIProgrammingLanguage::MAX; ++i) {
             if (mRuntimes[i])
-                mRuntimes[i]->BeginCycleCollection(builder);
+                mRuntimes[i]->BeginCycleCollection(builder, PR_TRUE);
         }
 
         // But just for extra information, add entries from the purple
@@ -2762,6 +2881,12 @@ nsCycleCollector::ExplainLiveExpectedGarbage()
 
         mScanInProgress = PR_FALSE;
 
+        for (PRUint32 i = 0; i <= nsIProgrammingLanguage::MAX; ++i) {
+            if (mRuntimes[i]) {
+                mRuntimes[i]->FinishTraverse();
+            }
+        }
+
         PRBool describeExtraRefcounts = PR_FALSE;
         PRBool findCycleRoots = PR_FALSE;
         {
@@ -2930,7 +3055,7 @@ nsCycleCollector::ExplainLiveExpectedGarbage()
                     PRUint32 currentSCC = 1;
 
                     while (DFSPostOrder.GetSize() > 0) {
-                        SetSCCWalker(currentSCC).Walk((PtrInfo*)DFSPostOrder.PopFront());
+                        GraphWalker<SetSCCVisitor>(SetSCCVisitor(currentSCC)).Walk((PtrInfo*)DFSPostOrder.PopFront());
                         ++currentSCC;
                     }
                 }
@@ -2947,7 +3072,7 @@ nsCycleCollector::ExplainLiveExpectedGarbage()
                                             child_end = pi->mLastChild;
                              child != child_end; ++child) {
                             if ((*child)->mSCCIndex != pi->mSCCIndex) {
-                                SetNonRootGreyWalker().Walk(*child);
+                                GraphWalker<SetNonRootGreyVisitor>(SetNonRootGreyVisitor()).Walk(*child);
                             }
                         }
                     }
@@ -3055,13 +3180,17 @@ nsCycleCollector::DestroyReversedEdges()
 void
 nsCycleCollector::ShouldBeFreed(nsISupports *n)
 {
-    mExpectedGarbage.PutEntry(n);
+    if (n) {
+        mExpectedGarbage.PutEntry(n);
+    }
 }
 
 void
 nsCycleCollector::WasFreed(nsISupports *n)
 {
-    mExpectedGarbage.RemoveEntry(n);
+    if (n) {
+        mExpectedGarbage.RemoveEntry(n);
+    }
 }
 #endif
 
@@ -3122,50 +3251,12 @@ NS_CycleCollectorForget2(nsPurpleBufferEntry *e)
     return sCollector ? sCollector->Forget2(e) : PR_TRUE;
 }
 
-
-PRUint32
-nsCycleCollector_collect()
-{
-    return sCollector ? sCollector->Collect() : 0;
-}
-
 PRUint32
 nsCycleCollector_suspectedCount()
 {
     return sCollector ? sCollector->SuspectedCount() : 0;
 }
 
-PRBool 
-nsCycleCollector_beginCollection()
-{
-    return sCollector ? sCollector->BeginCollection() : PR_FALSE;
-}
-
-PRBool 
-nsCycleCollector_finishCollection()
-{
-    return sCollector ? sCollector->FinishCollection() : PR_FALSE;
-}
-
-nsresult 
-nsCycleCollector_startup()
-{
-    NS_ASSERTION(!sCollector, "Forgot to call nsCycleCollector_shutdown?");
-
-    sCollector = new nsCycleCollector();
-    return sCollector ? NS_OK : NS_ERROR_OUT_OF_MEMORY;
-}
-
-void 
-nsCycleCollector_shutdown()
-{
-    if (sCollector) {
-        sCollector->Shutdown();
-        delete sCollector;
-        sCollector = nsnull;
-    }
-}
-
 #ifdef DEBUG
 void
 nsCycleCollector_DEBUG_shouldBeFreed(nsISupports *n)
@@ -3185,3 +3276,188 @@ nsCycleCollector_DEBUG_wasFreed(nsISupports *n)
 #endif
 }
 #endif
+
+class nsCycleCollectorRunner : public nsRunnable
+{
+    nsCycleCollector *mCollector;
+    nsICycleCollectorListener *mListener;
+    Mutex mLock;
+    CondVar mRequest;
+    CondVar mReply;
+    PRBool mRunning;
+    PRBool mShutdown;
+    PRBool mCollected;
+
+public:
+    NS_IMETHOD Run()
+    {
+#ifdef XP_WIN
+        TlsSetValue(gTLSThreadIDIndex,
+                    (void*) mozilla::threads::CycleCollector);
+#elif defined(NS_TLS)
+        gTLSThreadID = mozilla::threads::CycleCollector;
+#else
+        gCycleCollectorThread = PR_GetCurrentThread();
+#endif
+
+        NS_ASSERTION(NS_IsCycleCollectorThread() && !NS_IsMainThread(),
+                     "Wrong thread!");
+
+        MutexAutoLock autoLock(mLock);
+
+        if (mShutdown)
+            return NS_OK;
+
+        mRunning = PR_TRUE;
+
+        while (1) {
+            mRequest.Wait();
+
+            if (!mRunning) {
+                mReply.Notify();
+                return NS_OK;
+            }
+
+            mCollected = mCollector->BeginCollection(PR_FALSE, mListener);
+
+            mReply.Notify();
+        }
+
+        return NS_OK;
+    }
+
+    nsCycleCollectorRunner(nsCycleCollector *collector)
+        : mCollector(collector),
+          mListener(nsnull),
+          mLock("cycle collector lock"),
+          mRequest(mLock, "cycle collector request condvar"),
+          mReply(mLock, "cycle collector reply condvar"),
+          mRunning(PR_FALSE),
+          mShutdown(PR_FALSE),
+          mCollected(PR_FALSE)
+    {
+        NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    }
+
+    PRUint32 Collect(nsICycleCollectorListener* aListener)
+    {
+        NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+
+        MutexAutoLock autoLock(mLock);
+
+        if (!mRunning)
+            return 0;
+
+        nsAutoTPtrArray<PtrInfo, 4000> whiteNodes;
+        if (!mCollector->PrepareForCollection(&whiteNodes))
+            return 0;
+
+        NS_ASSERTION(!mListener, "Should have cleared this already!");
+        mListener = aListener;
+
+        mRequest.Notify();
+        mReply.Wait();
+
+        mListener = nsnull;
+
+        if (mCollected) {
+            mCollected = mCollector->FinishCollection();
+
+            mCollector->CleanupAfterCollection();
+
+            return mCollected ? mCollector->mCollectedObjects : 0;
+        }
+
+        return 0;
+    }
+
+    void Shutdown()
+    {
+        NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+
+        MutexAutoLock autoLock(mLock);
+
+        mShutdown = PR_TRUE;
+
+        if (!mRunning)
+            return;
+
+        mRunning = PR_FALSE;
+        mRequest.Notify();
+        mReply.Wait();
+    }
+};
+
+// Holds a reference.
+static nsCycleCollectorRunner* sCollectorRunner;
+
+// Holds a reference.
+static nsIThread* sCollectorThread;
+
+nsresult
+nsCycleCollector_startup()
+{
+    NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    NS_ASSERTION(!sCollector, "Forgot to call nsCycleCollector_shutdown?");
+
+    sCollector = new nsCycleCollector();
+
+    nsRefPtr<nsCycleCollectorRunner> runner =
+        new nsCycleCollectorRunner(sCollector);
+
+    nsCOMPtr<nsIThread> thread;
+    nsresult rv = NS_NewThread(getter_AddRefs(thread), runner);
+    NS_ENSURE_SUCCESS(rv, rv);
+
+    runner.swap(sCollectorRunner);
+    thread.swap(sCollectorThread);
+
+    return rv;
+}
+
+PRUint32
+nsCycleCollector_collect(nsICycleCollectorListener *aListener)
+{
+    NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    nsCOMPtr<nsICycleCollectorListener> listener(aListener);
+#ifdef DEBUG_CC
+    if (!aListener && sCollector && sCollector->mParams.mDrawGraphs) {
+        listener = new nsCycleCollectorLogger();
+    }
+#endif
+
+    if (sCollectorRunner)
+        return sCollectorRunner->Collect(listener);
+    return sCollector ? sCollector->Collect(1, listener) : 0;
+}
+
+void
+nsCycleCollector_shutdownThreads()
+{
+    NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    if (sCollectorRunner) {
+        nsRefPtr<nsCycleCollectorRunner> runner;
+        runner.swap(sCollectorRunner);
+        runner->Shutdown();
+    }
+
+    if (sCollectorThread) {
+        nsCOMPtr<nsIThread> thread;
+        thread.swap(sCollectorThread);
+        thread->Shutdown();
+    }
+}
+
+void
+nsCycleCollector_shutdown()
+{
+    NS_ASSERTION(NS_IsMainThread(), "Wrong thread!");
+    NS_ASSERTION(!sCollectorRunner, "Should have finished before!");
+    NS_ASSERTION(!sCollectorThread, "Should have finished before!");
+
+    if (sCollector) {
+        sCollector->Shutdown();
+        delete sCollector;
+        sCollector = nsnull;
+    }
+}
