@@ -41,9 +41,6 @@
 """A parser for cross-platform IDL (XPIDL) files."""
 
 import sys, os.path, re
-
-sys.path.append(os.path.join(os.path.dirname(__file__), os.pardir, os.pardir,
-                             'other-licenses', 'ply'))
 from ply import lex, yacc
 
 """A type conforms to the following pattern:
@@ -102,6 +99,22 @@ def paramAttlistToIDL(attlist):
     return '[%s] ' % ', '.join(["%s%s" % (name, value is not None and ' (%s)' % value or '')
                                 for name, value, aloc in sorted])
 
+def unaliasType(t):
+    while t.kind == 'typedef':
+        t = t.realtype
+    assert t is not None
+    return t
+
+def getBuiltinOrNativeTypeName(t):
+    t = unaliasType(t)
+    if t.kind == 'builtin':
+        return t.name
+    elif t.kind == 'native':
+        assert t.specialtype is not None
+        return '[%s]' % t.specialtype
+    else:
+        return None
+
 class BuiltinLocation(object):
     def get(self):
         return "<builtin type>"
@@ -392,7 +405,8 @@ class Native(object):
         'domstring': 'nsAString',
         'utf8string': 'nsACString',
         'cstring': 'nsACString',
-        'astring': 'nsAString'
+        'astring': 'nsAString',
+        'jsval': 'jsval'
         }
 
     def __init__(self, name, nativename, attlist, location):
@@ -626,7 +640,10 @@ class Attribute(object):
     noscript = False
     notxpcom = False
     readonly = False
+    implicit_jscontext = False
     binaryname = None
+    null = None
+    undefined = None
 
     def __init__(self, type, name, attlist, readonly, location, doccomments):
         self.type = type
@@ -645,19 +662,50 @@ class Attribute(object):
                 self.binaryname = value
                 continue
 
-            if value is not None:
-                raise IDLError("Unexpected attribute value", aloc)
-
-            if name == 'noscript':
-                self.noscript = True
-            elif name == 'notxpcom':
-                self.notxpcom = True
+            if name == 'Null':
+                if value is None:
+                    raise IDLError("'Null' attribute requires a value", aloc)
+                if readonly:
+                    raise IDLError("'Null' attribute only makes sense for setters",
+                                   aloc);
+                if value not in ('Empty', 'Null', 'Stringify'):
+                    raise IDLError("'Null' attribute value must be 'Empty', 'Null' or 'Stringify'",
+                                   aloc);
+                self.null = value
+            elif name == 'Undefined':
+                if value is None:
+                    raise IDLError("'Undefined' attribute requires a value", aloc)
+                if readonly:
+                    raise IDLError("'Undefined' attribute only makes sense for setters",
+                                   aloc);
+                if value not in ('Empty', 'Null'):
+                    raise IDLError("'Undefined' attribute value must be 'Empty' or 'Null'",
+                                   aloc);
+                self.undefined = value
             else:
-                raise IDLError("Unexpected attribute '%s'", aloc)
+                if value is not None:
+                    raise IDLError("Unexpected attribute value", aloc)
+
+                if name == 'noscript':
+                    self.noscript = True
+                elif name == 'notxpcom':
+                    self.notxpcom = True
+                elif name == 'implicit_jscontext':
+                    self.implicit_jscontext = True
+                else:
+                    raise IDLError("Unexpected attribute '%s'", aloc)
 
     def resolve(self, iface):
         self.iface = iface
         self.realtype = iface.idl.getName(self.type, self.location)
+        if (self.null is not None and
+            getBuiltinOrNativeTypeName(self.realtype) != '[domstring]'):
+            raise IDLError("'Null' attribute can only be used on DOMString",
+                           self.location)
+        if (self.undefined is not None and
+            getBuiltinOrNativeTypeName(self.realtype) != '[domstring]'):
+            raise IDLError("'Undefined' attribute can only be used on DOMString",
+                           self.location)
 
     def toIDL(self):
         attribs = attlistToIDL(self.attlist)
@@ -677,6 +725,8 @@ class Method(object):
     noscript = False
     notxpcom = False
     binaryname = None
+    implicit_jscontext = False
+    optional_argc = False
 
     def __init__(self, type, name, attlist, paramlist, location, doccomments, raises):
         self.type = type
@@ -703,6 +753,10 @@ class Method(object):
                 self.noscript = True
             elif name == 'notxpcom':
                 self.notxpcom = True
+            elif name == 'implicit_jscontext':
+                self.implicit_jscontext = True
+            elif name == 'optional_argc':
+                self.optional_argc = True
             else:
                 raise IDLError("Unexpected attribute '%s'", aloc)
 
@@ -744,6 +798,8 @@ class Param(object):
     retval = False
     shared = False
     optional = False
+    null = None
+    undefined = None
 
     def __init__(self, paramtype, type, name, attlist, location, realtype=None):
         self.paramtype = paramtype
@@ -763,6 +819,20 @@ class Param(object):
                 if value is None:
                     raise IDLError("'iid_is' must specify a parameter", aloc)
                 self.iid_is = value
+            elif name == 'Null':
+                if value is None:
+                    raise IDLError("'Null' must specify a parameter", aloc)
+                if value not in ('Empty', 'Null', 'Stringify'):
+                    raise IDLError("'Null' parameter value must be 'Empty', 'Null', or 'Stringify'",
+                                   aloc);
+                self.null = value
+            elif name == 'Undefined':
+                if value is None:
+                    raise IDLError("'Undefined' must specify a parameter", aloc)
+                if value not in ('Empty', 'Null'):
+                    raise IDLError("'Undefined' parameter value must be 'Empty' or 'Null'",
+                                   aloc);
+                self.undefined = value
             else:
                 if value is not None:
                     raise IDLError("Unexpected value for attribute '%s'" % name,
@@ -785,6 +855,14 @@ class Param(object):
         self.realtype = method.iface.idl.getName(self.type, self.location)
         if self.array:
             self.realtype = Array(self.realtype)
+        if (self.null is not None and
+            getBuiltinOrNativeTypeName(self.realtype) != '[domstring]'):
+            raise IDLError("'Null' attribute can only be used on DOMString",
+                           self.location)
+        if (self.undefined is not None and
+            getBuiltinOrNativeTypeName(self.realtype) != '[domstring]'):
+            raise IDLError("'Undefined' attribute can only be used on DOMString",
+                           self.location)
 
     def nativeType(self):
         kwargs = {}
